import os

import numpy as np
from skconfig.config import Configuration

# TODO as part of the installation, need ALTIUS_L1PP_USASK_DATABASE environment variable creation
# TODO separate this out into individual configs for each calibration (follow Nick's way)

if 'ALTIUS_L1PP_USASK_DATABASE' not in os.environ:
    raise RuntimeError('Environment variable ALTIUS_L1PP_USASK_DATABASE needs to be set')

user_defined_config_dir = os.path.join(os.environ['ALTIUS_L1PP_USASK_DATABASE'], 'altius_config')
if not os.path.exists(user_defined_config_dir):
    os.makedirs(user_defined_config_dir)

if not os.path.exists(os.path.join(user_defined_config_dir, 'readme.txt')):
    readme = open(os.path.join(user_defined_config_dir, 'readme.txt'), 'w')
    readme.write(
        'Place all variations of .yaml configuration files in this folder.\n'
        'For all altius_l1pp_inflight applications, it is only necessary to pass\n'
        'the name of the .yaml file in, as the software will only look in this directory.')
    readme.close()


class AltiusL1ppInflightConfig:

    def __init__(self, config_name=None, test_config=False):
        self._filename = config_name
        self._config = self._load_config(config_name, test_config)

    def _load_config(self, config_name=None, test_config=False) -> Configuration:
        config_directory = user_defined_config_dir
        if config_name is None:
            config_directory = os.path.dirname(__file__)
            config_name = 'globalkey.yaml'
            if test_config:
                config_name = 'test_globalkey.yaml'
        # import traceback
        # traceback.print_stack()
        config_path = os.path.join(config_directory, config_name)
        print('Loading configuration file: {}'.format(config_path))
        config = Configuration(file_locator=config_path)
        return config

    def _get_wavelength_ranges(self):
        uv_wl_range = self._config.as_array('uv/spectral/wavelength_range')
        vis_wl_range = self._config.as_array('vis/spectral/wavelength_range')
        nir_wl_range = self._config.as_array('nir/spectral/wavelength_range')
        return uv_wl_range, vis_wl_range, nir_wl_range

    def get_wavelength_range(self, channel='uv'):
        if channel not in ['uv', 'vis', 'nir']:
            raise ValueError('Channel must be one of uv, vis, or nir')
        return self._config.as_array('{}/front_aperture_area'.format(channel))

    def get_processed_solar_prnu_file_directory(self):
        processed_dir = self._config.as_string('prnu_calibration/solar_prnu_processed_file_dir')
        return processed_dir

    def get_processed_lunar_prnu_file_directory(self):
        processed_dir = self._config.as_string('prnu_calibration/lunar_prnu_processed_file_dir')
        return processed_dir

    def get_processed_wavelength_reg_file_directory(self):
        processed_file_dir = self._config.as_string('srf_calibration/processed_file_dir')
        return processed_file_dir

    def get_hmi_home_directory(self):
        hmi_home = self._config.as_string('folders/sdo_hmii_data')
        return hmi_home

    def get_detector_pixel_dimensions(self, channel='uv', mode='native'):
        """
            Parameters
            ----------
            channel : str
                'uv', 'vis', or 'nir'

            mode : str
                'native' or 'science'

            Returns
            -------
            [Y, X] dimensions of detector at chosen resolution, overlapping with the science mode
        """

        science_size = self._config.as_array('{}/window/science_native_size'.format(channel))
        science_binning = self._config.as_array('{}/window/science_binning'.format(channel))

        if mode == 'science':
            return science_size

        return science_size * science_binning

    def get_detector_fov(self, channel='uv'):
        """
            Parameters
            ----------

            channel : str, Optional
                'uv', 'vis', or 'nir'

            Returns
            -------
                detector_fov: np.array([float, float])
                    Vertical and horizontal FOV, in degrees, of the detector overlapping with
                    the science window.
        """
        native_size = self._config.as_array('{}/window/native_size'.format(channel))
        science_size = self._config.as_array('{}/window/science_native_size'.format(channel))
        science_origin = self._config.as_array('{}/window/science_native_origin'.format(channel))
        science_binning = self._config.as_array('{}/window/science_binning'.format(channel))
        fov_degrees = self._config.as_array('{}/window/fov_degrees'.format(channel))

        return fov_degrees * (science_origin + science_binning * science_size) / native_size

    def _get_band_from_wavelength(self, wavelength):
        uv_wl_range, vis_wl_range, nir_wl_range = self._get_wavelength_ranges()
        if uv_wl_range[0] <= wavelength <= uv_wl_range[1]:
            band = 'uv'
        elif vis_wl_range[0] <= wavelength <= vis_wl_range[1]:
            band = 'vis'
        elif nir_wl_range[0] <= wavelength <= nir_wl_range[1]:
            band = 'nir'
        else:
            raise RuntimeError('Wavelength {0:3f} outside ALTIUS wavelength ranges'.format(wavelength))
        return band

    def get_max_DN(self):
        return self._config.as_float('prnu_calibration/max_DN_well')

    def get_max_photons_per_sec_per_nm_per_cm2(self):
        return self._config.as_float('prnu_calibration/max_photons_per_sec_per_cm2_per_nm')

    def get_max_lunar_radiance(self):
        return self._config.as_float('prnu_calibration/max_lunar_photons_per_sec_per_cm2_per_nm')

    def get_num_vertical_raster_scans(self):
        return self._config.as_int('prnu_calibration/num_vertical_scans')

    def get_num_horizontal_raster_scans(self):
        return self._config.as_int('prnu_calibration/num_horizontal_scans')

    def get_nominal_satellite_height_km(self):
        return self._config.as_float('orbit/nominal_altitude_km')

    def get_pixel_boresights(self, channel='uv', mode='native'):
        """
            Parameters
            ----------

            channel : str, Optional
                'uv', 'vis', or 'nir'

            mode : str, Optional
                'native' or 'science'

            Returns
            -------
                vertical_boresight, horizontal_boresight: float, float
                    Vertical and horizontal pixel positions of the boresight within the science window
        """
        native_boresight = self._config.as_array('{}/window/native_boresight'.format(channel))
        science_binning = self._config.as_array('{}/window/science_binning'.format(channel))
        science_native_origin = self._config.as_array('{}/window/science_native_origin'.format(channel))

        boresight = native_boresight - science_native_origin

        if mode == 'science':
            boresight /= science_binning

        return boresight[0], boresight[1]

    def get_srf_directory(self):
        return self._config.as_string('srf_calibration/srf_directory')

    def get_aperture_front_area(self, channel='uv'):
        return self._config.as_float('{}/front_aperture_area'.format(channel))

    def get_binning(self, channel='uv'):
        return self._config.as_array('{}/window/science_binning'.format(channel))

    def get_spatial_psf_pathname(self, channel='uv'):
        return self._config.as_pathname('{}/spatial_psf/data_file'.format(channel))

    def get_rolo_wavelengths(self, channel='uv'):
        return self._config.as_array('prnu_image_generation/lunar/{}_rolo_calibration_wavelengths'.format(channel))

    def get_lunar_prnu_num_images(self):
        num_vert = self._config.as_int('prnu_image_generation/lunar/num_vert_positions')
        num_horz = self._config.as_int('prnu_image_generation/lunar/num_horz_positions')
        return num_vert, num_horz

    def get_lunar_phase_angle(self):
        return self._config.as_float('prnu_image_generation/lunar/phase_angle_deg')

    def get_solar_scenario(self):
        return self._config.as_int('prnu_image_generation/solar/scenario')

    def get_solar_prnu_calibration_wavelength(self):
        return self._config.as_float('prnu_image_generation/solar/calibration_wavelength')

    def get_channel(self, calibration_wavelength):
        found_channel = ''
        for channel in ['uv', 'vis', 'nir']:
            wl_range = self._config.as_array('{}/spectral/wavelength_range'.format(channel))
            if wl_range[0] <= calibration_wavelength <= wl_range[1]:
                found_channel = channel

        if found_channel == '':
            raise ValueError('Wavelength ', calibration_wavelength, ' is out of instrument channel bounds')

        return found_channel

    def get_prnu_detector_mode(self):
        return self._config.as_string('prnu_calibration/detector_mode')

    def get_srf_calibration_channel(self):
        return self._config.as_string('srf_calibration/channel')

    def get_wavelength_fitting_band(self):
        return self._config.as_float('srf_calibration/wavelength_fitting_band')

    def get_num_measurements_per_band(self):
        return self._config.as_int('srf_calibration/num_measurements_per_band')

    def get_srf_calibration_wavelengths(self, channel: str):
        return self._config.as_array('srf_calibration/{}/wavelengths'.format(channel))

    def get_srf_detector_mode(self):
        return self._config.as_string('srf_calibration/detector_mode')

    def get_srf_measurement_wavelengths(self, channel: str):
        cal_wl = self.get_srf_calibration_wavelengths(channel)
        band = self.get_wavelength_fitting_band()
        n = self.get_num_measurements_per_band()

        all_measurement_wavelengths = np.zeros((len(cal_wl), n))

        for k, wl in enumerate(cal_wl):
            meas = np.linspace(wl - band / 2, wl + band / 2, n)
            all_measurement_wavelengths[k] = meas

        return all_measurement_wavelengths

    def get_processed_solar_srf_file_directory(self):
        return self._config.as_string('srf_calibration/processed_file_dir')

    def get_solar_srf_detector_positions(self):
        # TODO will likely make a different class for handling SRF positions
        vsteps = self._config.as_array('srf_calibration/unit_vertical_positions')
        hsteps = self._config.as_array('srf_calibration/unit_horizontal_positions')

        channel = self.get_srf_calibration_channel()
        mode = self.get_srf_detector_mode()
        pixel_size = self.get_detector_pixel_dimensions(channel=channel, mode=mode)
        horizontal_width = pixel_size[1]
        vertical_width = pixel_size[0]

        detector_positions = []

        for uv, uh in zip(vsteps, hsteps):
            detector_positions.append((uv * vertical_width, uh * horizontal_width))

        detector_positions = np.asarray(detector_positions)

        return detector_positions

    def get_srf_center_pixel_radius(self):
        return self._config.as_int('srf_calibration/center_pixels_radius')

    def get_config_source_filename(self):
        return self._filename

    def get_satellite_ascending_node_hours(self) -> float:
        hours = self._config.as_float('orbit/ascending_node_hours')
        minutes = self._config.as_float('orbit/ascending_node_minutes')
        seconds = self._config.as_float('orbit/ascending_node_seconds')
        return hours + minutes / 60 + seconds / 3600
