import csv
import os
import sys
from typing import Tuple, Union

import numpy as np
import scipy
import xarray as xr
from scipy.interpolate import interp2d

from altius_l1pp_inflight.instrument.srf.srf_interface import SrfInterface

RESOLUTION = 0.04


class AltiusAotfSrf(SrfInterface):

    def __init__(self, data_directory):
        self._data_directory = data_directory
        self._tunable_wavelengths, self._wavelength_grid, self._aotf_srf, self._f = self._load_aotf_srf()

    def _load_aotf_srf(self):
        csv.field_size_limit(int(sys.maxsize / 10000000000))
        data = []
        fname = os.path.join(self._data_directory, 'aotf_srf.csv')
        with open(fname) as f:
            rd = csv.reader(f)
            for i, row in enumerate(rd):
                line = []
                # x = row[0].split()
                for v in row:
                    line.append(float(v))
                data.append(line)

        data = np.array(data)

        tunable_wavelengths = data[1:, 0]
        wavelengths = data[0, 1:]
        srf = data[1:, 1:]

        interp_wl_grid = np.arange(-40, 40, step=0.02)
        z = np.zeros((len(tunable_wavelengths), len(interp_wl_grid)))
        for idx, wl in enumerate(tunable_wavelengths):
            wl_grid = wavelengths - wl
            srf_grid = np.interp(interp_wl_grid, wl_grid, srf[idx, :])
            z[idx] = srf_grid

        f = interp2d(interp_wl_grid, tunable_wavelengths, z)
        return tunable_wavelengths.flatten(), wavelengths.flatten(), srf, f

    # TODO need to extract this yet
    def _load_detector_qe(self, wavelengths) -> np.ndarray:
        fname = os.path.join(self._data_directory, 'SRF_uv_data.mat')
        srf_data_dictionary = scipy.io.loadmat(fname)
        data = srf_data_dictionary["SRF"]
        wl = srf_data_dictionary["SRF_WL"].flatten() * 1e9

        detector_srf = data[0][5][:, 0, 0].flatten()
        qe = np.interp(wavelengths.flatten(), wl, detector_srf)
        qe = qe[:, np.newaxis]
        return qe

    def get_srf(self, center_wavelength_nm, new_wavelengths=None,
                interp_method=SrfInterface.LINEAR_INTERP, normalize=True) -> Tuple[xr.DataArray, float]:
        """
        Returns the SRF closest to the specified wavelength, and the actual wavelength

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength to obtain the SRF from the tunable wavelengths

        new_wavelengths : np.ndarray, Optional
            The grid of wavelengths, in nm, to evaluate/interpolate the SRF over.  If not
            specified, will return the SRF at the original measured wavelengths in wavelength_grid

        interp_method : str, Optional
            'linear' or 'log'.  Interpolation method for evaluating the SRF over new_wavelengths

        normalize : bool, Optional
            If True, normalizes the integral of the SRF to be 1.0.
            If False, the values of the SRF are adjusted so that the integral matches the original
            data's integral

        Returns
        -------
        +srf : xr.DataArray
            The SRF with tuned wavelength near to center_wavelength_nm.  x-coordinate 'wavelength' set
            by the new_wavelengths parameter
        actual_wavelength : float
            The nearest tunable wavelength to the input center_wavelength_nm
        """
        # tunable_idx = np.abs(self._tunable_wavelengths - center_wavelength_nm).argmin()
        # srf = self._aotf_srf[tunable_idx, :]  # * self._detector_qe * self._optics_srf

        wl = np.arange(-40, 40, step=0.02)

        srf = self._f(wl, center_wavelength_nm)

        wl += center_wavelength_nm

        orig_sum = np.sum(srf)

        if new_wavelengths is not None:
            if interp_method is self.LINEAR_INTERP:
                srf = np.interp(new_wavelengths.flatten(), wl, srf.flatten(), left=0,
                                right=0)
            else:
                log_srf = np.log(srf.flatten())
                log_srf = np.interp(new_wavelengths.flatten(), wl, log_srf, left=0,
                                    right=0)
                srf = np.exp(log_srf)

            srf = srf * orig_sum / np.sum(srf)  # make sure it scales to the original SRF data
            srf = xr.DataArray(srf, dims={'wavelength'}, coords={'wavelength': new_wavelengths})
        else:
            srf = xr.DataArray(srf, dims={'wavelength'}, coords={'wavelength': wl})

        if normalize:
            srf /= np.sum(srf)

        return srf, center_wavelength_nm  # self._tunable_wavelengths[tunable_idx]

    def convolve_spectrum_with_srf(self, convolution_wavelengths: Union[float, np.ndarray],
                                   spectrum_wavelengths, spectrum) -> np.ndarray:
        """
            Convolves the spectrum with the SRF at each wavelength.

            Parameters
            ----------
            spectrum_wavelengths : np.ndarray
                Wavelengths in nm of the input spectrum

            spectrum : np.ndarray
                Input high-resolution spectrum to be convolved with the corrected SRF

            convolution_wavelengths : Union[float, np.ndarray]
                Wavelengths (in nm) to evaluate the resulting convolution at

            Returns
            -------
            convolved_spectrum : np.ndarray
                The resulting spectrum convolved with SRFs at each convolution wavelength
        """
        convolved_spectrum = np.zeros_like(convolution_wavelengths)
        for k, w in enumerate(convolution_wavelengths):
            srf_w, _ = self.get_srf(w, spectrum_wavelengths, normalize=True)
            srf_contribution_spectrum = srf_w.values * spectrum
            srf_contribution = np.sum(srf_contribution_spectrum)
            convolved_spectrum[k] = srf_contribution
        return convolved_spectrum

    def get_wavelength_grid(self, resolution=None):
        default_grid = np.linspace(390, 750, int(410 / RESOLUTION) + 1)
        if resolution is None:
            return default_grid

        return np.arange(390, 750, step=resolution)

    def get_fwhm(self, wavelengths: np.ndarray):
        raise NotImplementedError
