import os

import click
import numpy as np

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.interfaces.lunar_prnu import SpsToUsaskLunarPrnuSceneInterface, LunarPrnuSpotCheckOutput
from altius_l1pp_inflight.scene.disk_geometry import LunarGeometry
from altius_l1pp_inflight.scene.disk_geometry import center_solar_disk
from altius_l1pp_inflight.scene.generation.lunar.lunar_calibration_util import get_lunar_observations_directory, \
    get_lunar_prnu_directory
from altius_l1pp_inflight.scene.generation.lunar.rolo_wrapper import RoloModel
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf


@click.command()
@click.option('-i', '--input_measurement_file', type=str,
              help='Name of .nc file containing the lunar measurements for PRNU, contained in '
                   'ALTIUS_L1PP_DATABASE director for lunar measurements')
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-o', '--output_file', type=str,
              help='Output filename to be stored in the ALTIUS_L1PP_DATABASE directory for processed lunar PRNU files')
@click.option('-c', '--channel', type=str,
              help='uv, vis, or nir')
def cli(input_measurement_file, config_file, output_file, channel):
    config = altius_config.AltiusL1ppInflightConfig(config_name=config_file)
    run_lunar_prnu_cal(input_measurement_file, output_file, config, channel)


def run_lunar_prnu_cal(input_file, output_file, config: altius_config.AltiusL1ppInflightConfig, channel):
    scene_interface = SpsToUsaskLunarPrnuSceneInterface()
    lunar_directory = get_lunar_observations_directory(config, channel)
    input_file_path = os.path.join(lunar_directory, input_file)
    scene_interface.open_read_dataset(input_file_path)
    channel = scene_interface.get_channel(config)
    mode = config.get_prnu_detector_mode()
    input_dir = get_lunar_observations_directory(config, channel)

    input_path = os.path.join(input_dir, input_file)
    scene_interface.open_read_dataset(input_path)

    num_scenes = scene_interface.get_num_scenes()
    num_vert_pixels, num_horz_pixels = scene_interface.get_num_pixels_scene()

    output_dir = get_lunar_prnu_directory(config, channel)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_path = os.path.join(output_dir, output_file)

    max_radiance = config.get_max_lunar_radiance()
    max_DN = config.get_max_DN()
    cf = max_radiance / max_DN

    # +psf = PsfSimpleGaussian(2.1)

    vb, hb = config.get_pixel_boresights(channel, mode)
    v_fov, h_fov = config.get_detector_fov(channel)
    lunar_geometry = LunarGeometry(num_horz_pixels, num_vert_pixels, vb, hb, h_fov, v_fov)

    # TODO divide out current PRNU

    calibration_wavelengths = scene_interface.get_wavelengths()
    num_pos_vert, num_pos_horz = config.get_lunar_prnu_num_images()
    grid_pixel_positions = []
    for row in range(0, num_pos_vert):
        for col in range(0, num_pos_horz):
            grid_pixel_positions.append((int((row + 0.5) * num_vert_pixels / num_pos_vert),
                                         int((col + 0.5) * num_horz_pixels / num_pos_horz)))

    lunar_prnu_output = LunarPrnuSpotCheckOutput(output_path, num_vert_pixels, num_horz_pixels,
                                                 wavelengths=calibration_wavelengths,
                                                 grid_pixel_positions=np.asarray(grid_pixel_positions))
    click.secho('WARNING: PSF blurring not included in current version', fg='yellow')
    click.secho('WARNING: SRF assumed Gaussian - real data not incorporated yet', fg='yellow')
    for wl_idx, wl in enumerate(calibration_wavelengths):

        click.echo('Processing wavelength {}, {}/{}'.format(wl, wl_idx + 1, len(calibration_wavelengths)))

        for scene_index in range(0, num_scenes):

            click.echo('    PRNU spot check for scene {}/{}', scene_index + 1, num_scenes)
            altius_scene = scene_interface.load_scene(wl, scene_index)
            r, hc, vc, alg_points = center_solar_disk(
                altius_scene.get_altius_image())  # TODO change this over when ready
            scene_irradiance = altius_scene.compute_irradiance(hc, vc, cf)

            # TODO temporary:
            grid_wl = np.arange(wl - 50, wl + 50, step=0.1)
            srf = GaussianSrf(grid_wl, 3.0 * np.ones_like(grid_wl))

            rolo = RoloModel()
            irradiance_spectrum_moon = []
            lunar_wl = np.arange(wl - 30, wl + 30, step=0.1)

            for wavelen in lunar_wl:
                irrad = rolo.irradiance(wavelength_nm=wavelen,
                                        distance_moon_sun_au=altius_scene.get_moon_sun_distance_au(),
                                        distance_moon_spacecraft_km=altius_scene.get_satellite_to_moon_distance_km(),
                                        phase_angle_deg=altius_scene.get_lunar_phase_angle_deg(),
                                        selen_latitude_deg=altius_scene.get_subsatellite_selenographic_latitude(),
                                        selen_longitude_deg=altius_scene.get_subsatellite_selenographic_longitude(),
                                        selen_solar_longitude_deg=altius_scene.get_subsolar_selenographic_longitude())
                irradiance_spectrum_moon.append(irrad)

            ideal_lunar_irradiance = srf.convolve_spectrum_with_srf(np.array([wl]),
                                                                    lunar_wl, np.array(irradiance_spectrum_moon))

            ratio = scene_irradiance / ideal_lunar_irradiance

            lunar_geometry.set_disk(altius_scene.get_expected_altius_lunar_radius(), hc, vc,
                                    altius_scene.get_moon_sun_distance_au())
            pixel_mask = lunar_geometry.disk_mask()

            # TODO error estimate?
            lunar_prnu_output.write_next_ratio_and_mask(ratio, ratio * 0.01,
                                                        pixel_mask.astype(int), (vc, hc), wl)

    lunar_prnu_output.close()
    click.secho('*******************************************', fg='green')
    click.secho('Wrote lunar PRNU spot check file to {}'.format(output_path), fg='green')
    click.secho('*******************************************', fg='green')


# Temporary place for CLI
if __name__ == '__main__':
    run_lunar_prnu_cal(input_file='lunar_test_measurements.nc',
                       output_file='lunar_prnu_test.nc',
                       config=altius_config.AltiusL1ppInflightConfig(),
                       channel='uv')
