import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import minimize, minimize_scalar

from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977


def find_best_center(solar_image,
                     solar_disk_pixel_parameters,
                     raster_pixel_bounds,
                     wavelength,
                     solar_geometry: SolarGeometry,
                     distance_to_sun_km: float,
                     optimize_radius=False):
    """
    Parameters
    ----------
    solar_image : np.ndarray
        m x n 2D array of the detector image with the solar disk

    solar_disk_pixel_parameters : tuple
        (vertical_center_pixels, horizontal_center_pixels, radius_pixels) current knowledge of the solar
        disk using other methods

    raster_pixel_bounds : tuple
        (lower_left_vertical_pixel, lower_left_horizontal_pixel, vertical_width, horizontal_width)
        evaluation area of the limb-flattening

    wavelength : float
        Wavelength in nm

    solar_geometry : SolarGeometry
        Instance of class used to specify the solar disk center and radius and all related pixel coordinates

    distance_to_sun_km : float

    optimize_radius : bool, Optional
        If true, try to find the best radius that will produce the flattest region

    Returns
    -------
    result: dict
        vertical_center : float

        horizontal_center : float

        radius : float

    """

    limb_darkening_model = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1, 50)
    ld_profile = limb_darkening_model.limb_darkening(wavelength, mu_profile)

    llv = raster_pixel_bounds[0]
    llh = raster_pixel_bounds[1]
    rwv = raster_pixel_bounds[2]
    rwh = raster_pixel_bounds[3]

    vcc = solar_disk_pixel_parameters[0]
    hcc = solar_disk_pixel_parameters[1]
    rr = solar_disk_pixel_parameters[2]

    optimal_center = minimize(fun=compute_entropy_center_find, x0=np.array([vcc, hcc]),
                              method='Powell',
                              args=(mu_profile, ld_profile, solar_image, raster_pixel_bounds,
                                    rr, solar_geometry, distance_to_sun_km),
                              bounds=((llv, llv + rwv), (llh, llh + rwh)))

    vertical_center = optimal_center.x[0]
    horizontal_center = optimal_center.x[1]

    if optimize_radius:
        optimal_radius = minimize_scalar(fun=compute_adhoc_corner_gradient,
                                         method='Bounded',
                                         args=(mu_profile, ld_profile, solar_image, raster_pixel_bounds,
                                               (vertical_center, horizontal_center), solar_geometry,
                                               distance_to_sun_km),
                                         bounds=(0.95 * rr, 1.05 * rr))

        radius = optimal_radius.x
    else:
        radius = rr

    result = {'vertical_center': vertical_center,
              'horizontal_center': horizontal_center,
              'radius': radius}

    return result


def compute_adhoc_corner_gradient(radius_guess, mu_profile, ld_profile, solar_image, raster_pixel_bounds, center,
                                  solar_geometry: SolarGeometry, distance_to_sun_km: float):
    llv = raster_pixel_bounds[0]
    llh = raster_pixel_bounds[1]
    rwv = raster_pixel_bounds[2]
    rwh = raster_pixel_bounds[3]

    vc = center[0]
    hc = center[1]

    solar_geometry.set_disk(radius_guess, hc, vc, distance_to_sun_km)
    mu = solar_geometry._mu()

    ld = np.interp(mu[llv:(llv + rwv), llh:(llh + rwh)], mu_profile, ld_profile)
    f = solar_image[llv:(llv + rwv), llh:(llh + rwh)] / ld
    midv = int(vc - llv)
    midh = int(hc - llh)
    # Account for noise by averaging over a small range of pixels (TODO determine range of pixels)
    midval = np.mean(f[midv - 4:midv + 4, midh - 4:midh + 4])
    g1 = np.mean(f[0:8, 0:8]) - midval
    g2 = np.mean(f[-9:-1, 0:8]) - midval
    g3 = np.mean(f[0:8, -9:-1]) - midval
    g4 = np.mean(f[-9:-1, -9:-1]) - midval
    return np.sqrt(g1 ** 2 + g2 ** 2 + g3 ** 2 + g4 ** 2)


def compute_entropy_center_find(center_guess, mu_profile, ld_profile, solar_image,
                                raster_pixel_bounds, radius, solar_geometry, distance_to_sun_km):
    guess_vals = (center_guess[0], center_guess[1], radius)
    return compute_entropy(guess_vals, mu_profile, ld_profile, solar_image, raster_pixel_bounds,
                           solar_geometry, distance_to_sun_km)


def compute_entropy(guess_vals, mu_profile, ld_profile, solar_image, raster_pixel_bounds, solar_geometry: SolarGeometry,
                    distance_to_sun_km: float):
    llv = raster_pixel_bounds[0]
    llh = raster_pixel_bounds[1]
    rwv = raster_pixel_bounds[2]
    rwh = raster_pixel_bounds[3]

    vc = guess_vals[0]
    hc = guess_vals[1]
    rr = guess_vals[2]

    solar_geometry.set_disk(rr, hc, vc, distance_to_sun_km)
    mu = solar_geometry.mu()
    ld = np.interp(mu[llv:(llv + rwv), llh:(llh + rwh)], mu_profile, ld_profile)
    f = solar_image[llv:(llv + rwv), llh:(llh + rwh)] / ld
    return entropy(f)


def entropy(array_in):
    return np.nansum(array_in * np.log(array_in))


def plot_entropy_map(solar_image,
                     image_angular_size_deg,
                     solar_disk_pixel_parameters,
                     raster_pixel_bounds,
                     angular_uncertainty_degrees,
                     pixel_res,
                     wavelength,
                     expected_center,
                     found_center,
                     solar_geometry: SolarGeometry,
                     distance_to_sun_km: float):
    """
    Parameters
    ----------
    solar_image : np.ndarray
        m x n 2D array of the detector image with the solar disk

    image_angular_size_deg : tuple
        (vertical_size_degrees, horizontal_size_degrees) mapped to the m x n image array

    solar_disk_pixel_parameters : tuple
        (vertical_center_pixels, horizontal_center_pixels, radius_pixels) current knowledge of the solar
        disk using other methods

    raster_pixel_bounds : tuple
        (lower_left_vertical_pixel, lower_left_horizontal_pixel, vertical_width, horizontal_width)
        evaluation area of the limb-flattening

    angular_uncertainty_degrees : float
        Accuracy of the center passed in above - used to determine the size of the array of search points
        for the optimal center

    pixel_res : float

    wavelength : float

    expected_center : tuple
        (vert, horz) pixel coordinates of expected center from known data

    found_center : tuple
        (vert, horz) pixel coordinates found by simple centering algorithm

    """

    limb_darkening_model = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1, 50)
    ld_profile = limb_darkening_model.limb_darkening(wavelength, mu_profile)

    llv = raster_pixel_bounds[0]
    llh = raster_pixel_bounds[1]
    rwv = raster_pixel_bounds[2]
    rwh = raster_pixel_bounds[3]

    vcc = solar_disk_pixel_parameters[0]
    hcc = solar_disk_pixel_parameters[1]
    rr = solar_disk_pixel_parameters[2]

    m = len(solar_image)
    n = len(solar_image[0])

    angular_uncertainty_pixels = m / image_angular_size_deg[0] * angular_uncertainty_degrees

    center_offsets = np.arange(-angular_uncertainty_pixels, angular_uncertainty_degrees, pixel_res)

    entropy_map = np.zeros((len(center_offsets), len(center_offsets)))

    for ih, dh in enumerate(center_offsets):
        for iv, dv in enumerate(center_offsets):
            hc = hcc + dh
            vc = vcc + dv
            solar_geometry.set_disk(rr, hc, vc, distance_to_sun_km)
            mu = solar_geometry._mu()
            ld = np.interp(mu, mu_profile, ld_profile)
            f = solar_image / ld
            entropy_map[iv, ih] = entropy(f[llv:(llv + rwv), llh:(llh + rwh)])

    min_idx = np.unravel_index(np.argmin(entropy_map), np.shape(entropy_map))

    vertical_center = center_offsets[min_idx[0]] + vcc
    horizontal_center = center_offsets[min_idx[1]] + hcc

    fig, ax = plt.subplots()

    lh = pixel_res * float(len(entropy_map[0])) / 2
    lv = pixel_res * float(len(entropy_map)) / 2
    V, H = np.meshgrid(np.arange(-lv, lv, step=pixel_res, dtype=float),
                       np.arange(-lh, lh, step=pixel_res, dtype=float))
    V += vertical_center
    H += horizontal_center

    im = ax.contourf(V, H, entropy_map)
    fig.colorbar(im)
    ax.plot(horizontal_center, vertical_center, 'r.')
    ax.plot(expected_center[1], expected_center[0], 'mx')
    ax.plot(found_center[1], found_center[0], 'go')
    plt.legend(['Centering algorithm', 'Actual Center', 'Entropy Min. Center'])
    plt.title('Location of Entropy Minimum vs. Actual Center')
    ax.plot()

    plt.show()
