import numpy as np

from altius_l1pp_inflight.scene.disk_geometry import LunarGeometry, SolarGeometry

"""
This module will include different ways to describe the ALTIUS
scenes used for calibration.  At their core they contain a 2D
array of DN values, and include the effects of the SRF


Set of classes and methods for describing an image taken by the
ALTIUS detector.  Still a work in progress.


TODO Can be used to build further classes that show different stages of calibration correction
"""


class AltiusSolarScene:

    def __init__(self,
                 altius_image,
                 detector_width_deg,
                 satellite_to_sun_distance_km,
                 expected_altius_solar_center_horz,
                 expected_altius_solar_center_vert,
                 expected_altius_solar_radius,
                 wavelength,
                 boresight,
                 integration_time,
                 start_pointing_vector,
                 end_pointing_vector,
                 time):
        self._altius_image = altius_image
        self._detector_width_deg = detector_width_deg
        self._satellite_to_sun_distance_km = satellite_to_sun_distance_km
        self._expected_altius_solar_center_horz = expected_altius_solar_center_horz
        self._expected_altius_solar_center_vert = expected_altius_solar_center_vert
        self._expected_altius_solar_radius = expected_altius_solar_radius
        self._wavelength = wavelength
        self._boresight = boresight
        self._integration_time = integration_time
        self._starting_pointing_vector = start_pointing_vector
        self._end_pointing_vector = end_pointing_vector
        self._time = time

    def get_time(self) -> int:
        return self._time

    def get_start_boresight_pointing_vector(self):
        return self._starting_pointing_vector

    def get_end_boresight_pointing_vector(self):
        return self._end_pointing_vector

    def get_integration_time(self):
        return self._integration_time

    def get_boresight(self):
        return self._boresight

    def get_wavelength(self):
        return self._wavelength

    def get_altius_image(self):
        return self._altius_image

    def get_detector_width_deg(self):
        return self._detector_width_deg

    def get_satellite_to_sun_distance_km(self):
        return self._satellite_to_sun_distance_km

    def get_expected_altius_solar_center_horz(self):
        return self._expected_altius_solar_center_horz

    def get_expected_altius_solar_center_vert(self):
        return self._expected_altius_solar_center_vert

    def get_expected_altius_solar_radius(self):
        return self._expected_altius_solar_radius

    # TODO deprecated, I think
    # def compute_irradiance(self, horizontal_center, vertical_center, conversion_factor, mask_radius=None):
    #     """
    #         Compute the irradiance of this solar disk for the assumed center of the sun,
    #         and with an optional mask applied to only get a portion of the disk.
    #     """
    #
    #     h = np.arange(0.5, len(self._altius_image), step=1.0)
    #     v = np.arange(0.5, len(self._altius_image[0]), step=1.0)
    #     solar_geometry = SolarGeometry(len(self._altius_image[0]), len(self._altius_image), self._boresight[0],
    #                                    self._boresight[1],
    #                                    self._detector_width_deg[0], self._detector_width_deg[1])
    #     solar_geometry.set_disk(self._expected_altius_solar_radius, self._expected_altius_solar_center_horz,
    #                             self._expected_altius_solar_center_vert, self._satellite_to_sun_distance_km)
    #     delta_omega = solar_geometry.delta_omega_view()
    #     rad = np.copy(self._altius_image)
    #     vv, hh = np.meshgrid(h, v)
    #     pixel_radii = np.sqrt((vv - vertical_center) ** 2 + (hh - horizontal_center) ** 2)
    #
    #     if mask_radius is not None:
    #         mask = pixel_radii <= mask_radius
    #     else:
    #         mask = pixel_radii <= self._expected_altius_solar_radius
    #
    #     rad *= mask
    #
    #     return np.nansum(rad * delta_omega) * conversion_factor, mask


class AltiusLunarScene:
    def __init__(self,
                 altius_image,
                 detector_width_deg,
                 satellite_to_moon_distance_km,
                 expected_altius_lunar_center_horz,
                 expected_altius_lunar_center_vert,
                 expected_altius_lunar_radius,
                 wavelength,
                 boresight,
                 integration_time,
                 start_pointing_vector,
                 end_pointing_vector,
                 sun_moon_distance_au,
                 lunar_phase_angle_deg,
                 subsatellite_selenographic_latitude,
                 subsatellite_selenographic_longitude,
                 subsolar_selenographic_longitude):
        self._subsolar_selenographic_longitude = subsolar_selenographic_longitude
        self._subsatellite_selenographic_longitude = subsatellite_selenographic_longitude
        self._subsatellite_selenographic_latitude = subsatellite_selenographic_latitude
        self._lunar_phase_angle_deg = lunar_phase_angle_deg
        self._moon_sun_distance_au = sun_moon_distance_au
        self._altius_image = altius_image
        self._detector_width_deg = detector_width_deg
        self._satellite_to_moon_distance_km = satellite_to_moon_distance_km
        self._expected_altius_lunar_center_horz = expected_altius_lunar_center_horz
        self._expected_altius_lunar_center_vert = expected_altius_lunar_center_vert
        self._expected_altius_lunar_radius = expected_altius_lunar_radius
        self._wavelength = wavelength
        self._boresight = boresight
        self._integration_time = integration_time
        self._starting_pointing_vector = start_pointing_vector
        self._end_pointing_vector = end_pointing_vector

    def get_start_boresight_pointing_vector(self):
        return self._starting_pointing_vector

    def get_end_boresight_pointing_vector(self):
        return self._end_pointing_vector

    def get_integration_time(self):
        return self._integration_time

    def get_boresight(self):
        return self._boresight

    def get_wavelength(self):
        return self._wavelength

    def get_altius_image(self):
        return self._altius_image

    def get_detector_width_deg(self):
        return self._detector_width_deg

    def get_satellite_to_moon_distance_km(self):
        return self._satellite_to_moon_distance_km

    def get_expected_altius_lunar_center_horz(self):
        return self._expected_altius_lunar_center_horz

    def get_expected_altius_lunar_center_vert(self):
        return self._expected_altius_lunar_center_vert

    # TODO some work needs to be done with the lunar center, since it's not perfectly circular
    def get_expected_altius_lunar_radius(self):
        return self._expected_altius_lunar_radius

    def compute_irradiance(self, horizontal_center, vertical_center, conversion_factor):
        """
            Compute the irradiance of this lunar disk for the assumed center of the moon,
            and with an optional mask applied to only get a portion of the disk.
        """

        h = np.arange(0.5, len(self._altius_image), step=1.0)
        v = np.arange(0.5, len(self._altius_image[0]), step=1.0)
        lunar_geometry = LunarGeometry(len(self._altius_image[0]), len(self._altius_image),
                                       self._boresight[0], self._boresight[1],
                                       self._detector_width_deg[0], self._detector_width_deg[1])

        lunar_geometry.set_disk(self._expected_altius_lunar_radius,
                                self._expected_altius_lunar_center_horz,
                                self._expected_altius_lunar_center_vert,
                                self._satellite_to_moon_distance_km)
        delta_omega = lunar_geometry.delta_omega_view()

        rad = np.copy(self._altius_image)
        mask = lunar_geometry.disk_mask()

        rad *= mask

        return np.nansum(rad * delta_omega) * conversion_factor

    def get_moon_sun_distance_au(self):
        return self._moon_sun_distance_au

    def get_lunar_phase_angle_deg(self):
        return self._lunar_phase_angle_deg

    def get_subsatellite_selenographic_latitude(self):
        return self._subsatellite_selenographic_latitude

    def get_subsatellite_selenographic_longitude(self):
        return self._subsatellite_selenographic_longitude

    def get_subsolar_selenographic_longitude(self):
        return self._subsolar_selenographic_longitude
