from datetime import timedelta

import astropy.coordinates
import numpy as np
import sasktran as sk
from arg_time import ut_to_mjd, datetime64_to_datetime
from astropy import units as u
from astropy.coordinates import Longitude, Latitude, EarthLocation, solar_system, SkyCoord
from astropy.time import Time
from astropy.units import Quantity
from skretrieval.platforms.satellite import SatelliteSunSync
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap

from altius_l1pp_inflight.configuration.altius_config import AltiusL1ppInflightConfig

start_utc_time = datetime64_to_datetime(np.datetime64('2019-07-26 19:59:00'))  # TODO get from input
config = AltiusL1ppInflightConfig()

calibration_height_m = 60000  # TODO from config

satellite_nominal_altitude_m = config.get_nominal_satellite_height_km() * 1000.0
ascending_node_time_hours = config.get_satellite_ascending_node_hours()

# Compute the points on a complete orbit:
satellite = SatelliteSunSync(utc=start_utc_time,
                             orbittype='sgp4',
                             period_from_altitude=satellite_nominal_altitude_m,
                             localtime_of_ascending_node_hours=ascending_node_time_hours)
period: timedelta = satellite.period()
orbits_per_day = 86400 / period.total_seconds()
images_per_orbit = 360  # 2 per deg. latitude
num_images = int(images_per_orbit * orbits_per_day)
sat_earthlocation_positions = []
time_per_image = 86400 / num_images
dt = timedelta(seconds=time_per_image)

geometry = sk.geometry.Geometry()

lats = []
lons = []

for i in range(num_images):
    tnow = start_utc_time + i * dt
    mjd = ut_to_mjd(tnow)
    satellite.update_position(tnow)

    sun_itrs: SkyCoord = solar_system.get_body('sun',
                                               Time(mjd, format='mjd'))  # location=None assumes a geocentric observer

    sun_geo = sk.Geodetic()
    sun_geo.from_lat_lon_alt(sun_itrs.dec.value, sun_itrs.ra.value, 0)
    llh = satellite.lat_lon_height

    lats.append(llh[0])
    lons.append(llh[1])

    # print('Latitude: {}, Longitude: {}, Altitude: {}'.format(llh[0], llh[1], llh[2]))
    sat_gcrs: EarthLocation = astropy.coordinates.EarthLocation.from_geodetic(
        lon=Longitude(angle=llh[1], unit=u.deg),
        lat=Latitude(angle=llh[0], unit=u.deg),
        height=Quantity(value=llh[2], unit=u.meter))

    sat_to_earth_itrs: SkyCoord = solar_system.get_body('earth', Time(mjd, format='mjd'), location=sat_gcrs)
    sat_to_earth_itrs.representation_type = 'cartesian'
    sat_itrs = sat_to_earth_itrs.transform_to('itrs')
    # sat_itrs = sat_gcrs.to_geocentric()

    chi = sat_itrs.separation(sun_itrs)

    # finding the intersection of the 94.5 sza ring with the 60 km tangent altitude ring

    # Determine the co-latitude of the 60 km tangent point assuming radius of the earth is the same as at the
    #  satellite sub-point
    sat_geo = sk.Geodetic()
    sat_geo.from_lat_lon_alt(llh[1], llh[0], 0)
    earth_radius = np.linalg.norm(sat_geo.location)
    tp_radius = earth_radius + calibration_height_m
    sat_radius = earth_radius + satellite_nominal_altitude_m

    y_hat = np.cross(sat_geo.location, sun_geo.location)
    y_hat /= np.linalg.norm(y_hat)
    z_hat = sat_geo.local_up
    x_hat = np.cross(y_hat, z_hat)

    # assumes the radius of the earth doesn't vary significantly between the satellite and tangent point - which is reasonable when we're looking across the terminator, but less so when looking north/south
    # for those cases - will double-check the calculations and adjust the TP as needed to give the correct look vector
    phi = np.arccos(tp_radius / sat_radius)

    phi_prime = np.deg2rad(94.52)  # this is a starting point

    cos_theta_prime = (np.cos(phi_prime) * np.cos(chi) - np.cos(phi)) / (np.sin(phi_prime) * np.sin(chi))

    theta = np.arccos(
        (cos_theta_prime * np.sin(phi_prime) * np.cos(chi) + np.cos(phi_prime) * np.sin(chi)) / np.sin(phi))

    # tangent point intersection of satellite 60 km TP circle and solar 94.5 deg circle (assuming spherical
    # earth same radius as sub-satellite point)
    r_bar = np.cos(theta) * np.sin(phi) * x_hat + np.sin(theta) * np.cos(phi) * y_hat + np.cos(phi) * z_hat

    r_geo = sk.Geodetic()
    r_geo.from_xyz(r_bar)

    r_geo.from_lat_lon_alt(r_geo.latitude, r_geo.longitude, calibration_height_m)

    sza = np.rad2deg(np.arccos(np.dot(r_geo.location, sun_geo.location) /
                               (np.linalg.norm(r_geo.location) * np.linalg.norm(sun_geo.location))))

    # if 94 < sza < 96:
    #     print('Tangent point SZA: {} deg'.format(sza))

map = Basemap(projection='mill', resolution='l', lon_0=np.mean(lons), lat_0=0)  # ortho gives a globe
map.drawcoastlines(linewidth=0.25)
map.fillcontinents(color='coral', lake_color='aqua')
map.drawmapboundary(fill_color='aqua')
map.drawmeridians(np.arange(0, 360, 30))
map.drawparallels(np.arange(-90, 90, 30))
map.plot(lats, lons)
CS = map.nightshade(start_utc_time)
plt.show()
