import os

import click
import numpy as np
from matplotlib import image
from skimage.transform import rescale

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.instrument.psf.psf_interface import PsfSimpleGaussian
from altius_l1pp_inflight.scene.interfaces.lunar_prnu import SpsToUsaskLunarPrnuSceneInterface
from altius_l1pp_inflight.scene.disk_geometry import LunarGeometry, LUNAR_RADIUS, EARTH_TO_MOON_DISTANCE_KM
from altius_l1pp_inflight.scene.generation.lunar.lunar_calibration_util import get_lunar_observations_directory
from altius_l1pp_inflight.scene.generation.lunar.rolo_wrapper import RoloModel
from altius_l1pp_inflight.scene.lunar_geometry import center_lunar_image
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf


def get_lunar_positioning_geometry(measurement_time, satellite_position, config):
    # TODO this might go in the calibration portion (or some util section), to determine the
    #  lunar lat-lon coordinates from astropy
    phase_angle_deg = config.get_lunar_phase_angle()
    lunar_pos = {'distance_moon_sun_au': 1.0,
                 'sat_to_moon_distance_km': EARTH_TO_MOON_DISTANCE_KM,
                 'phase_angle_deg': phase_angle_deg,
                 'selen_latitude_deg': 0.0,
                 'selen_longitude_deg': 0.0,
                 'selen_solar_longitude_deg': 0.0}
    return lunar_pos


def create_hr_lunar_image(phase_angle_deg):
    image_dir = os.path.dirname(os.path.abspath(__file__))
    im = image.imread(os.path.join(image_dir, 'fullmoon.png'))
    im = 0.2989 * im[:, :, 0] + 0.5870 * im[:, :, 1] + 0.1140 * im[:, :, 2]
    r, p_horz, p_vert, alg_points = center_lunar_image(im)

    # Find the angles of each pixel
    y_pixel = np.arange(0, len(im[0])) - p_horz
    z_pixel = np.arange(0, len(im)) - p_vert
    y_pixel, z_pixel = np.meshgrid(y_pixel, z_pixel)
    x_pixel = np.sqrt(r ** 2 - y_pixel ** 2 - z_pixel ** 2)
    phi = np.rad2deg(np.arctan(y_pixel / x_pixel))
    theta = np.rad2deg(np.arctan(np.sqrt(x_pixel ** 2 + y_pixel ** 2) / z_pixel))
    if phase_angle_deg > 0:
        shadow_start_angle = 90 - phase_angle_deg
        im[phi >= shadow_start_angle] = 0.0
    else:
        shadow_start_angle = np.abs(phase_angle_deg) - 90
        im[phi <= shadow_start_angle] = 0.0

    return im


def compute_altius_irradiance(altius_image, lunar_geometry_altius: LunarGeometry):
    delta_omega_altius = lunar_geometry_altius.delta_omega_view()
    check_irradiance = np.nansum(delta_omega_altius * altius_image)
    return check_irradiance


# config_file = None
# satellite_ephemeris_file = 'satellite_ephemeris.csv'
#
# channel = 'vis'
# mode = 'native'

# scale the moon to be the correct irradiance - convolve rolo with +srf
# convolve it down to the detector resolution for the correct place on the detector
# convolve with +psf
# save to netcdf
@click.command()
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config',
              default=None)
@click.option('-e', '--satellite_ephemeris_file', type=str,
              help='Path to the satellite ephemeris file, with columns [x, y, z, mjd] and the same number of rows'
                   'as the number of calibration wavelengths times the number of lunar grid positions',
              default=None)
@click.option('-o', '--output_file', type=str,
              help='Output file name for the lunar observations to be placed in the ALTIUS_L1PP_DATABASE '
                   'lunar observation directory')
@click.option('-c', '--channel', type=str, help='uv, vis, or nir')
def cli(output_file: str, channel, config_file: str = None, satellite_ephemeris_file: str = None):
    config = altius_config.AltiusL1ppInflightConfig(config_file)
    run_make_lunar_scenes(config, satellite_ephemeris_file, channel, output_file)


def run_make_lunar_scenes(config: altius_config.AltiusL1ppInflightConfig,
                          satellite_ephemeris_file: str = None,
                          channel: str = 'vis',
                          output_file: str = 'lunar_out.nc'):
    # sat_ephem = np.genfromtxt(satellite_ephemeris_file, delimiter=',')  # TODO eventually use
    fov = config.get_detector_fov(channel)
    detector_horz_deg = fov[1]
    detector_vert_deg = fov[0]
    mode = config.get_prnu_detector_mode()
    dim = config.get_detector_pixel_dimensions(channel, mode)
    num_pixels_altius_horz = dim[1]
    num_pixels_altius_vert = dim[0]
    rolo_wavelengths = config.get_rolo_wavelengths(channel=channel)
    # angular width of moon:
    lunar_angular_radius_deg = np.rad2deg(np.arctan(LUNAR_RADIUS / EARTH_TO_MOON_DISTANCE_KM))
    im = create_hr_lunar_image(7.0)
    r, p_horz, p_vert, alg_points = center_lunar_image(im)
    # approximate:
    num_virtual_pixels_horz = int(r * detector_horz_deg / lunar_angular_radius_deg)
    num_virtual_pixels_vert = int(r * detector_vert_deg / lunar_angular_radius_deg)
    # place the moon in various positions:
    num_pos_vert, num_pos_horz = config.get_lunar_prnu_num_images()
    rolo = RoloModel()
    bv, bh = config.get_pixel_boresights(channel=channel, mode=mode)
    psf = PsfSimpleGaussian(2.1)  # TODO pull in the proper PSF
    vert_scale = num_pixels_altius_vert / num_virtual_pixels_vert
    horz_scale = num_pixels_altius_horz / num_virtual_pixels_horz
    scene_interface = SpsToUsaskLunarPrnuSceneInterface()

    dir_out = get_lunar_observations_directory(config, channel)
    if not os.path.exists(dir_out):
        os.makedirs(dir_out)
    file_out = os.path.join(dir_out, output_file)
    bv_virt = bv / vert_scale
    bh_virt = bh / horz_scale
    grid_pixel_positions = []
    for row in range(0, num_pos_vert):
        for col in range(0, num_pos_horz):
            grid_pixel_positions.append((int((row + 0.5) * num_pixels_altius_vert / num_pos_vert),
                                         int((col + 0.5) * num_pixels_altius_horz / num_pos_horz)))
    scene_interface.open_file_for_write(file_out,
                                        rolo_wavelengths,
                                        detector_fov=(detector_vert_deg, detector_horz_deg),
                                        detector_size_pixels=(num_pixels_altius_vert, num_pixels_altius_horz),
                                        detector_boresight_pixels=(bv, bh),
                                        grid_pixel_positions=np.asarray(grid_pixel_positions))
    im_width = len(im[0])
    im_height = len(im)
    lunar_geometry = LunarGeometry(num_virtual_pixels_horz, num_virtual_pixels_vert,
                                   vertical_pixel_boresight=bv_virt, horizontal_pixel_boresight=bh_virt,
                                   vert_fov_degrees=detector_vert_deg, horz_fov_degrees=detector_horz_deg)
    lunar_geometry_altius = LunarGeometry(num_pixels_altius_horz, num_pixels_altius_vert, bv, bh,
                                          vert_fov_degrees=detector_vert_deg,
                                          horz_fov_degrees=detector_horz_deg)

    click.secho('WARNING: PSF blurring not included in current version', fg='yellow')
    click.secho('WARNING: SRF assumed Gaussian - real data not incorporated yet', fg='yellow')
    click.secho('WARNING: Satellite ephemeris ignored in current version', fg='yellow')
    # Order is: at each position, image all calibration wavelengths, then move the satellite
    for pos in grid_pixel_positions:

        for calibration_wavelength in rolo_wavelengths:
            ch = pos[1] / horz_scale
            cv = pos[0] / vert_scale

            tl_corner_h = int(ch - p_horz)
            tl_corner_v = int(cv - p_vert)
            measurement_time = np.datetime64('2019-11-14T12:35')  # TODO this will be read in from file
            satellite_position = np.array([0, 0, 0])  # TODO this will be read in from file
            lunar_pos = get_lunar_positioning_geometry(measurement_time, satellite_position, config)

            im = create_hr_lunar_image(lunar_pos['phase_angle_deg'])

            lunar_geometry.set_disk(r, ch, cv, lunar_pos['sat_to_moon_distance_km'])
            delta_omega = lunar_geometry.delta_omega_view()

            # Compute the irradiance for this geometry
            grid_wl = np.arange(calibration_wavelength - 20, calibration_wavelength + 20, step=0.1)
            srf = GaussianSrf(grid_wl, 3.0 * np.ones_like(grid_wl))
            observed_irrad = compute_lunar_irradiance(calibration_wavelength, lunar_pos, rolo, srf)

            scale = observed_irrad / np.nansum(
                im * delta_omega[tl_corner_v:tl_corner_v + im_height, tl_corner_h:tl_corner_h + im_width])
            im *= scale

            # convolve down the image:
            altius_res_moon = rescale(im, (vert_scale, horz_scale))
            altius_image = np.zeros((num_pixels_altius_vert, num_pixels_altius_horz))

            tl_h = int(tl_corner_h * horz_scale)
            tl_v = int(tl_corner_v * vert_scale)

            altius_lunar_center_horz = p_horz * horz_scale + tl_h
            altius_lunar_center_vert = p_vert * vert_scale + tl_v

            # place the downgraded moon in the image:
            altius_image[tl_v:len(altius_res_moon) + tl_v, tl_h:len(altius_res_moon[0]) + tl_h] = altius_res_moon
            altius_lunar_radius = r * horz_scale

            # Check the irradiance:
            check_irradiance = compute_altius_irradiance(altius_image, lunar_geometry_altius)
            altius_image *= observed_irrad / check_irradiance
            check_irradiance = compute_altius_irradiance(altius_image, lunar_geometry_altius)
            # altius_image = psf.blur_image(altius_image)  TODO include blurring

            # add noise sources:
            max_radiance = config.get_max_lunar_radiance()
            radiance_to_DN_factor = config.get_max_DN() / max_radiance
            altius_image = convert_to_DN_and_add_noise(altius_image, delta_omega, radiance_to_DN_factor)

            scene_interface.write_next_lunar_scene_to_netcdf(time=measurement_time,
                                                             scene=altius_image,
                                                             wavelength=calibration_wavelength,
                                                             expected_lunar_center=(altius_lunar_center_vert,
                                                                                    altius_lunar_center_horz),
                                                             expected_lunar_radius=altius_lunar_radius,
                                                             sat_moon_distance_km=lunar_pos['sat_to_moon_distance_km'],
                                                             start_boresight_pointing=np.array([1, 0, 0]),
                                                             end_boresight_pointing=np.array([1, 0, 0]),
                                                             integration_time=1.0,
                                                             subsolar_selen_longitude=lunar_pos[
                                                                 'selen_solar_longitude_deg'],
                                                             subsat_selen_latitude=lunar_pos['selen_latitude_deg'],
                                                             subsat_selen_longitude=lunar_pos['selen_longitude_deg'],
                                                             sun_to_moon_distance_au=lunar_pos[
                                                                 'sat_to_moon_distance_km'],
                                                             lunar_phase_angle_deg=lunar_pos['phase_angle_deg'],
                                                             satellite_position=satellite_position)
    scene_interface.close_write_nc()
    print('Finished writing lunar calibration data to ', file_out)


def compute_lunar_irradiance(calibration_wavelength, lunar_pos, rolo, srf):
    irradiance_spectrum_moon = []
    lunar_wl = np.arange(calibration_wavelength - 30, calibration_wavelength + 30, step=0.1)
    for wl in lunar_wl:
        irrad = rolo.irradiance(wavelength_nm=wl,
                                distance_moon_sun_au=lunar_pos['distance_moon_sun_au'],
                                distance_moon_spacecraft_km=lunar_pos['sat_to_moon_distance_km'],
                                phase_angle_deg=lunar_pos['phase_angle_deg'],
                                selen_latitude_deg=lunar_pos['selen_latitude_deg'],
                                selen_longitude_deg=lunar_pos['selen_longitude_deg'],
                                selen_solar_longitude_deg=lunar_pos['selen_solar_longitude_deg'])
        irradiance_spectrum_moon.append(irrad)
    observed_irrad = srf.convolve_spectrum_with_srf(np.array([calibration_wavelength]),
                                                    lunar_wl, np.array(irradiance_spectrum_moon))
    # done computing irradiance
    return observed_irrad


if __name__ == '__main__':
    fileout = 'lunar_test_measurements.nc'
    config = altius_config.AltiusL1ppInflightConfig()
    run_make_lunar_scenes(config, '', 'uv', fileout)
