import os

import numpy as np
# import spiceypy as spice
from scipy.interpolate import interp1d
from scipy.stats import norm


class RoloModel:
    """
        Implements the lunar irradiance modelled from the RObotic Lunar Observatory (ROLO)
        observations of the moon.  The irradiance is a function of phase angle, solar angle, and
        spacecraft position relative to the selenographic surface origin.  Phase angles close
        to opposition (0.0 deg) are not well modelled by ROLO and should be avoided.

        The coefficients are contained in data/rolo_coefficients.csv
        The solar model used is from Wehrli (1985), contained in data/wehrli1985.csv, (W/m2/nm),
        and is stored in photons/s/cm2/nm
        Lunar soil spectra are also available with this model, though has not yet been
        incorporated into the lunar irradiance model.  Data contained in  data/lunar_soil_62231.csv

        H.H. Kieffer & T.C. Stone 2005.  THE SPECTRAL IRRADIANCE OF THE MOON, The Astronomical Journal (2005 June) 129:2887-2901
    """

    def __init__(self):
        os.chdir(os.path.dirname(os.path.abspath(__file__)))
        data = np.genfromtxt('data/rolo_coefficients.csv', skip_header=1, delimiter=",")
        self.coeffs = data[:, 1:]
        self.wavelen_nm = np.squeeze(data[:, 0])
        self.SRF_FWHM_nm = 2.0
        solar = np.genfromtxt('data/wehrli85.csv', skip_header=1, delimiter=None)
        self.solar_irrad_wehrli_wavelen_nm = np.squeeze(solar[:, 0])

        self._lunar_soil = np.genfromtxt('data/lunar_soil_62231.csv', skip_header=1, delimiter=None)

        # energy per photon:
        h = 6.62607004e-34
        c = 2.99792458e8
        e_per_photon = h * c / self.solar_irrad_wehrli_wavelen_nm / 1e-9
        # TODO convert m^2 to cm^2 (x10000)

        self.solar_irrad_wehrli_irrad = np.squeeze(solar[:, 1]) / e_per_photon

    def lunar_soil(self):
        """
            Returns
            -------
            The lunar soil reflectance spectrum (2D array of wavelength and reflectance)
        """
        return self._lunar_soil

    def disk_reflectance(self, wavelength_index: int, phase_angle_deg: float, selen_latitude_deg: float,
                         selen_longitude_deg: float, selen_solar_longitude_deg: float):
        """
            Computes the lunar disk reflectance (from 0.0 to 1.0) given the wavelength and geometry
            Phase angles close to opposition (0.0) are not well modelled, and should be avoided

            Utilizes the coefficents from Kieffer & Stone (2005) to compute the reflectance.

            Parameters
            ----------

            wavelength_index: int
                Index of the ROLO model wavelength being computed

            phase_angle_deg: float
                Phase angle of the moon in degrees.  0.0 degrees is full moon

            selen_latitude_deg: float
                Sub-satellite selenographic latitude, in degrees.

            selen_longitude_deg: float
                Sub-satellite selenographic longitude, in degrees.

            selen_solar_longitude_deg: float
                Sub-solar selenographic longitude, in degrees

            Returns
            -------
            reflectance : float
                Value in range [0.0, 1.0] for the reflectance of the lunar disk.

        """
        a = self.coeffs[wavelength_index, 0:4]
        b = self.coeffs[wavelength_index, 4:7]
        d = self.coeffs[wavelength_index, 7:10]
        c = self.coeffs[wavelength_index, 10:14]
        p = self.coeffs[wavelength_index, 14:]
        g = np.abs(np.deg2rad(phase_angle_deg))
        phi_solar = np.deg2rad(selen_solar_longitude_deg)
        theta = np.deg2rad(selen_latitude_deg)
        phi = np.deg2rad(selen_longitude_deg)
        i = a[0] + a[1] * g + a[2] * g ** 2 + a[3] * g ** 3
        ii = b[0] * phi_solar + b[1] * phi_solar ** 3 + b[2] * phi_solar ** 5
        iii = c[0] * theta + c[1] * phi + c[2] * phi_solar * theta + c[3] * phi_solar * phi
        iv = d[0] * np.exp(-g / p[0]) + d[1] * np.exp(-g / p[1]) + d[2] * np.cos((g - p[2]) / p[3])

        return np.exp(i + ii + iii + iv)

    def irradiance(self, wavelength_nm: float, distance_moon_spacecraft_km: float,
                   distance_moon_sun_au: float, phase_angle_deg: float,
                   selen_latitude_deg: float, selen_longitude_deg: float,
                   selen_solar_longitude_deg: float):
        """
            Computes the irradiance of the moon for the given geometry and wavelength.
            Utilizes the reflectance model, and essentially scales the reflectance by the
            solar model, which is convolved with the ROLO spectral response function

            The function computes the irradiance at all ROLO wavelengths, and determines the
            average irradiance, in photons/s/cm2/nm at the intended wavelength by multiplying the spectrum
            by the ROLO SRF and taking the sum.

            Parameters
            ----------

            wavelength_nm: float
                Wavelength in nm to be modelled.  For better accuracy, it is recommended to
                use only wavelengths that are directly modelled by ROLO - there are large spans
                between some wavelengths, and interpolation will have a high degree of uncertainty.

            distance_moon_spacecraft_km: float
                Distance from the observing spacecraft to the center of the moon, in km.

            distance_moon_sun_au: float
                Distance from the center of the moon to the center of the sun, in astronomical units.

            phase_angle_deg: float
                Phase angle of the moon in degrees.  0.0 degrees is full moon

            selen_latitude_deg: float
                Sub-satellite selenographic latitude, in degrees.

            selen_longitude_deg: float
                Sub-satellite selenographic longitude, in degrees.

            selen_solar_longitude_deg: float
                Sub-solar selenographic longitude, in degrees

            Returns
            -------
            irrad : float
                The average irradiance in photons/s/cm2/nm.
        """
        omega_m = 6.4177e-5  # moon solid angle at 384400 km TODO might correct this calculation
        fd = (distance_moon_spacecraft_km * distance_moon_sun_au / 384400) ** 2
        wl_grid, srf = self.get_instrument_response(wavelength_nm)
        irrad = np.zeros(np.shape(self.wavelen_nm))
        f_solar = interp1d(self.solar_irrad_wehrli_wavelen_nm, self.solar_irrad_wehrli_irrad,
                           kind='cubic', fill_value=0.0)
        solar_spectrum = f_solar(wl_grid)

        # TODO need to smooth reflectance by lunar soil and breccia spectra (TBD)
        for wlidx, wl in enumerate(self.wavelen_nm):
            refl = self.disk_reflectance(wlidx, phase_angle_deg, selen_latitude_deg,
                                         selen_longitude_deg, selen_solar_longitude_deg)

            ek = np.sum(solar_spectrum * srf)
            irrad[wlidx] = refl * omega_m * ek / np.pi / fd

        f_irrad = interp1d(self.wavelen_nm, irrad, kind='linear', fill_value=0.0, bounds_error=False)
        irrad_avg = np.nansum(f_irrad(wl_grid) * srf)

        return irrad_avg

    def get_instrument_response(self, wavelength_nm: float):
        """Returns a Gaussian SRF based on self.SRF_FWHM_nm

        Parameters
        ----------
        wavelength_nm: float
            Wavelength, in nm, to get the SRF

        Returns
        -------
        wl_grid: np.ndarray
            Array of wavelengths in nm, from 200 to 1300 in 0.1 nm steps

        srf: np.ndarray
            Gaussian centered at wavelength_nm, same length as wl_grid
        """
        wl_grid = np.arange(200.0, 1300.0, step=0.1)  # 0.1 nm resolution
        sigma = self.SRF_FWHM_nm / 2.355

        z = (wl_grid - wavelength_nm) / sigma
        srf = norm.pdf(z, 0, 1)
        srf /= np.sum(srf)
        return wl_grid, srf


class SpiceAngleCalc:

    # TODO port this over to astropy
    def __init__(self):
        pass
        # spice.furnsh("data/latest_leapseconds.tls")
        # spice.furnsh("data/moon_pa_de421_1900-2050.bpc")
        # spice.furnsh("data/moon_080317.tf")
        # spice.furnsh("data/moon_assoc_pa.tf")
        # spice.furnsh("data/pck00010.tpc")
        # spice.furnsh("data/earth_latest_high_prec.bpc")
        # spice.furnsh("data/naif0012.tls")
        # spice.furnsh("data/de421.bsp")
        # spice.furnsh("data/earth_assoc_itrf93.tf")
        # This way we're not calling furnsh every time we make a geometry calculation

    # def compute_lunar_angles(self, observation_times_tai_seconds: np.ndarray, satellite_positions: np.ndarray,
    #                          satellite_ref_frame="J2000"):
    #
    #     et = observation_times_tai_seconds.reshape((len(observation_times_tai_seconds), 1)) + 32.184
    #     n_obs = len(et)
    #     satellite_positions = satellite_positions.reshape((n_obs, 3))
    #
    #     # Computing in high-res ECEF coordinates:
    #     sun_to_moon_km, a = spice.spkpos("MOON", et, satellite_ref_frame, "NONE", "SUN")
    #     sun_to_earth_km, a = spice.spkpos("EARTH", et, satellite_ref_frame, "NONE", "SUN")
    #
    #     earth_to_moon_km = np.zeros((n_obs, 3))
    #     lunar_velocity_km_s = np.zeros((n_obs, 3))
    #     for i in range(0, n_obs):
    #         sv, a = spice.spkgeo(301, et[i], satellite_ref_frame, 399)
    #         earth_to_moon_km[i] = sv[0:3]
    #         lunar_velocity_km_s[i] = sv[3:6]
    #
    #     # assume ecef_satellite_positions is in km
    #     observer_to_moon_km = earth_to_moon_km - satellite_positions
    #
    #     # Phase angle (this might be standardized to the center of earth rather than the observer)
    #     g = np.arccos(np.sum(sun_to_moon_km * observer_to_moon_km, axis=1) /
    #                   (np.linalg.norm(sun_to_moon_km, axis=1) * np.linalg.norm(observer_to_moon_km, axis=1)))
    #
    #     seleno_latitudes, seleno_longitudes = self.compute_sub_observer_point_moon(
    #         -earth_to_moon_km, -observer_to_moon_km, satellite_ref_frame, et)
    #
    #     solar_latitudes, solar_longitudes = self.compute_sub_observer_point_moon(
    #         -earth_to_moon_km, -sun_to_moon_km, satellite_ref_frame, et)
    #     au = 1.495978707e8  # km
    #     moon_sun_distance_au = np.linalg.norm(sun_to_moon_km, axis=1) / au
    #
    #     return g, solar_longitudes, seleno_latitudes, seleno_longitudes, moon_sun_distance_au,
    #     sun_to_moon_km, sun_to_earth_km, earth_to_moon_km
    #
    # @staticmethod
    # def compute_sub_observer_point_moon(moon_to_earth: np.ndarray,
    #                                     moon_to_observer: np.ndarray,
    #                                     reference_system: str,
    #                                     ephemeris_times: np.ndarray):
    #     """computes lat, lon of sub-observer point on moon assuming it is a sphere"""
    #     n_obs, n_dim = np.shape(moon_to_earth)
    #
    #     rm = np.zeros((n_obs, 3, 3))
    #     for i in range(0, n_obs):
    #         rm[i] = spice.pxform(reference_system, "J2000", ephemeris_times[i])
    #
    #     moon_to_observer_j2000 = np.array([np.dot(m, v.T) for m, v in zip(rm, moon_to_observer)])
    #
    #     look_vector = moon_to_observer_j2000 / np.linalg.norm(moon_to_observer_j2000, axis=1).reshape((n_obs, 1))
    #
    #     rrr = np.array([spice.pxform(reference_system, "MOON_ME", t) for t in ephemeris_times])
    #     moon_to_observer_me = np.array([np.dot(m, v.T) for m, v in zip(rrr, look_vector)])
    #
    #     lon = np.zeros((n_obs, 1))
    #     lat = np.zeros((n_obs, 1))
    #     for k, v in enumerate(moon_to_observer_me):
    #         r, lon[k], lat[k] = spice.reclat(v)
    #
    #     return lat, lon

    # https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d/897677#897677
    @staticmethod
    def compute_rotation_matrix(from_vectors: np.ndarray,
                                to_vectors: np.ndarray):
        n_obs, n_dim = np.shape(to_vectors)

        a = from_vectors / np.linalg.norm(from_vectors, axis=1).reshape((n_obs, 1))
        b = to_vectors / np.linalg.norm(to_vectors, axis=1).reshape((n_obs, 1))
        dot_a_b = np.sum(a * b, axis=1).reshape((n_obs,))
        cross_a_b = np.cross(a, b)

        rotmat = np.zeros((n_obs, 3, 3))

        for i in range(0, n_obs):
            d = dot_a_b[i]
            x = cross_a_b[i]

            gg = np.array([[d, -np.linalg.norm(x), 0],
                           [np.linalg.norm(x), d, 0],
                           [0, 0, 1]])

            m = b[i] - d * a[i]
            ff = np.squeeze(np.array([[a[i]],
                                      [m / np.linalg.norm(m)],
                                      [-x]])).T
            uu = ff @ gg @ np.linalg.inv(ff)
            rotmat[i] = uu

        return rotmat

    @staticmethod
    def determine_me_coordinates(moon_to_earth_j2000: np.ndarray,
                                 lunar_velocity_j2000: np.ndarray,
                                 moon_to_earth_me: np.ndarray):
        n_obs, n_dim = np.shape(moon_to_earth_j2000)
        x_axis = moon_to_earth_j2000
        x_axis = x_axis / np.linalg.norm(x_axis, axis=1).reshape((n_obs, 1))

        y_axis = -lunar_velocity_j2000 / np.linalg.norm(lunar_velocity_j2000, axis=1).reshape((n_obs, 1))
        z_axis = np.cross(x_axis, y_axis)

        # TODO compute rotation matrix between moon_to_earth_J2000 and moon_to_earth_me
        # from https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d:
        a = moon_to_earth_j2000 / np.linalg.norm(moon_to_earth_j2000, axis=1).reshape((n_obs, 1))
        b = moon_to_earth_me / np.linalg.norm(moon_to_earth_me, axis=1).reshape((n_obs, 1))
        cc = np.cos(np.sum(a * b, axis=1))
        vv = np.cross(a, b)

        rr = np.zeros((n_obs, 3, 3))
        for k in range(0, n_obs):
            c = cc[k]
            v = vv[k]
            vx = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
            r = np.identity(3) + vx + np.dot(vx, vx.T) / (1 + c)
            rr[k] = r

        return x_axis, y_axis, z_axis, rr
