import click
import numpy as np
import os
from pathlib import Path

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry, center_solar_disk
from altius_l1pp_inflight.scene.interfaces.solar_srf import SpsToUsaskSolarSrfSceneInterface, \
    SpsSingleSolarImageInterface
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise


@click.command()
@click.option('-i', '--input_directory', type=str,
              help='Full directory containing all image outputs defined by SpsSingleSolarImageInterface class')
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-o', '--output_file', type=str,
              help='Unique output file name where solar scenes are saved in the ALTIUS_L1PP_DATABASE directory')
def cli(output_file: str, config_file: str = None, input_directory: str = None):
    convert_altius_images_to_srf_input_format(input_directory, config_file, output_file)


def convert_altius_images_to_srf_input_format(input_directory: str, config_file: str, output_file: str):
    config = altius_config.AltiusL1ppInflightConfig(config_file)

    out_dir = config.get_processed_solar_srf_file_directory()
    output_filepath = os.path.join(out_dir, output_file)

    channel = config.get_srf_calibration_channel()
    mode = config.get_srf_detector_mode()
    calibration_wavelengths = config.get_srf_calibration_wavelengths(channel)
    detector_positions = config.get_solar_srf_detector_positions()
    pixel_size = config.get_detector_pixel_dimensions(channel, mode)
    boresight = config.get_pixel_boresights(channel, mode)
    fov = config.get_detector_fov(channel)
    num_meas_per_band = config.get_num_measurements_per_band()

    interface = SpsToUsaskSolarSrfSceneInterface()
    interface.open_file_for_write(output_filepath,
                                  calibration_wavelengths,
                                  detector_positions,
                                  pixel_size,
                                  boresight,
                                  fov,
                                  num_meas_per_band)

    solar_geometry = SolarGeometry(pixel_size[1], pixel_size[0],
                                   boresight[0], boresight[1],
                                   fov[0], fov[1])

    center_radius_pixels = config.get_srf_center_pixel_radius()

    data_path = Path(input_directory)

    results = []
    for _ in detector_positions:
        results.append([])

    n_files = 0

    for f in data_path.iterdir():
        if f.is_file():
            n_files += 1
            image_interface = SpsSingleSolarImageInterface.from_netcdf(f.as_posix())
            raw_data = image_interface.raw_data()
            kwargs = {'detector_angular_width_degrees': fov,
                      'distance_to_sun_km': raw_data['distance_to_sun']}
            image = raw_data['image']
            solar_radius, solar_center_horz, solar_center_vert, ag = center_solar_disk(image, **kwargs)

            center_pixels = compute_center_pixel_irradiance(center_radius_pixels, image, raw_data['distance_to_sun'],
                                                            solar_center_horz,
                                                            solar_center_vert, solar_geometry,
                                                            solar_radius)

            result = {'time': raw_data['time'],
                      'center_pixel_sum': center_pixels,
                      'calibration_wavelength': raw_data['calibration_wavelength'],
                      'measurement_wavelength': raw_data['measurement_wavelength'],
                      'solar_center': (solar_center_vert, solar_center_horz),
                      'solar_radius': solar_radius,
                      'satellite_to_sun_distance_km': raw_data['distance_to_sun'],
                      'start_pointing': raw_data['start_pointing'],
                      'end_pointing': raw_data['end_pointing'],
                      'integration_time': raw_data['integration_time'],
                      'srf_central_lobe_sum': 1.0}
            idx = np.argmin(np.sum((raw_data['detector_position'].values - detector_positions) ** 2))
            results[idx].append(result)

    if n_files != num_meas_per_band * len(calibration_wavelengths) * len(detector_positions):
        raise AssertionError('Missing necessary files for full calibration')

    # sort the results by calibration wavelength and detector position
    for pos_idx, pos in enumerate(detector_positions):
        for cal_wl, meas_wl in zip(calibration_wavelengths, config.get_srf_measurement_wavelengths(channel)):
            interface.write_current_measurement_wavelengths(pos, cal_wl, meas_wl)
            for v in results[pos_idx]:
                interface.write_next_solar_scene_to_netcdf(v['time'],
                                                           v['center_pixel_sum'],
                                                           v['measurement_wavelength'],
                                                           v['solar_center'],
                                                           v['solar_radius'],
                                                           v['satellite_to_sun_distance_km'],
                                                           v['start_pointing'],
                                                           v['end_pointing'],
                                                           v['integration_time'],
                                                           v['srf_central_lobe_sum'])

    interface.close_write()


def compute_center_pixel_irradiance(center_radius_pixels, image, distance_to_sun,
                                    solar_center_horz, solar_center_vert,
                                    solar_geometry, solar_radius):
    solar_geometry.set_disk(disk_pixel_radius=solar_radius,
                            disk_horizontal_center=solar_center_horz,
                            disk_vertical_center=solar_center_vert,
                            distance_to_body_km=distance_to_sun)
    mask = solar_geometry.center_pixel_mask(center_radius_pixels)
    center_pixels = np.nansum(image[mask])
    return center_pixels
