"""
Command line script for writing the .nc files used for the PRNU calibration scenario.
It computes the high-resolution solar disk at each time step of the calibration
and saves them to a file that would potentially be used by QinetiQ SPS as a
stimuli input.

The script then takes the high-res images and converts them to
ALTIUS detector scenes of the sun, with the sun at each position on the detector
to be used to form a PRNU image by the processing script.


Methods
-------
compute_altius_image_and_save : None
    Simulates a single detector image of the sun and saves it to file.  Images of the sun are created at
    high resolution to closely simulate actual solar images from the HMI database.

    The high resolution image of the sun is interpolated at a given time step, and convolves the SRF with the
    SAO2010 solar spectrum to scale
    the irradiance of the solar disk to the correct value.  The image is then convolved down to ALTIUS detector
    resolution and placed at the correct position on the detector, as well as applying the spatial PSF
    to correctly blur the image.  Lastly, shot noise and digitization noise is added to the image for the sun
    at (near) full well DN.

    See Also
    --------
    :func:`~.compute_altius_image_and_save`

run_make_solar_images : None
    Function to compute the detector images for each position of the sun. The effect is
    to simulate the sun at every point on the detector to get coverage on every pixel.

Notes
-----
Returns two netcdf files - one containing the high-resolution stimuli defined by the interface in
:code:`SpsHighResolutionUsaskToSpsSceneInterface`, and one containing the ALTIUS detector images
defined by the interface in :code:`SpsToUsaskSolarSceneInterface`
"""

import os

import click
import numpy as np
from sasktran import SolarSpectrum

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.instrument.psf.psf_interface import PsfInterface, PsfSimpleGaussian
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf
from altius_l1pp_inflight.instrument.srf.srf_interface import SrfInterface
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry
from altius_l1pp_inflight.scene.generation.solar.calibration_util import get_solar_prnu_simulated_observation_dir, \
    compute_solar_angular_radius
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDataScenario
from altius_l1pp_inflight.scene.generation.solar.solar_scan import SolarScan
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import RasterScanGrid
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import SolarImage
from altius_l1pp_inflight.scene.interfaces.solar_high_res import SpsHighResolutionUsaskToSpsSceneInterface
from altius_l1pp_inflight.scene.interfaces.solar_prnu import SpsToUsaskSolarPrnuSceneInterface
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise


def compute_altius_image_and_save(solar_scan, raster_idx, scenario,
                                  horizontal_width, vertical_width,
                                  srf: SrfInterface,
                                  psf: PsfInterface,
                                  wavelength, measurement_times,
                                  sps_to_usask_interface: SpsToUsaskSolarPrnuSceneInterface,
                                  usask_to_sps_interface: SpsHighResolutionUsaskToSpsSceneInterface,
                                  irradiance_to_dn_factor: float,
                                  delta_omega: np.ndarray):
    """
    Compute the solar scene at the current raster scan index.  The call to make_altius_observation is the slowest
    portion as it involves sub-pixel matching of the high-resolution HMI images to the lower-resolution ALTIUS +scene.
    That is why once this computation is performed for this scenario, it is saved to file.

    Parameters
    ----------
    solar_scan : SolarScan
        SolarScan class containing all necessary information to generate a solar image for a
        given viewing position on the detector and sunspot scenario

    horizontal_width : int
        Width in pixels of the detector

    vertical_width : int
        Height in pixels of the detector

    raster_idx : int
        Given the raster-scanning scheme for the PRNU, this indexes the position of the sun on the detector
        according to a back-and-forth raster scan, with 0 at the bottom left.

    scenario : SolarDataScenario
        Scenario describing the sunspots from the HMI repository

    srf : SrfInterface
        The SRF of the instrument used for generating images.

    psf : PsfInterface
        The PSF of the instrument, used to blur the final image on the detector.

    wavelength : float
        Wavelength of the measurement, in nm

    measurement_times : np.ndarray[np.datetime64]
        Array of times of all previous simulated solar disks, only used outside this function

    sps_to_usask_interface : SpsToUsaskSolarPrnuSceneInterface
        Interface that will be receiving files from the SPS for the solar PRNU.  For consistency,
        here the interface writes the results to file in the same format that will be provided by the SPS

    usask_to_sps_interface : SpsHighResolutionUsaskToSpsSceneInterface
        Interface to write high-resolution stimuli files

    irradiance_to_dn_factor : float

    Returns
    -------
    None : None
        Writes the computed image to the currently open netcdf files for the high resolution
        and detector
    """
    solar_image: SolarImage = solar_scan.get_indexed_image(scenario, raster_idx)
    altius_image, image_geometry = solar_image.make_altius_observation(num_pixels_horizontal=horizontal_width,
                                                                       num_pixels_vertical=vertical_width,
                                                                       srf=srf,
                                                                       wavelength=wavelength)

    # scale the limb darkening, and write to high res netcdf file for SPS stimuli input:
    solar_disk_radiance = solar_image.get_solar_disk_radiance_result()
    horz_virtual_shift, vert_virtual_shift = \
        solar_image._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(raster_idx)
    limb_corrected_radiance = solar_disk_radiance.to_scaled_image_result(wavelength, srf, horz_virtual_shift,
                                                                         vert_virtual_shift)
    time: np.datetime64 = solar_image._dataset['time'].values
    icf_solar_center = solar_image.compute_icf_solar_center().flatten()
    usask_to_sps_interface.write_next_solar_disk_to_netcdf(time,
                                                           limb_corrected_radiance.radiance(),
                                                           wavelength,
                                                           limb_corrected_radiance.center_pixel(),
                                                           limb_corrected_radiance.average_solar_radius_in_pixels(),
                                                           icf_solar_center)

    measurement_times.append(time)
    satellite_to_sun_distance_km = solar_image._dataset['satellite_to_sun_distance_km'].values
    expected_altius_solar_center_horz = image_geometry['altius_solar_center_horz']
    expected_altius_solar_center_vert = image_geometry['altius_solar_center_vert']
    expected_altius_solar_radius = image_geometry['altius_solar_radius']

    altius_image_dn = convert_to_DN_and_add_noise(altius_image, delta_omega, irradiance_to_dn_factor)

    sps_to_usask_interface.write_next_solar_scene_to_netcdf(time,
                                                            altius_image_dn,
                                                            wavelength,
                                                            (expected_altius_solar_center_vert,
                                                             expected_altius_solar_center_horz),
                                                            expected_altius_solar_radius,
                                                            satellite_to_sun_distance_km,
                                                            np.array([1, 0, 0]),
                                                            np.array([1, 0, 0]),
                                                            0.1)


@click.command()
@click.option('-e', '--satellite_ephemeris_file', type=str,
              help='Path to the satellite ephemeris file, with columns [x, y, z, mjd] and the same number of rows'
                   'as the number of calibration wavelengths times the number of lunar grid positions',
              default=None)
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-o', '--output_file', type=str,
              help='Output filename to be stored in the ALTIUS_L1PP_DATABASE directory for processed solar scene files')
def cli(output_file: str, satellite_ephemeris_file: str = None, config_file: str = None):
    config = altius_config.AltiusL1ppInflightConfig(config_file)
    run_make_solar_scenes_for_prnu(output_file, config, satellite_ephemeris_file)


def run_make_solar_scenes_for_prnu(output_file: str, config: altius_config.AltiusL1ppInflightConfig,
                                   satellite_ephemeris_file: str = None):
    """
        Parameters
        ----------
        output_file : str

        satellite_ephemeris_file : str

        config : AltiusL1ppInflightConfig
    """
    calibration_wavelength = config.get_solar_prnu_calibration_wavelength()
    mode = config.get_prnu_detector_mode()
    channel = config.get_channel(calibration_wavelength)
    pixel_size = config.get_detector_pixel_dimensions(channel=channel, mode=mode)
    horizontal_width = pixel_size[1]
    vertical_width = pixel_size[0]
    num_vert_scans = config.get_num_vertical_raster_scans()
    num_horz_scans = config.get_num_horizontal_raster_scans()
    prnu_cal_image = np.zeros((vertical_width, horizontal_width))
    fit_cal_image = np.zeros_like(prnu_cal_image)
    bv, bh = config.get_pixel_boresights(channel, mode)
    bh_offset = 2 * bh / horizontal_width - 1.0
    bv_offset = 2 * bv / vertical_width - 1.0
    fov = config.get_detector_fov(channel=channel)  # (vertical, horizontal)

    solar_geometry = SolarGeometry(horizontal_width, vertical_width,
                                   bv, bh, fov[0], fov[1])

    maxDN = config.get_max_DN()
    max_radiance = config.get_max_photons_per_sec_per_nm_per_cm2()
    factor = maxDN / max_radiance

    srf_wavelen = np.linspace(calibration_wavelength - 25, calibration_wavelength + 25, 5001)  # 0.01 nm res

    srf = GaussianSrf(srf_wavelen, np.ones_like(srf_wavelen) * 3.0)
    psf = PsfSimpleGaussian(2.1)  # sort of arbitrary blurring
    click.secho('WARNING: PSF blurring not included in current version', fg='yellow')
    click.secho('WARNING: SRF assumed Gaussian - real data not incorporated yet', fg='yellow')
    click.secho('WARNING: Satellite ephemeris ignored in current version', fg='yellow')

    satellite_height = config.get_nominal_satellite_height_km()

    # Configure the entire solar scan and its geometry:
    scenario = SolarDataScenario(config.get_solar_scenario())
    output_directory = get_solar_prnu_simulated_observation_dir(config, scenario.to_string_directory())
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    sps_output_calibration_file = os.path.join(output_directory, output_file)

    num_scans = num_vert_scans * num_horz_scans

    solar_scan = SolarScan(fov[1], fov[0], bh_offset, bv_offset, satellite_height, config)

    solar_angular_radius = compute_solar_angular_radius(scenario.get_scenario_mjd())

    solar_scan_grid = RasterScanGrid(solar_angular_radius, 0.1, 0.1, fov[1], fov[0],
                                     num_vert_scans, num_horz_scans, 0)
    solar_scan.configure_satellite_orbit_for_scan(solar_scan_grid, scenario)

    measurement_times = []

    # centering_error = np.zeros((num_horz_scans * num_vert_scans, 3))
    sps_to_usask_interface = SpsToUsaskSolarPrnuSceneInterface()
    usask_to_sps_interface = SpsHighResolutionUsaskToSpsSceneInterface()
    sao = SolarSpectrum()

    sps_to_usask_interface.open_file_for_write(sps_output_calibration_file,
                                               calibration_wavelength,
                                               (fov[0], fov[1]),
                                               (vertical_width, horizontal_width),
                                               (vertical_width / 2, horizontal_width / 2))
    solar_spectrum_wavelengths = srf.get_wavelength_grid(0.01)
    raw_stimuli_output = os.path.join(output_directory, 'prnu_calibration_raw_stimuli_{}'.format(output_file))
    usask_to_sps_interface.open_file_for_write(raw_stimuli_output, (fov[0], fov[1]), solar_spectrum_wavelengths,
                                               sao.irradiance(solar_spectrum_wavelengths))

    click.echo('Initiating simulation for {} images of the sun'.format(num_scans))
    for raster_idx in range(0, num_scans):
        click.echo('Computing raster scan {} out of {}'.format(raster_idx + 1, num_scans))
        compute_altius_image_and_save(solar_scan, raster_idx, scenario,
                                      horizontal_width, vertical_width,
                                      srf, psf,
                                      calibration_wavelength, measurement_times,
                                      sps_to_usask_interface,
                                      usask_to_sps_interface,
                                      factor,
                                      solar_geometry.delta_omega_view())

    sps_to_usask_interface.close_write_nc()
    usask_to_sps_interface.close()

    click.secho('*****************************************************************************', fg='green')
    click.secho('Finished writing solar PRNU observation data to {}'.format(sps_output_calibration_file), fg='green')
    click.secho('Raw stimuli solar files written to: {}'.format(raw_stimuli_output), fg='green')
    click.secho('*****************************************************************************', fg='green')


# Temporary place for CLI
if __name__ == '__main__':
    config = altius_config.AltiusL1ppInflightConfig()
    run_make_solar_scenes_for_prnu('prnu_solar_measurements.nc', config, None)
