"""
Command line script for writing the .nc files used for the PRNU calibration scenario.
It computes the high-resolution solar disk at each time step of the calibration
and saves them to a file that would potentially be used by QinetiQ SPS as a
stimuli input.

The script then takes the high-res images and converts them to
ALTIUS detector scenes of the sun, with the sun at each position on the detector
to be used to form a PRNU image by the processing script.


Methods
-------
compute_altius_image_and_save : None
    Simulates a single detector image of the sun and saves it to file.  Images of the sun are created at
    high resolution to closely simulate actual solar images from the HMI database.

    The high resolution image of the sun is interpolated at a given time step, and convolves the SRF with the
    SAO2010 solar spectrum to scale
    the irradiance of the solar disk to the correct value.  The image is then convolved down to ALTIUS detector
    resolution and placed at the correct position on the detector, as well as applying the spatial PSF
    to correctly blur the image.  Lastly, shot noise and digitization noise is added to the image for the sun
    at (near) full well DN.

    See Also
    --------
    :func:`~.compute_altius_image_and_save`

run_make_solar_images : None
    Function to compute the detector images for each position of the sun. The effect is
    to simulate the sun at every point on the detector to get coverage on every pixel.

Notes
-----
Returns two netcdf files - one containing the high-resolution stimuli defined by the interface in
:code:`SpsHighResolutionUsaskToSpsSceneInterface`, and one containing the ALTIUS detector images
defined by the interface in :code:`SpsToUsaskSolarSceneInterface`
"""
import glob
import os
import shutil

import click
import numpy as np
from sasktran.solarspectrum import SolarSpectrum
from sktime.mjd import mjd_to_datetime64

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.instrument.psf.psf_interface import PsfInterface, PsfSimpleGaussian
from altius_l1pp_inflight.instrument.srf.aotf_srf import AltiusAotfSrf
from altius_l1pp_inflight.instrument.srf.srf_interface import SrfInterface
from altius_l1pp_inflight.instrument.srf.uv_srf import AltiusUVSrf
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry
from altius_l1pp_inflight.scene.generation.solar import calibration_util
import altius_l1pp_inflight.scene.generation.solar.calibration_util
from altius_l1pp_inflight.scene.generation.solar.extract_center_solar_pixels import compute_center_pixel_irradiance
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDataScenario
from altius_l1pp_inflight.scene.generation.solar.solar_scan import SolarScan
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import SrfScanGrid
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import SolarImage
from altius_l1pp_inflight.scene.interfaces.solar_srf import SpsToUsaskSolarSrfSceneInterface, \
    SpsSingleSolarImageInterface
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise


def compute_altius_image_and_save(filepath, solar_scan, scene_index, scenario,
                                  horizontal_width, vertical_width,
                                  srf: SrfInterface,
                                  psf: PsfInterface,
                                  wavelength, calibration_wavelength,
                                  radiance_to_dn_factor: float,
                                  solar_geometry: SolarGeometry):
    """
    Compute the solar scene at the current raster scan index.  The call to make_altius_observation is the slowest
    portion as it involves sub-pixel matching of the high-resolution HMI images to the lower-resolution ALTIUS +scene.
    That is why once this computation is performed for this scenario, it is saved to file.

    Parameters
    ----------
    filepath: str

    solar_scan : SolarScan

    horizontal_width : int

    vertical_width : int

    scene_index : int

    scenario : SolarDataScenario

    srf : SrfInterface

    psf : PsfInterface

    wavelength : float

    radiance_to_dn_factor : float

    solar_geometry: SolarGeometry

    calibration_wavelength: float

    Returns
    -------
    None : None
        Writes the computed image to the currently open netcdf files for the high resolution
        and detector

    """

    solar_image: SolarImage = solar_scan.get_indexed_image(scenario, scene_index)
    altius_image, image_geometry = solar_image.make_altius_observation(num_pixels_horizontal=horizontal_width,
                                                                       num_pixels_vertical=vertical_width,
                                                                       srf=srf,
                                                                       wavelength=wavelength,
                                                                       fast_scaling=False)

    # scale the limb darkening, and write to high res netcdf file for SPS stimuli input:
    solar_disk_radiance = solar_image.get_solar_disk_radiance_result()
    horz_virtual_shift, vert_virtual_shift = \
        solar_image._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(scene_index)
    limb_corrected_radiance = solar_disk_radiance.to_scaled_image_result(wavelength, srf, horz_virtual_shift,
                                                                         vert_virtual_shift)
    time: np.datetime64 = solar_image._dataset['time'].values
    # icf_solar_center = solar_image.compute_icf_solar_center().flatten()
    # usask_to_sps_interface.write_next_solar_disk_to_netcdf(time,
    #                                                        limb_corrected_radiance.radiance(),
    #                                                        wavelength,
    #                                                        limb_corrected_radiance.center_pixel(),
    #                                                        limb_corrected_radiance.average_solar_radius_in_pixels(),
    #                                                        icf_solar_center)

    satellite_to_sun_distance_km = solar_image._dataset['satellite_to_sun_distance_km'].values
    expected_altius_solar_center_horz = image_geometry['altius_solar_center_horz']
    expected_altius_solar_center_vert = image_geometry['altius_solar_center_vert']
    expected_altius_solar_radius = image_geometry['altius_solar_radius']

    solar_geometry.set_disk(disk_pixel_radius=image_geometry['altius_solar_radius'],
                            disk_horizontal_center=image_geometry['altius_solar_center_horz'],
                            disk_vertical_center=image_geometry['altius_solar_center_vert'],
                            distance_to_body_km=solar_image.get_satellite_to_sun_distance_km())

    altius_image_dn = convert_to_DN_and_add_noise(altius_image,
                                                  solar_geometry.delta_omega_view(),
                                                  radiance_to_dn_factor)

    output_image = SpsSingleSolarImageInterface(altius_image_dn=altius_image_dn,
                                                calibration_wavelength=calibration_wavelength,
                                                measurement_wavelength=wavelength,
                                                distance_to_sun=satellite_to_sun_distance_km,
                                                start_pointing=np.asarray([1, 0, 0]),
                                                end_pointing=np.asarray([1, 0, 0]),
                                                integration_time=0.1,
                                                time=time,
                                                detector_position=(expected_altius_solar_center_vert,
                                                                   expected_altius_solar_center_horz))

    output_image.to_netcdf(filepath)


@click.command()
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-o', '--output_directory', type=str,
              help='Optional custom output directory for the simulated solar image files.'
                   'If not specified, files are output to $ALTIUS_L1PP_INFLIGHT$/l1pp_datafiles/'
                   'srf_registration_processed/observation/<scenario>/<channel>/raw/')
@click.option('--output_high_res', is_flag=True,
              help='If set, orders the script to save the high-resolution intermediate output of the solar images for '
                   'later ingestion into SPS as stimuli.')
@click.option('-e', '--satellite_ephemeris_file', type=str,
              help='Path to the satellite ephemeris file, with columns [x, y, z, mjd] and the same number of rows'
                   'as the number of calibration wavelengths times measurement wavelengths times the number '
                   'of detector solar positions', default=None)
def cli(config_file: str = None, output_directory: str = None,
        satellite_ephemeris_file: str = None, output_high_res: bool = False):
    config = altius_config.AltiusL1ppInflightConfig(config_file)
    run_make_solar_scenes_for_srf(config, output_directory, satellite_ephemeris_file, output_high_res)


def run_make_solar_scenes_for_srf(config: altius_config.AltiusL1ppInflightConfig,
                                  output_directory: str = None,
                                  satellite_ephemeris_file: str = None,
                                  output_high_res=False):
    """
        For all calibration wavelengths in the config file and specified channel,
        produces a netcdf file with all measurement
        wavelengths around that calibration wavelength of the sun at each location on the detector.
        The locations on the detector are TBD - radial symmetry is assumed, so it may just be a series
        of locations from the center to one of the corners of the detector.

        Parameters
        ----------
        config : AltiusL1ppInflightConfig

        output_high_res : bool
    """

    bypass_mode = False

    channel = config.get_srf_calibration_channel()
    calibration_wavelengths = config.get_srf_calibration_wavelengths(channel)
    all_measurement_wavelengths = config.get_srf_measurement_wavelengths(channel)
    num_meas_per_band = config.get_num_measurements_per_band()

    mode = config.get_srf_detector_mode()
    pixel_size = config.get_detector_pixel_dimensions(channel=channel, mode=mode)
    horizontal_width = pixel_size[1]
    vertical_width = pixel_size[0]

    boresight = config.get_pixel_boresights(channel, mode)
    scenario = config.get_solar_scenario()

    maxDN = config.get_max_DN()
    max_irradiance = config.get_max_photons_per_sec_per_nm_per_cm2()
    factor = maxDN / max_irradiance
    fov = config.get_detector_fov(channel=channel)  # (vertical, horizontal)
    satellite_height = config.get_nominal_satellite_height_km()
    scenario = SolarDataScenario(config.get_solar_scenario())
    if output_directory is None:
        output_directory = calibration_util.get_solar_srf_simulated_observation_dir(config,
                                                                                    scenario.to_string_directory())
    click.echo('Removing existing .nc files from directory {}...'.format(output_directory))

    # Configure the entire solar scan and its geometry:
    if os.path.exists(output_directory):
        try:
            shutil.rmtree(output_directory)
        except OSError as e:
            print("Error: %s : %s" % (output_directory, e.strerror))
    os.makedirs(output_directory)

    detector_positions = config.get_solar_srf_detector_positions()

    # usask_to_sps_interface = SpsHighResolutionUsaskToSpsSceneInterface()
    # sao = SolarSpectrum()

    bv, bh = config.get_pixel_boresights(channel, mode)
    bh_offset = 2 * bh / horizontal_width - 1.0
    bv_offset = 2 * bv / vertical_width - 1.0

    if channel == 'uv':
        srf_dir = os.path.join(config.get_srf_directory(), 'fpi')
        srf_for_fwhm = AltiusUVSrf(data_directory=srf_dir)
        fwhms = []
        for wl in srf_for_fwhm.tunable_wavelengths():
            fwhm, _ = srf_for_fwhm.compute_srf_approximate_fwhm(wl, srf_for_fwhm.get_wavelength_grid())
            fwhms.append(fwhm)
        # srf = GaussianSrf(srf_for_fwhm.tunable_wavelengths(), np.asarray(fwhms))
        # srf = GaussianSrf(srf_for_fwhm.tunable_wavelengths(), 3.0*np.ones_like(srf_for_fwhm.tunable_wavelengths()))
        srf = srf_for_fwhm
    else:
        srf_dir = os.path.join(config.get_srf_directory(), 'aotf')
        srf = AltiusAotfSrf(data_directory=srf_dir)

    disk_radius_apparent_angle = calibration_util.compute_solar_angular_radius(scenario.get_scenario_mjd())

    detector_positions = config.get_solar_srf_detector_positions()

    solar_scan_grid = SrfScanGrid(disk_radius_apparent_angle,
                                  0.1,
                                  0.1,
                                  fov[1], fov[0],
                                  unit_horz_positions=detector_positions[:, 1],
                                  unit_vert_positions=detector_positions[:, 0],
                                  measurements_per_position=len(all_measurement_wavelengths[0]))

    solar_scan = SolarScan(fov[1], fov[0], bh_offset, bv_offset, satellite_height, config)
    solar_scan.configure_satellite_orbit_for_scan(solar_scan_grid, scenario)

    num_measurements = len(all_measurement_wavelengths[0]) * len(detector_positions)
    click.echo('Initiating simulation for {} images of the sun'.format(num_measurements))

    solar_geometry = SolarGeometry(horizontal_width, vertical_width,
                                   bv, bh, fov[0], fov[1])

    center_pixel_radius = config.get_srf_center_pixel_radius()

    solar_cal = calibration_util.SolarCalibrationUtil(solar_geometry, srf)
    if bypass_mode:
        sps_to_usask = SpsToUsaskSolarSrfSceneInterface()
        bypass_outdir = calibration_util.get_solar_srf_intermediate_observation_dir(config,
                                                                                    scenario.to_string_directory())
        output_file = os.path.join(bypass_outdir, 'bypass_srf_center_pixels.nc')
        sps_output_calibration_file = os.path.join(output_directory, output_file)

        sps_to_usask.open_file_for_write(sps_output_calibration_file,
                                         calibration_wavelengths,
                                         detector_positions,
                                         pixel_size,
                                         boresight,
                                         fov,
                                         num_meas_per_band)
    for scene_index, scene_pos in enumerate(detector_positions):

        for cal_idx, (meas_wavelengths, cal_wl) in enumerate(zip(all_measurement_wavelengths, calibration_wavelengths)):

            psf = PsfSimpleGaussian(2.1)  # sort of arbitrary blurring
            click.secho('WARNING: PSF blurring not included in current version', fg='yellow')
            click.secho('WARNING: Satellite ephemeris ignored in current version', fg='yellow')

            click.echo('Computing raster scan {} from set {}'.format(scene_index, scene_pos))

            if bypass_mode:
                sps_to_usask.write_current_measurement_wavelengths(detector_position=scene_pos,
                                                                   calibration_wavelength=cal_wl,
                                                                   measurement_wavelengths=meas_wavelengths)
                click.echo('Running in BYPASS mode')
                solar_radius = solar_scan.get_solar_angular_radius(scene_index) * horizontal_width / fov[1]
                distance_to_sun = solar_scan.get_distance_to_sun(scene_index)
                solar_geometry.set_disk(disk_pixel_radius=solar_radius,
                                        disk_horizontal_center=scene_pos[1],
                                        disk_vertical_center=scene_pos[0],
                                        distance_to_body_km=distance_to_sun)
                gran_snr = solar_scan.get_granulation_snr()
                convolved_spectrum_irradiance = []
                ideal_irrad = []
                center_lobe_sums = []

                for mw in meas_wavelengths:
                    disk_image = solar_cal.model_noiseless_solar_disk(scene_pos[0], scene_pos[1],
                                                                      solar_radius, mw, distance_to_sun)
                    # _, _, center_lobe_sum = srf.partition_srf(mw, srf.get_wavelength_grid())
                    center_lobe_sum = 1.0
                    center_lobe_sums.append(center_lobe_sum)
                    delta_omega_view = solar_geometry.delta_omega_view()
                    ideal_irrad.append(np.nansum(disk_image * delta_omega_view))
                    disk_image = np.random.normal(disk_image, disk_image / gran_snr)
                    mask = solar_geometry.center_pixel_mask(center_pixel_radius)
                    delta_omega_view = delta_omega_view[mask]
                    center_pixels = disk_image[mask]
                    center_pixels = convert_to_DN_and_add_noise(center_pixels, delta_omega_view, factor)
                    convolved_spectrum_irradiance.append(np.nansum(center_pixels))

                # convolved_spectrum_irradiance = np.asarray(convolved_spectrum_irradiance)
                mjd_time = solar_scan.get_average_time(scene_index)

                # center_pixel_factor = compute_irradiance_center_pixel_scaling_factor(wavelengths=meas_wavelengths,
                #                                                                      solar_geometry=solar_geometry,
                #                                                                      center_radius_pixels=center_pixel_radius)

                # center_pixel_summed_irradiance = center_pixel_factor*convolved_spectrum_irradiance
                click.echo('Computing center pixels for all wavelengths around {} nm'.format(cal_wl))
                for wavelength, cls, center_pixels_irradiance in zip(meas_wavelengths, center_lobe_sums,
                                                                     convolved_spectrum_irradiance):
                    sps_to_usask.write_next_solar_scene_to_netcdf(mjd_to_datetime64(mjd_time),
                                                                  center_pixels_irradiance,
                                                                  wavelength,
                                                                  (scene_pos[0], scene_pos[1]),
                                                                  solar_radius,
                                                                  distance_to_sun,
                                                                  np.array([1, 0, 0]),
                                                                  np.array([1, 0, 0]),
                                                                  0.1,
                                                                  cls)
            else:
                solar_geometry = SolarGeometry(horizontal_width, vertical_width,
                                               bv, bh, fov[0], fov[1])

                for wl_idx, wl in enumerate(meas_wavelengths):
                    click.echo('{}/{} - computing solar image for {} nm'.format(wl_idx + 1, len(meas_wavelengths), wl))
                    filepath = os.path.join(output_directory,
                                            'observation_{}_{}_{}.nc'.format(scene_index, cal_idx, wl_idx))
                    compute_altius_image_and_save(filepath, solar_scan, scene_index, scenario,
                                                  horizontal_width, vertical_width,
                                                  srf, psf,
                                                  wl, cal_wl,
                                                  factor, solar_geometry)

    sps_to_usask.close_write()
    # usask_to_sps_interface.close()

    click.secho('*****************************************************************************', fg='green')
    click.secho('Finished writing solar SRF observation data to {}'.format(sps_output_calibration_file), fg='green')
    # click.secho('Raw stimuli solar files written to: {}'.format(raw_stimuli_output), fg='green')
    click.secho('*****************************************************************************', fg='green')


# Temporary place for CLI
if __name__ == '__main__':
    config = altius_config.AltiusL1ppInflightConfig()
    scenario = SolarDataScenario(config.get_solar_scenario())
    output_directory = calibration_util.get_solar_srf_simulated_observation_dir(config, scenario.to_string_directory())
    run_make_solar_scenes_for_srf(config, output_directory)
