import os
from datetime import datetime, timedelta

import numpy as np
import sktime.mjd as mjd_conv
from astropy import units as u
from astropy.coordinates import EarthLocation
from astropy.coordinates import Longitude, Latitude
from astropy.units import Quantity
from skretrieval.platforms import ObservationPolicy
from skretrieval.platforms.satellite import SatelliteSunSync
import pickle

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.generation.solar.hmi_files import HmiFileManager, SolarDataScenario, \
    SolarDiskRadianceResult
from altius_l1pp_inflight.scene.generation.solar.interpolate_solar_disk import SimulatedSolarDiskInterpolator
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import SolarScanGrid
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import VirtualDetectorPixelArray


class SolarScan:
    """
    A class to construct a large set of test images of the sun that scan across the detector.
    The FOV of the detector is fixed to some value, and will determine the number of images
    necessary to scan the ~0.25x0.25 degree center portion of the sun.

    For example, a 2.0x2.0 degree detector will need exactly 64 (8x8) images of the sun raster
    scanned across the detector.

    Due to the variability of solar granulation, which occurs on the order of 12 minutes, solar
    observations are performed every 12 minutes by HMI.  Since the PRNU calibration occurs over the
    span of half an orbit (45 minutes), it is not feasible to produce a single image every 12 minutes
    to be in line with the HMI data - the instrument will likely fit in all 64 (or however many) solar
    measurements for a single wavelength within that half-orbit time frame.  This lends to interpolating
    the solar disk between 12-minute scans.  Given the "scan rate" of the detector and integration time
    per image, this class interpolates the solar disk at incremental times between each 12-minute
    HMI measurement.

    HMI data is stored in the $SOLAR_HMI_DIR$ directory set as an environment variable.  Data is fixed
    by the <solar downloader> class, which downloads data from three separate occasions if it is not in the
    file location specified.  Otherwise, the data is loaded in for each of the four scenarios (no sunspots,
    one sunspot, two sunspots, and several sunspots).

    Resulting solar image scenes can be saved to netcdf in the directory specified by the $SOLAR_PROCESSED_SCENE_DIR$
    environment variable.
    """

    def __init__(self, fov_horiz_deg, fov_vert_deg, boresight_horiz_portion, boresight_vert_portion,
                 satellite_altitude_km, config: altius_config.AltiusL1ppInflightConfig):
        """
           Parameters
           ----------
           fov_horiz_deg : float
              Angular width of the detector in the horizontal direction, in degrees.

           fov_vert_deg : float
              Angular width of the detector in the vertical direction, in degrees.

            boresight_horiz_portion : float
                Empirical number between -1.0 and 1.0 describing relative horizontal offset of boresight from
                the center of the detector space

            boresight_vert_portion : float
                Empirical number between -1.0 and 1.0 describing relative vertical offset of boresight from
                the center of the detector space

            satellite_altitude_km : float
                Altitude of satellite above surface of the earth in km.
        """
        self._fov_horiz_deg = fov_horiz_deg
        self._fov_vert_deg = fov_vert_deg
        self._boresight_horiz_portion = boresight_horiz_portion
        self._boresight_vert_portion = boresight_vert_portion
        self._config = config
        self._data_dir = config.get_hmi_home_directory()
        self._file_manager = HmiFileManager(config)
        self._solar_interp: dict[SolarDataScenario, SimulatedSolarDiskInterpolator] = {}
        self._altitude_km = satellite_altitude_km
        self._current_scenario = None
        self._platform = None
        self._sat_to_sun_distances_km = None
        self._virtual_pixels = None
        self._image_snapshot_times = None

    def get_granulation_snr(self, solar_data_scenario: SolarDataScenario = SolarDataScenario.NO_SUNSPOTS):
        return self._solar_interp[solar_data_scenario].get_snr()

    def configure_satellite_at_bottom_of_orbit(self, initial_time_utc):
        """
        Based off the initial time passed in corresponding to the beginning of the HMI
        dataset, configures a sun-synchronous satellite to begin its orbit near the south pole
        so that the entire time for the solar observation is in sunlight.

        Parameters
        ----------
        initial_time_utc : datetime
           First time in the HMI dataset for the given scenario

        Returns
        -------
        satellite : SatelliteSunSync
           The satellite location class definining the position of the satellite for the orbit
        """
        # This roughly places the initial measurement at the bottom of the orbit, I think
        dummy_sat = SatelliteSunSync(utc=initial_time_utc, orbittype='sgp4',
                                     period_from_altitude=self._altitude_km * 1000.0,
                                     localtime_of_ascending_node_hours=initial_time_utc.hour + initial_time_utc.minute / 60)
        orbit_period = dummy_sat.period()
        ascending_node = initial_time_utc + orbit_period / 4
        satellite = SatelliteSunSync(utc=initial_time_utc, orbittype='sgp4',
                                     period_from_altitude=self._altitude_km * 1000.0,
                                     localtime_of_ascending_node_hours=ascending_node.hour + ascending_node.minute / 60)
        return satellite

    def configure_satellite_times_and_positions(self, solar_scan_grid: SolarScanGrid, initial_utc,
                                                satellite, total_time_seconds):
        """
        Propagates a satellite along an orbit defined by :code:`satellite`

        Parameters
        ----------

        initial_utc : datetime
           The intial time of the satellite orbit placed at the "bottom" of the orbit

        satellite : PlatformLocation
            Satellite orbit propagator, likely an instance of :code:`SatelliteSunSynchronous`

        total_time_seconds :
            Total time for imaging the necessary number of raster images.

        Returns
        -------
        np.ndarray, np.ndarray
           First array is image snapshot times (datetime format), second arry is the satellite positions as the
           astropy type :code:`EarthLocation`
        """
        sat_earthlocation_positions = []
        num_images = solar_scan_grid.get_num_grid_positions() * solar_scan_grid.get_num_measurements_per_position()
        dt = timedelta(seconds=total_time_seconds / num_images)
        image_snapshot_times = []
        for i in range(num_images):
            tnow = initial_utc + i * dt
            image_snapshot_times.append(tnow)
            satellite.update_position(tnow)
            llh = satellite.lat_lon_height
            spos: EarthLocation = EarthLocation.from_geodetic(
                lon=Longitude(angle=llh[1], unit=u.deg),
                lat=Latitude(angle=llh[0], unit=u.deg),
                height=Quantity(value=llh[2], unit=u.meter))
            sat_earthlocation_positions.append(spos)
        sat_earthlocation_positions = np.array(sat_earthlocation_positions, dtype=EarthLocation)
        return np.array(image_snapshot_times, dtype=datetime), np.array(sat_earthlocation_positions,
                                                                        dtype=EarthLocation)

    def get_indexed_image(self, scenario: SolarDataScenario, index: int, save_data=False):

        self._validate_cache(False, scenario)
        if self._current_scenario == scenario:

            print("**Computing orbit and images for scenario ", scenario.to_string_directory(), '**')

            obs: ObservationPolicy = self._platform.observation_policy  # Get the observation set used for this back_end_radiance calculation                                                                                    # Get the number of instantaneous lines of sight from the second dimension of the array. LOS are specified in the instrument control frame

            rot = obs.icf_to_ecef_rotation_matrix()  # Copy the rotation matrices for each sample in the observation set.

            return self._virtual_pixels.compute_hi_res_image_at_scan_index(self._platform, index,
                                                                           rot[index],
                                                                           self._sat_to_sun_distances_km[index],
                                                                           save_data, scenario,
                                                                           self._solar_interp[scenario],
                                                                           self._image_snapshot_times[index],
                                                                           self._config)
        else:
            print('Geometry not configured for scenario ' + scenario.to_string_directory())
            return -1

    def configure_satellite_orbit_for_scan(self, solar_scan_grid: SolarScanGrid,
                                           scenario: SolarDataScenario):
        self._solar_scan_grid = solar_scan_grid
        self._validate_cache(False, scenario)
        grid_mjd_times = self._solar_interp[scenario].grid_mjd_times()
        # Build the satellite positions by starting the satellite looking at the sun for the first image at
        # the time of the first grid time
        platform_utc = mjd_conv.mjd_to_datetime(grid_mjd_times[0])
        satellite = self.configure_satellite_at_bottom_of_orbit(platform_utc)
        self._image_snapshot_times, sat_earthlocation_positions = self.configure_satellite_times_and_positions(
            solar_scan_grid,
            platform_utc,
            satellite,
            solar_scan_grid.get_total_time_seconds())
        self._virtual_pixels = VirtualDetectorPixelArray(
            self._solar_interp[scenario].solar_disk_average_radius_in_pixels(),
            solar_scan_grid, self._boresight_horiz_portion,
            self._boresight_vert_portion)
        self._platform, self._sat_to_sun_distances_km = self._virtual_pixels.make_platform(satellite,
                                                                                           self._image_snapshot_times,
                                                                                           sat_earthlocation_positions)
        self._current_scenario = scenario

    # def produce_and_save_all_test_images(self, recompute_cache=False):
    #     """
    #     Top level method to produce the solar scenes for a given center wavelength, SRF, and satellite configuration.
    #     Will produce a file for each solar position on the raster scan of the sun.  All files are stored in the same
    #     directory.
    #
    #     Result will be a set of n files for the wavelength in each sunspot scenario.
    #
    #     Parameters
    #     ----------
    #     time_per_image_sec : float
    #         Integration time of the detector for each solar image in seconds.
    #
    #     scan_rate_deg_per_sec : float
    #         Scan rate in degrees per second which the satellite is capable of moving.
    #
    #     recompute_cache : bool
    #         If true, reloads the HMI raw data files.  If false, loads the pickle files that speed up the
    #         initial calculation time.
    #     """
    #     # self._validate_cache(recompute_cache)
    #
    #     for scenario in HmiSunspotScenario:
    #         self._compute_and_save_cache(scenario)
    #         self.produce_images_for_scenario(scenario)

    def _validate_cache(self, recompute_cache, scenario):
        """
        Method to speed up the loading of HMI data TODO not done yet
        """
        self._cache_file_present = os.path.exists(self._cache_filename(scenario))

        if scenario not in self._solar_interp.keys():
            if self._cache_file_present is False:
                # if there's no cache and no file, it needs to be computed
                self._solar_interp[scenario] = self._compute_and_save_cache(scenario)
            else:
                with open(self._cache_filename(scenario), 'rb') as f:
                    self._solar_interp[scenario] = pickle.load(f)
                    f.close()

        elif recompute_cache is True:
            self._solar_interp.pop(scenario)
            self._solar_interp[scenario] = self._compute_and_save_cache(scenario)

    def _cache_filename(self, scenario):
        return os.path.join(self._data_dir, 'disk_interpolator_cache_{}.pickle'.format(scenario.to_string_directory()))

    def _compute_and_save_cache(self, scenario):
        """
        Populate the cache with the solar disk interpolators for each scenario - TODO not done yet
        """
        if scenario not in self._solar_interp.keys():

            results = self._file_manager.load_scenario_files_as_images(scenario)

            all_scaled_images = []
            for res in results:
                if not isinstance(res, SolarDiskRadianceResult):
                    raise AssertionError('Object is not of type HmiSolarDiskRadianceResult')
                all_scaled_images.append(res)

            all_scaled_images = np.array(all_scaled_images)
            solar_interp_entry = SimulatedSolarDiskInterpolator(all_scaled_images)
            with open(self._cache_filename(scenario), 'wb') as f:
                pickle.dump(solar_interp_entry, f)
                f.close()
            return solar_interp_entry

        # self._solar_disk_interpolator.dump()

    def get_average_time(self, scene_index):
        idx = self._solar_scan_grid.get_scan_indices()
        snapshot_idx = np.argwhere(idx == scene_index)
        from sktime.mjd import datetime_to_mjd
        all_t = np.asarray([datetime_to_mjd(t) for t in self._image_snapshot_times])
        return np.mean(all_t[snapshot_idx])

    def get_distance_to_sun(self, scene_index):
        t = self.get_average_time(scene_index)
        grid_mjd_times = self._solar_interp[self._current_scenario].grid_mjd_times()
        idx = np.argmin(np.abs(t - grid_mjd_times))

        sun: SolarDiskRadianceResult = self._solar_interp[self._current_scenario].get_sim_ideal_grid_images()[idx]

        return sun.actual_solar_distance()

    def get_solar_angular_radius(self, scene_index):
        t = self.get_average_time(scene_index)
        grid_mjd_times = self._solar_interp[self._current_scenario].grid_mjd_times()
        idx = np.argmin(np.abs(t - grid_mjd_times))

        sun: SolarDiskRadianceResult = \
            self._solar_interp[self._current_scenario].get_sim_ideal_grid_images()[idx]

        return sun.angular_radius_degrees()
