import numpy as np


class SolarScanGrid:
    """
    Abstract class representing the way the satellite scans the sun across
    the surface of the detector.
    """

    def __init__(self, solar_radius_angular_size_deg: float,
                 scan_rate_deg_per_sec: float,
                 time_per_image_sec: float,
                 fov_horizontal_deg: float,
                 fov_vertical_deg: float):
        # we will use the box of size disk_radius_pixels in the center of the sun
        self._integration_time_sec = time_per_image_sec
        self._scan_rate_deg_per_sec = scan_rate_deg_per_sec
        self._solar_radius_angular_size_deg = solar_radius_angular_size_deg
        self._fov_horiz_deg = fov_horizontal_deg
        self._fov_vert_deg = fov_vertical_deg

    def get_num_grid_positions(self):
        pass

    def get_num_measurements_per_position(self):
        pass

    def get_total_time_seconds(self):
        pass

    def solar_disc_radius_angle(self):
        return self._solar_radius_angular_size_deg

    def scan_rate_degrees_per_sec(self):
        return self._scan_rate_deg_per_sec

    def integration_time_sec(self):
        return self._integration_time_sec

    def fov_horz_deg(self) -> float:
        return self._fov_horiz_deg

    def fov_vert_deg(self) -> float:
        return self._fov_vert_deg

    def get_boresight_location(self, disk_radius_pixels: float,
                               boresight_horz_portion: float,
                               boresight_vert_portion: float) -> (float, float):
        pass

    def get_infield_virtual_pixel_size(self, disk_radius_pixels: float):
        pass

    def get_downscaling_grid_boundaries(self,
                                        disk_radius_pixels: float,
                                        pix_per_vertical: float,
                                        pix_per_horizontal: float):
        pass

    def get_detector_location(self, disk_radius_pixels, index):
        pass

    def get_corner_pixel_coords(self, disk_radius_pixels) -> dict:
        pass

    def get_scan_indices(self):
        pass


class SrfScanGrid(SolarScanGrid):
    """
    Initializes the positions of the sun for the SRF calibration, which are arbitrary
    positions on the detector defined in a normalized detector space, [0, 1] on the
    horizontal and [0, 1] on the vertical axis.
    e.g. to define a set of 4 measurements along the horizontal axis in the vertical
    center of the detector, one could pass in:
        unit_horz_positions: [0.1, 0.3, 0.7, 0.9]
        unit_vert_positions: [0.5, 0.5, 0.5, 0.5]
    """

    def __init__(self, solar_radius_angular_size_deg: float,
                 scan_rate_deg_per_sec: float,
                 time_per_image_sec: float,
                 fov_horizontal_deg: float,
                 fov_vertical_deg: float,
                 unit_horz_positions: np.ndarray,
                 unit_vert_positions: np.ndarray,
                 measurements_per_position: int):
        super().__init__(solar_radius_angular_size_deg,
                         scan_rate_deg_per_sec,
                         time_per_image_sec,
                         fov_horizontal_deg,
                         fov_vertical_deg)

        self._measurements_per_position = measurements_per_position
        self._num_grid_pos = len(unit_horz_positions)
        time_per_grid_pos = solar_radius_angular_size_deg / scan_rate_deg_per_sec + time_per_image_sec
        self._total_time_seconds = self._num_grid_pos * time_per_grid_pos

        self._unit_horz_positions = unit_horz_positions
        self._unit_vert_positions = unit_vert_positions

    def get_num_grid_positions(self):
        return self._num_grid_pos

    def get_num_measurements_per_position(self):
        return self._measurements_per_position

    def get_total_time_seconds(self):
        return self._total_time_seconds

    def solar_disc_radius_angle(self):
        return self._solar_radius_angular_size_deg

    def get_boresight_location(self, disk_radius_pixels: float,
                               boresight_horz_portion: float,
                               boresight_vert_portion: float) -> (float, float):
        n_vert, n_horz = self.get_infield_virtual_pixel_size(disk_radius_pixels)

        boresight_virtual_px_horz = (n_horz / 2) * (1 - boresight_horz_portion)  # minus because +y is to the left
        boresight_virtual_pixel_vert = (n_vert / 2) * (1 + boresight_vert_portion)

        return boresight_virtual_pixel_vert, boresight_virtual_px_horz

    def get_infield_virtual_pixel_size(self, disk_radius_pixels: float):
        n_horz = disk_radius_pixels * self.fov_horz_deg() / self.solar_disc_radius_angle()
        n_vert = disk_radius_pixels * self.fov_vert_deg() / self.solar_disc_radius_angle()

        return n_vert, n_horz

    def get_downscaling_grid_boundaries(self,
                                        disk_radius_pixels: float,
                                        pix_per_vertical: float,
                                        pix_per_horizontal: float):
        n_vert, n_horz = self.get_infield_virtual_pixel_size(disk_radius_pixels)
        grid_boundaries_vert = np.arange(0, n_vert, pix_per_vertical)
        grid_boundaries_horz = np.arange(0, n_horz, pix_per_horizontal)
        grid_boundaries_vert = np.append(grid_boundaries_vert, n_vert)
        grid_boundaries_horz = np.append(grid_boundaries_horz, n_horz)

        return grid_boundaries_vert, grid_boundaries_horz

    def get_detector_location(self, disk_radius_pixels, index):
        n_vert, n_horz = self.get_infield_virtual_pixel_size(disk_radius_pixels)

        return self._unit_vert_positions[index] * n_vert, self._unit_horz_positions[index] * n_horz

    def get_corner_pixel_coords(self, disk_radius_pixels) -> dict:
        n_vert, n_horz = self.get_infield_virtual_pixel_size(disk_radius_pixels)

        bl_y, bl_z = 0, 0
        br_y, br_z = n_horz, 0
        tl_y, tl_z = 0, n_vert
        tr_y, tr_z = n_horz, n_vert

        coords = {'bottom_left': [bl_y, bl_z],
                  'bottom_right': [br_y, br_z],
                  'top_left': [tl_y, tl_z],
                  'top_right': [tr_y, tr_z]}

        return coords

    def get_scan_indices(self):
        idx = []
        for i, _ in enumerate(self._unit_horz_positions):
            idx.append(i * np.ones((self._measurements_per_position)).astype(int))

        return np.asarray(idx)


class RasterScanGrid(SolarScanGrid):
    """
    Container class holding the raster imaging grid of the sun for PRNU calibration, starting at the bottom right
    of the detector, moving right, then up one row, then moving left across the detector, and so on.
    Each grid cell is a box of width radius of the sun in the image.  The sun is centered in each
    box at each raster position, so that the center portion of the sun is used.

    The grid is configured to have a row and column of images directly outside the FOV, so that if
    8 x 8 grid boxes fit inside the FOV, the final raster scan grid will be 10 x 10.
    """

    def __init__(self, solar_radius_angular_size_deg: float, scan_rate_deg_per_sec: float, time_per_image_sec: float,
                 fov_horizontal_deg: float, fov_vertical_deg: float, v_in: int, h_in: int, padding=0):
        super().__init__(solar_radius_angular_size_deg,
                         scan_rate_deg_per_sec,
                         time_per_image_sec,
                         fov_horizontal_deg,
                         fov_vertical_deg)
        # we will use the box of size disk_radius_pixels in the center of the sun
        self._padding = padding

        # h_in = int(np.ceil(fov_horizontal_deg/solar_radius_angular_size_deg))
        # v_in = int(np.ceil(fov_vertical_deg/solar_radius_angular_size_deg))
        self._num_grid_cells_horiz = h_in + 2 * self._padding
        self._num_grid_cells_vert = v_in + 2 * self._padding

        self._grid_angular_step_size_horiz = fov_horizontal_deg / h_in
        self._grid_angular_step_size_vert = fov_vertical_deg / v_in
        self._num_images = self._num_grid_cells_horiz * self._num_grid_cells_vert
        self._total_time_seconds = self._num_images * solar_radius_angular_size_deg / scan_rate_deg_per_sec + self._num_images * time_per_image_sec

        self._column_indices, self._row_indices = self.get_row_col_meshgrid()

    def grid_horizontal_step_in_pixels(self, solar_pixel_radius):
        return self._grid_angular_step_size_horiz * solar_pixel_radius / self._solar_radius_angular_size_deg

    def grid_vertical_step_in_pixels(self, solar_pixel_radius):
        return self._grid_angular_step_size_vert * solar_pixel_radius / self._solar_radius_angular_size_deg

    def _num_scans_horizontal_infield(self):
        return self._num_grid_cells_horiz - 2 * self._padding

    def _num_scans_vertical_infield(self):
        return self._num_grid_cells_vert - 2 * self._padding

    def get_row_col_meshgrid(self):
        COL, ROW = np.meshgrid(np.arange(self._num_grid_cells_vert),
                               np.arange(self._num_grid_cells_horiz))

        # reverse the order of every second row of the column array -> this is scanning right to left
        COL[1::2, :] = np.fliplr(COL[1::2, :])
        ROW = ROW.flatten()
        COL = COL.flatten()
        return COL, ROW

    def get_num_grid_positions(self):
        return self._num_images

    def get_num_measurements_per_position(self):
        return 1

    def get_total_time_seconds(self):
        return self._total_time_seconds

    def _get_total_pixels(self, disk_radius_pixels):
        grid_step_horz_px_size = self.grid_horizontal_step_in_pixels(disk_radius_pixels)
        grid_step_vert_px_size = self.grid_vertical_step_in_pixels(disk_radius_pixels)

        num_horz_virtual_px = grid_step_horz_px_size * self._num_grid_cells_horiz
        num_vert_virtual_px = grid_step_vert_px_size * self._num_grid_cells_vert
        return num_vert_virtual_px, num_horz_virtual_px

    def get_detector_location(self, disk_radius_pixels, index):
        """
        Returns the pixel coordinate relative to the 0, 0 pixel position in virtual detector space
        of the center pixel specified by the scan index.

        :param disk_radius_pixels:
        :param index:
        :return:
        """
        grid_step_horz_px_size = self.grid_horizontal_step_in_pixels(disk_radius_pixels)
        grid_step_vert_px_size = self.grid_vertical_step_in_pixels(disk_radius_pixels)
        row_index = self._row_indices[index]
        column_index = self._column_indices[index]

        vert_pixel_coord = (row_index + 0.5) * grid_step_vert_px_size
        horz_pixel_coord = (column_index + 0.5) * grid_step_horz_px_size

        return vert_pixel_coord, horz_pixel_coord

    def get_boresight_location(self, disk_radius_pixels: float,
                               boresight_horz_portion: float,
                               boresight_vert_portion: float):
        grid_step_horz = self.grid_horizontal_step_in_pixels(disk_radius_pixels)
        grid_step_vert = self.grid_vertical_step_in_pixels(disk_radius_pixels)
        n_horz = grid_step_horz * self._num_grid_cells_horiz
        n_vert = grid_step_vert * self._num_grid_cells_vert

        boresight_virtual_px_horz = (n_horz / 2 - self._padding * grid_step_horz) * \
                                    (1 - boresight_horz_portion) + self._padding * grid_step_horz  # minus because +y is to the left
        boresight_virtual_pixel_vert = (n_vert / 2 - self._padding * grid_step_vert) * \
                                       (1 + boresight_vert_portion) + self._padding * grid_step_vert

        return boresight_virtual_pixel_vert, boresight_virtual_px_horz

    def get_infield_virtual_pixel_size(self, disk_radius_pixels: float):
        n_vertical = self._num_scans_vertical_infield() * self.grid_vertical_step_in_pixels(disk_radius_pixels)
        n_horizontal = self._num_scans_horizontal_infield() * self.grid_horizontal_step_in_pixels(disk_radius_pixels)

        return n_vertical, n_horizontal

    def get_downscaling_grid_boundaries(self,
                                        disk_radius_pixels: float,
                                        pix_per_vertical: float,
                                        pix_per_horizontal: float):
        step_v = self.grid_vertical_step_in_pixels(disk_radius_pixels)
        step_h = self.grid_horizontal_step_in_pixels(disk_radius_pixels)
        n_vert, n_horz = self._get_total_pixels(disk_radius_pixels)
        grid_boundaries_vert = np.arange(self._padding * step_v,
                                         n_vert - self._padding * step_v,
                                         pix_per_vertical)
        grid_boundaries_horz = np.arange(self._padding * step_h,
                                         n_horz - self._padding * step_h,
                                         pix_per_horizontal)
        grid_boundaries_vert = np.append(grid_boundaries_vert,
                                         n_vert - self._padding * step_v)
        grid_boundaries_horz = np.append(grid_boundaries_horz,
                                         n_horz - self._padding * step_h)

        return grid_boundaries_vert, grid_boundaries_horz

    def get_corner_pixel_coords(self, disk_radius_pixels) -> dict:
        step_v = self.grid_vertical_step_in_pixels(disk_radius_pixels)
        step_h = self.grid_horizontal_step_in_pixels(disk_radius_pixels)
        n_vertical = self._num_grid_cells_vert * step_v
        n_horizontal = self._num_grid_cells_horiz * step_h

        bl_y, bl_z = (step_v, step_h)
        br_y, br_z = (n_horizontal - step_h, step_v)
        tl_y, tl_z = (step_h, n_vertical - step_v)
        tr_y, tr_z = (n_horizontal - step_h, n_vertical - step_v)

        coords = {'bottom_left': [bl_y, bl_z],
                  'bottom_right': [br_y, br_z],
                  'top_left': [tl_y, tl_z],
                  'top_right': [tr_y, tr_z]}

        return coords

    def get_scan_indices(self):
        return np.arange(0, self._num_images)
