import os
from datetime import datetime

import matplotlib.patches as pch
import matplotlib.pyplot as plt
import numpy as np
import sktime.mjd as mjd_conv
import xarray as xr
from astropy.coordinates import EarthLocation
from astropy.coordinates import SkyCoord
from astropy.coordinates import solar_system, solar_system_ephemeris
from astropy.time import Time
from scipy.interpolate import interp2d
from skimage.transform import rescale
from skretrieval.platforms import Platform, PlatformLocation
from skretrieval.platforms.platform.rotationmatrix import RotationMatrix
from skretrieval.util import rotation_matrix

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.instrument.srf.srf_interface import SrfInterface
from altius_l1pp_inflight.scene.disk_geometry import SOLAR_RADIUS, SolarGeometry, D_1AU
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDiskRadianceResult, SolarDataScenario
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import SolarScanGrid, RasterScanGrid

solar_system_ephemeris.set('jpl')


def pixels_to_distance_km(pixels, disk_radius_pixels):
    return pixels * SOLAR_RADIUS / disk_radius_pixels


# Utility, if needed:
def _chunker(los, size):  # LOS is [3, nlos]
    return [los[:, pos:pos + size] for pos in range(0, len(los[0, :]), size)]


class VirtualDetectorPixelArray:
    """
    A class for handling the calculation of where the sun needs to be on the detector.
    The actual implementation works in "virtual pixels" - imagining the entire surface is
    a grid of pixels at the same resolution as the HMI image.

    For an HMI solar image of 4096 x 4096, the solar radius is about 1950 pixels.  For a
    2.0 by 2.0 degree detector FOV, and assuming the solar disk is 0.5 degrees across,
    this corresponds to 8 x 8 grid of raster boxes across the surface, plus padding rows on the sides,
    top, and bottom, giving a total of a 10 x 10 grid, and a "virtual pixel
    array" of 19500 x 19500 pixels.  These aren't actually all stored in memory though - only
    the 4096 x 4096 array of radiances for the solar disk are stored - they are simply translated
    to the appropriate location on the virtual pixel array for the given raster position.

    The (0, 0) virtual pixel is defined as the bottom left of the grid, and corresponds to the maximum
    y and minimum z value.  That is, for the 2.0 by 2.0 degree FOV with padding of ~0.25 deg on each side,
    the (0, 0) virtual pixel corresponds to (2.25 deg, -2.25 deg).  The boresight is determined from the offset
    parameters passed in - values from -1.0 to 1.0 that determine the position of the boresight relative to the
    detector space.  An example with easy numbers is shown below:

    HMI image: 4096 x 4096
    Detector: 2.0 x 2.0 deg
    Solar Disk: 0.5 deg (diameter), radius in pixels - 2000 - this becomes the raster grid step size
    Number of rows: 2.0 / 0.25 = 8, Number of columns: 2.0 / 0.25 = 8
    With padding, nrows = 10, ncols = 10
    Total grid size in virtual pixels: 10*2000 = 20000 x 20000 virtual pixels
    Actual detector space spans from [2000, 2000] to [18000, 18000], giving a width of 16000 virtual pixels
    For boresight parameters of 0.1, -0.2 passed in, the boresight would be located at virtual pixel location:
    0.1 * 16000 = 1600 pixels to the left of center
    -0.2 * 16000 = 3200 pixels below center
    Nominal center -> [10000, 10000].  Actual boresight -> [8400, 6800]

    This is an ideal situation - the sun radius doesn't fit exactly into the detector area.  A box of width
    the radius of the sun is used as a first guess and maximum bound to fit into the detector space.  For a solar radius
    of 0.26 deg and a detector width of 2.1 deg, we would get 8.07 boxes.  To keep the box size less than the maximum
    bound of the solar radius (in order to use the valid area of the sun limb darkening model), we round up the result
    to obtain 9 boxes.  This corresponds to a 2.1 / 9 = 0.233 deg wide box.

    Parameters
    ----------

        disk_radius_pixels : float
           Solar image radius in HMI pixels.

        solar_scan_grid : altius_l1pp_inflight.scene.generation.solar.solar_scan_grid.RasterScanGrid
           Raster scan grid used to define the direction and order of raster scanning and each grid location
           on the detector.  Implementation is to start at the bottom left, and work in an full-across/one-up
           approach.

        boresight_horz_portion: float
            Ratio describing the horizontal shift of the boresight of the detector.  +1.0 would place the boresight at
            the left edge of the detector, while -1.0 would place it at the right edge of the detector

        boresight_vert_portion: float
            Ratio describing the vertical shift of the boresight of the detector.  +1.0 would place the boresight at
            the top edge of the detector, while -1.0 would place it at the bottom edge of the detector

    """

    def __init__(self, disk_radius_pixels: float,
                 solar_scan_grid: SolarScanGrid,
                 boresight_horz_portion: float,
                 boresight_vert_portion: float):

        self._num_altius_pixels_vertical = 0
        self._num_altius_pixels_horizontal = 0
        self._boresight_horizontal_portion = boresight_horz_portion
        self._boresight_vertical_portion = boresight_vert_portion
        self._solar_scan_grid = solar_scan_grid
        self._solar_disk_radius_pixels = disk_radius_pixels
        # self._grid_step_horz_px_size = solar_scan_grid.grid_horizontal_step_in_pixels(disk_radius_pixels)
        # self._grid_step_vert_px_size = solar_scan_grid.grid_vertical_step_in_pixels(disk_radius_pixels)

        # COL, ROW = solar_scan_grid.get_row_col_meshgrid()

        # self._num_rows = solar_scan_grid.num_vert_grid_cells_including_padding()
        # self._num_cols = solar_scan_grid.num_horz_grid_cells_including_padding()

        # self._num_horz_virtual_px = self._grid_step_horz_px_size * solar_scan_grid.num_horz_grid_cells_including_padding()
        # self._num_vert_virtual_px = self._grid_step_vert_px_size * solar_scan_grid.num_vert_grid_cells_including_padding()

        # Figure out the boresight virtual pixel location based on the offsets, accounting for the padding
        b = self._solar_scan_grid.get_boresight_location(disk_radius_pixels,
                                                         boresight_vert_portion,
                                                         boresight_horz_portion)

        self._boresight_virtual_px_vert = b[0]
        self._boresight_virtual_px_horz = b[1]

        infield_size = self._solar_scan_grid.get_infield_virtual_pixel_size(disk_radius_pixels)

        self._num_virtual_detector_vertical = infield_size[0]
        self._num_virtual_detector_horizontal = infield_size[1]

        # self._column_indices = COL
        # self._row_indices = ROW

    def get_satellite_orientation_itrs(self, time: datetime, scene_index: int, spos: EarthLocation):
        """
        For the satellite position :code:`spos` and :code:`time`, uses astropy to determine the satellite-to-sun
        vector in ITRS coordinates, then use subsequent calls to determine the orientation of the satellite based
        on the raster index.

        Parameters
        ----------
        time : datetime
           Time of the simulated measurement.

        scene_index : int
           Index for the raster scan, used to determine where on the detector the sun is located.

        spos : EarthLocation
            Position of the satellite in lat/lon/height astropy coordinates at the time requested.

        Returns
        -------
        orientation : np.ndarray
           Orientation of the satellite in ITRS, described in a 6 element array by the x axis (boresight) and z axis (vertical
           detector plane) - [x[0], x[1], x[2], z[0], z[1], z[2]]

        solar_distance_km : float
            Distance from the sun to the satellite in km
        """
        mjd = mjd_conv.datetime_to_mjd(time)

        sat_to_solar_itrs: SkyCoord = solar_system.get_body(
            'sun', Time(mjd, format='mjd'), location=spos)
        sat_to_solar_itrs.representation_type = 'cartesian'

        sat_to_solar_itrs = sat_to_solar_itrs.transform_to('itrs')

        # Compute the virtual pixel shift:
        horizontal_center_shift_sun, vertical_center_shift_sun = self.get_solar_disk_detector_pixel_shift_from_boresight(
            scene_index)

        solar_distance_km = np.sqrt(sat_to_solar_itrs.x.value ** 2 +
                                    sat_to_solar_itrs.y.value ** 2 +
                                    sat_to_solar_itrs.z.value ** 2)

        # Using the virtual pixel shift, determine how we need to rotate the satellite:
        orientation = self.extract_satellite_orientation(horizontal_center_shift_sun, vertical_center_shift_sun,
                                                         sat_to_solar_itrs)
        return orientation, solar_distance_km

    def make_platform(self, satellite: PlatformLocation, utc_datetimes: np.ndarray, sat_earthlocations: np.ndarray):
        """
        Sets up the platform class of the satellite given the times and locations determined by a
        satellite propagator, and determines its orientation at each point in the orbit.

        Parameters
        ----------
        satellite : PlatformLocation
           Platform location class, typically a SatelliteSunSynchronous implementation

        utc_datetimes : np.ndarray
           Array of datetimes of the measurements/satellite locations as it is propagated in its orbit.

        sat_earthlocations : np.ndarray
            Array of astropy EarthLocation objects defining the satellite's orbit positions.

        Returns
        -------
        platform : Platform
            The final oriented platform, located at each point in its orbit and oriented so that the
            sun is in the appropriate place on the detector.
        sat_to_solar_distance_km : np.ndarray
            Distance from the satellite to the sun in km
        """
        unit_look_vectors = np.zeros((len(utc_datetimes), 6))
        sat_to_solar_distance_km = np.zeros(len(utc_datetimes))
        all_scene_indexes = self._solar_scan_grid.get_scan_indices()
        idx = 0
        for scene_indices in all_scene_indexes:
            for scene_index, time in zip(scene_indices, utc_datetimes):
                # Get the necessary orientation at this orbit position to place the sun in the correct raster position on the detector:
                u, d = self.get_satellite_orientation_itrs(time, scene_index, sat_earthlocations[scene_index])
                unit_look_vectors[idx, :] = u
                sat_to_solar_distance_km[idx] = d
                idx += 1

        platform = Platform(platform_locator=satellite)

        platform.add_measurement_set(utc=utc_datetimes, observer_definitions=('from_platform',),
                                     pointing_definitions=('unit_vectors', 'ignored_roll_control', unit_look_vectors))
        platform.make_observation_policy()
        return platform, sat_to_solar_distance_km

    def set_current_scanned_image(self, solar_image: SolarDiskRadianceResult,
                                  num_altius_pixels_vertical: int,
                                  num_altius_pixels_horizontal: int,
                                  scan_index: int):
        """
        Sets the current virtual image array and necessary parameters for binning to the chosen resolution
        Parameters
        ----------
        solar_image : SolarDiskRadianceResult

        num_altius_pixels_horizontal : int
            Number of pixels across the horizontal dimension of the ALTIUS detector

        num_altius_pixels_vertical : int
            Number of pixels across the vertical dimension of the ALTIUS detector
        """

        self._num_altius_pixels_horizontal = num_altius_pixels_horizontal
        self._num_altius_pixels_vertical = num_altius_pixels_vertical

        self._horz_scale = self._num_altius_pixels_horizontal / self._num_virtual_detector_horizontal
        self._vert_scale = self._num_altius_pixels_vertical / self._num_virtual_detector_vertical

        self._current_hr_disk_radiance = solar_image.radiance()
        self._current_hr_disk_radiance[np.isnan(self._current_hr_disk_radiance)] = 0.0

        self._set_current_virtual_pixel_centers(scan_index, solar_image)

        pix_per_vertical = 1 / self._vert_scale
        pix_per_horizontal = 1 / self._horz_scale

        self._altius_actual_radius = self._solar_disk_radius_pixels / pix_per_vertical

        # make sure to determine the whole grid before reducing it down to the solar disk:
        # Note: these are **pixel edges**
        gb = self._solar_scan_grid.get_downscaling_grid_boundaries(self._solar_disk_radius_pixels,
                                                                   pix_per_vertical,
                                                                   pix_per_horizontal)

        self._grid_boundaries_vert = gb[0]
        self._grid_boundaries_horz = gb[1]

        self._set_altius_resolution_solar_center()
        self._trim_arrays_to_disk_region()

    def _set_current_virtual_pixel_centers(self, scan_index, solar_image: SolarDiskRadianceResult):
        # horz_rel_to_boresight, vert_rel_to_boresight = \
        #     self.get_solar_disk_detector_pixel_shift_from_boresight(scan_index)
        center = solar_image.center_pixel()

        sol_center_virt = self._solar_scan_grid.get_detector_location(self._solar_disk_radius_pixels, scan_index)

        self._solar_center_virtual_vert = sol_center_virt[0]
        self._solar_center_virtual_horz = sol_center_virt[1]
        vert_rel_to_boresight = self._solar_center_virtual_vert - self._boresight_virtual_px_vert
        horz_rel_to_boresight = self._boresight_virtual_px_horz - self._solar_center_virtual_horz
        # self._solar_center_virtual_horz = self._boresight_virtual_px_horz - horz_rel_to_boresight
        # self._solar_center_virtual_vert = self._boresight_virtual_px_vert + vert_rel_to_boresight

        boresight_image_frame_vert = center[1] - vert_rel_to_boresight
        boresight_image_frame_horz = center[0] + horz_rel_to_boresight

        self._solar_geometry = SolarGeometry(len(solar_image.radiance()[0]), len(solar_image.radiance()),
                                             boresight_image_frame_vert, boresight_image_frame_horz,
                                             self._solar_scan_grid.fov_vert_deg(),
                                             self._solar_scan_grid.fov_horz_deg())
        virt_solar_center_vert = vert_rel_to_boresight + self._boresight_virtual_px_vert
        virt_solar_center_horz = self._boresight_virtual_px_horz - horz_rel_to_boresight  # because of the orientation of +y axis
        self._solar_geometry.set_disk(self._current_hr_disk_radiance, center[0], center[1], D_1AU)

        dh, dv = self._solar_geometry.get_delta_pixels_relative_disk_center()

        self._vertical_virtual_pixel_centers = dv[:, 0].flatten() + virt_solar_center_vert
        self._horizontal_virtual_pixel_centers = dh[0] + virt_solar_center_horz

        self._image_center_vert = center[1]
        self._image_center_horz = center[0]

    def _trim_arrays_to_disk_region(self):
        altius_horizontal_idx = np.intersect1d(
            np.argwhere(self._grid_boundaries_horz <= np.max(self._horizontal_virtual_pixel_centers)),
            np.argwhere(self._grid_boundaries_horz >= np.min(self._horizontal_virtual_pixel_centers)))
        altius_vertical_idx = np.intersect1d(
            np.argwhere(self._grid_boundaries_vert <= np.max(self._vertical_virtual_pixel_centers)),
            np.argwhere(self._grid_boundaries_vert >= np.min(self._vertical_virtual_pixel_centers)))
        # Trim them down to the solar disk area:
        self._grid_boundaries_vert = self._grid_boundaries_vert[altius_vertical_idx]
        self._grid_boundaries_horz = self._grid_boundaries_horz[altius_horizontal_idx]
        self._altius_vertical_idx = altius_vertical_idx  # altius_vertical_idx[:-1]
        self._altius_horizontal_idx = altius_horizontal_idx  # altius_horizontal_idx[:-1]

    def _set_altius_resolution_solar_center(self):

        lower_bnd_center_vert_idx = np.where(self._grid_boundaries_vert < self._image_center_vert)[0][-1]
        lower_bnd_center_horz_idx = np.where(self._grid_boundaries_horz < self._image_center_horz)[0][-1]
        lv = self._grid_boundaries_vert[lower_bnd_center_vert_idx]
        uv = self._grid_boundaries_vert[lower_bnd_center_vert_idx + 1]
        self._altius_solar_center_vert = lower_bnd_center_vert_idx + (self._solar_center_virtual_vert - lv) / (uv - lv)
        lh = self._grid_boundaries_horz[lower_bnd_center_horz_idx]
        uh = self._grid_boundaries_horz[lower_bnd_center_horz_idx + 1]
        self._altius_solar_center_horz = lower_bnd_center_horz_idx + (self._solar_center_virtual_horz - lh) / (uh - lh)

    def to_altius_resolution_image(self, fast_method=False):
        if fast_method is False:
            image = self._to_altius_resolution_image_slow()

        else:
            image = self._to_altius_resolution_image_fast()

        # image *= (self._horz_scale*self._vert_scale)
        # h, v = setup_solid_angle_indices(HORIZONTAL_HMI_PIXEL_INDEX,
        #                                  VERTICAL_HMI_PIXEL_INDEX,
        #                                  self._x_c,
        #                                  self._y_c,
        #                                  horizontal_pixel_shift,
        #                                  vertical_pixel_shift)
        #
        # return compute_delta_solid_angles(self.actual_solar_distance(),
        #                                   self._average_radius_pixels,
        #                                   h, v)
        # altius_solid_angles =
        # image /= altius_solid_angles

        image_geometry = {'altius_solar_center_horz': self._altius_solar_center_horz,
                          'altius_solar_center_vert': self._altius_solar_center_vert,
                          'altius_solar_radius': self._altius_actual_radius}
        # image_geometry['used_mask'] = used_mask

        return image, image_geometry

    def _to_altius_resolution_image_fast(self):
        """
        Utilizes the skimage modules to rescale the image
        """
        bv = self._boresight_virtual_px_vert * self._vert_scale
        bh = self._boresight_virtual_px_horz * self._horz_scale
        vi = self._vertical_virtual_pixel_centers * self._vert_scale - bv
        hi = self._horizontal_virtual_pixel_centers * self._horz_scale + bh

        v0 = int(np.round(vi[0]))
        v1 = int(np.round(vi[-1]))
        h0 = int(np.round(hi[0]))
        h1 = int(np.round(hi[-1]))

        resized_disk = rescale(self._current_hr_disk_radiance, (self._vert_scale, self._horz_scale))
        image = np.zeros((self._num_altius_pixels_vertical, self._num_altius_pixels_horizontal))
        image[v0:v1, h0:h1] = resized_disk

        scale = np.nansum(self._current_hr_disk_radiance) / np.nansum(image)

        num_per_pixel = 1 / (self._vert_scale * self._horz_scale)

        return image * scale / num_per_pixel

    def _to_altius_resolution_image_slow(self):
        """
        Converts the virtual pixels to actual pixels by summing the virtual pixel values
        into the coinciding actual ALTIUS detector pixels.  It is assumed that the pixel dimensions
        passed in map perfectly to the virtual detector space, not including the padding rows of
        raster scans.

        Parameters
        ----------

        Returns
        -------
        image : np.ndarray
            2D array of shape (num_pixels_horizontal, num_pixels_vertical) of the aggregated
            radiance values, in units of photons/cm^2/s

        image_geometry : dict
            raster_bounding_box : tuple
                (minimum_x, minimum_y, width_x, width_y) in pixels of the raster box, mostly for
                debugging/troubleshooting
            altius_solar_center_horz : float
                the expected center of the solar disk to a sub-pixel level on the ALTIUS image
            altius_solar_center_vert : float
                the expected center of the solar disk to a sub-pixel level on the ALTIUS image
            altius_solar_radius : float
                sub-pixel resolution of the solar disk radius on the ALTIUS image
        """

        image = np.zeros((self._num_altius_pixels_horizontal, self._num_altius_pixels_vertical))
        used_mask = np.zeros_like(image)
        debug_image = False

        disk_irradiance = self._current_hr_disk_radiance  # *delta_omega

        num_per_pixel = 1 / (self._vert_scale * self._horz_scale)

        # TODO the end column and row get missed, figure out a way to do this
        for gbv in range(len(self._grid_boundaries_vert) - 1):
            for gbh in range(len(self._grid_boundaries_horz) - 1):
                lv = self._grid_boundaries_vert[gbv]
                uv = self._grid_boundaries_vert[gbv + 1]
                lh = self._grid_boundaries_horz[gbh]
                uh = self._grid_boundaries_horz[gbh + 1]

                # get all pixels that are fully inside the ALTIUS pixel:
                # hli, hui, vli, vui = self._get_virtual_pixels_inside_altius_pixel(lh, lv, uh, uv)
                vi = np.intersect1d(np.argwhere(self._vertical_virtual_pixel_centers <= uv),
                                    np.argwhere(lv <= self._vertical_virtual_pixel_centers))
                hi = np.intersect1d(np.argwhere(self._horizontal_virtual_pixel_centers <= uh),
                                    np.argwhere(lh <= self._horizontal_virtual_pixel_centers))
                hli = hi[0]
                hui = hi[-1]
                vui = vi[-1]
                vli = vi[0]
                altius_pixel_radiance = np.sum(disk_irradiance[vli:vui, hli:hui])

                # now add in the partial pixels:

                delta_left = self._horizontal_virtual_pixel_centers[hli] - lh
                delta_right = uh - self._horizontal_virtual_pixel_centers[hui]
                delta_up = uv - self._vertical_virtual_pixel_centers[vui]
                delta_down = self._vertical_virtual_pixel_centers[vli] - lv

                # Add the sum of the partial rows, not counting the corners twice
                altius_pixel_radiance += delta_left * np.sum(disk_irradiance[vli:vui, hli - 1])
                altius_pixel_radiance += delta_right * np.sum(disk_irradiance[vli:vui, hui])
                altius_pixel_radiance += delta_up * np.sum(disk_irradiance[vui, hli:hui])
                altius_pixel_radiance += delta_down * np.sum(disk_irradiance[vli - 1, hli:hui])

                # now add in the little corners:
                altius_pixel_radiance += delta_left * delta_up * np.sum(disk_irradiance[vui, hli - 1])
                altius_pixel_radiance += delta_left * delta_down * np.sum(disk_irradiance[vli - 1, hli - 1])
                altius_pixel_radiance += delta_right * delta_up * np.sum(disk_irradiance[vui, hui])
                altius_pixel_radiance += delta_right * delta_down * np.sum(disk_irradiance[vli - 1, hui])

                # norm = len(vi) * len(hi) + (delta_left + delta_right) * len(vi) + (delta_up + delta_down) * len(hi) + \
                #        delta_left * delta_up + delta_left * delta_down + delta_right * delta_up + \
                #        delta_right * delta_down

                # altius_pixel_radiance /= norm

                # careful with the upper index for single row/column:
                if debug_image:
                    self._compute_used_mask(delta_down, delta_left, delta_right, delta_up,
                                            hli, hui, used_mask, vli, vui)

                iv = self._altius_vertical_idx[gbv]
                ih = self._altius_horizontal_idx[gbh]
                image[iv, ih] = altius_pixel_radiance

        image /= num_per_pixel

        return image

    def _compute_used_mask(self, delta_down, delta_left, delta_right, delta_up, hli, hui, used_mask, vli, vui):
        used_mask[vli:vui, hli:hui] += 1.0
        used_mask[vli:vui, hli - 1] += delta_left
        used_mask[vli:vui, hui] += delta_right
        used_mask[vui, hli:hui] += delta_up
        used_mask[vli - 1, hli:hui] += delta_down
        used_mask[vui, hli - 1] += delta_left * delta_up
        used_mask[vli - 1, hli - 1] += delta_left * delta_down
        used_mask[vui, hui] += delta_right * delta_down
        used_mask[vli - 1, hui] += delta_right * delta_up

    def _get_virtual_pixels_inside_altius_pixel(self, lh, lv, uh, uv):
        ydiffl = self._horizontal_virtual_pixel_centers - lh
        ydiffl[ydiffl < 0] = np.inf
        yli = ydiffl.argmin()
        ydiffu = uh - self._horizontal_virtual_pixel_centers
        ydiffu[ydiffu < 0] = np.inf
        yui = ydiffu.argmin()
        zdiffl = self._vertical_virtual_pixel_centers - lv
        zdiffl[zdiffl < 0] = np.inf
        zli = zdiffl.argmin()
        zdiffu = uv - self._vertical_virtual_pixel_centers
        zdiffu[zdiffu < 0] = np.inf
        zui = zdiffu.argmin()
        return yli, yui, zli, zui

    def get_solar_disk_detector_pixel_shift_from_boresight(self, scan_index):
        """
        Determines the horizontal and vertical shift in virtual pixels to place the solar image at
        the correct raster position on the detector, given by the raster index.

        Parameters
        ----------
        scan_index : int
           Index defining the position of the solar image on the raster grid.  0 corresponds to the bottom left of the
           detector.  The raster moves right across the detector, moves up one row, then moves left across the detector, etc.

        Returns
        -------
        horizontal_pixel_shift : float
            Number of virtual pixels to horizontally shift the solar HMI image by on the detector.

        vertical_pixel_shift : float
            Number of virtual pixels to vertically shift the solar HMI image by on the detector.
        """
        pos = self._solar_scan_grid.get_detector_location(self._solar_disk_radius_pixels, scan_index)

        vertical_pixel_shift = pos[0] - self._boresight_virtual_px_vert
        horizontal_pixel_shift = self._boresight_virtual_px_horz - pos[1]

        return horizontal_pixel_shift, vertical_pixel_shift

    def extract_satellite_orientation(self, horizontal_virtual_pixel_shift: float,
                                      vertical_virtual_pixel_shift: float,
                                      sat_to_solar_itrs: SkyCoord):
        """
        Determines a somewhat arbitrary orientation for a satellite in space to place the sun
        at the correct location on the detector.  It begins by placing the x-axis (boresight) along the
        satellite-to-sun ITRS vector, and constraining the z-axis to have z[0] = z[1] = 1, such that the
        z[2] component can be solved for by a dot-product between the x-axis and z-axis.  The y-axis is determined
        from the cross product.  From here, the horizontal and virtual pixel shifts are used to determine the azimuthal
        rotation matrix about the z-axis and elevation rotation matrix about the y-axis to place the sun at the
        necessary position on the detector.  These rotation matrices are applied to the x and z axis to give the
        final orientation of the satellite.

        Parameters
        ----------
        horizontal_virtual_pixel_shift : float
           Horizontal shift in virtual HMI pixels to place the sun at the correct azimuth on the detector

        vertical_virtual_pixel_shift : float
           Vertical shift in virtual HMI pixels to place the sun at the correct elevation on the detector

        sat_to_solar_itrs : np.ndarray
            Vector of the satellite to sun in ITRS coordinates

        Returns
        -------
        np.ndarray
           Six-element vector containing the x and z unit vectors in ITRS coordinates of the satellite
           orientation.
        """
        sat_to_sun_distance_km = np.sqrt(sat_to_solar_itrs.x.value ** 2 +
                                         sat_to_solar_itrs.y.value ** 2 +
                                         sat_to_solar_itrs.z.value ** 2)

        y_c = pixels_to_distance_km(horizontal_virtual_pixel_shift,
                                    self._solar_disk_radius_pixels)
        z_c = pixels_to_distance_km(vertical_virtual_pixel_shift,
                                    self._solar_disk_radius_pixels)

        theta = np.arctan(y_c / sat_to_sun_distance_km)
        colat = np.pi / 2 - theta
        phi = np.arctan(np.sqrt(sat_to_sun_distance_km ** 2 + y_c ** 2) / z_c ** 2)

        x_icf = np.array([sat_to_solar_itrs.x.value, sat_to_solar_itrs.y.value, sat_to_solar_itrs.z.value])
        x_icf /= np.sqrt(np.sum(x_icf ** 2))
        # set z_icf[0] = z_icf[1] = 1, then the dot product of x_icf and z_icf yields:
        zz = -(x_icf[0] + x_icf[1]) / x_icf[2]
        z_icf = np.array([1, 1, zz])
        z_icf /= np.sqrt(np.sum(z_icf ** 2))
        # rotate about this to get z_icf
        y_icf = np.cross(z_icf, x_icf)
        R_h = rotation_matrix(z_icf, -phi)
        R_v = rotation_matrix(y_icf, colat)
        x_icf = R_h @ R_v @ x_icf
        z_icf = R_h @ R_v @ z_icf
        return np.array([x_icf[0], x_icf[1], x_icf[2], z_icf[0], z_icf[1], z_icf[2]])

    def get_icf_corners(self, sat_to_sun_distance_km):
        """
        Configures the four vectors defining the corners of the detector FOV
        in ICF coordinates.

        Returns
        -------
        corners_icf : np.ndarray
           3 x 4 array of the four corner ICF unit vectors
        """

        coords = self._solar_scan_grid.get_corner_pixel_coords(self._solar_disk_radius_pixels)

        bl = coords['bottom_left']
        br = coords['bottom_right']
        tl = coords['top_left']
        tr = coords['top_right']

        x = sat_to_sun_distance_km
        r = self._solar_disk_radius_pixels

        bl = np.array([x, pixels_to_distance_km(self._boresight_virtual_px_horz - bl[0], r),
                       pixels_to_distance_km(bl[1] - self._boresight_virtual_px_vert, r)])
        br = np.array([x, pixels_to_distance_km(self._boresight_virtual_px_horz - br[0], r),
                       pixels_to_distance_km(br[1] - self._boresight_virtual_px_vert, r)])
        tl = np.array([x, pixels_to_distance_km(self._boresight_virtual_px_horz - tl[0], r),
                       pixels_to_distance_km(tl[1] - self._boresight_virtual_px_vert, r)])
        tr = np.array([x, pixels_to_distance_km(self._boresight_virtual_px_horz - tr[0], r),
                       pixels_to_distance_km(tr[1] - self._boresight_virtual_px_vert, r)])

        corners_icf = np.array([bl, br, tl, tr])
        corners_icf = corners_icf.T / np.sqrt(np.sum(corners_icf ** 2, axis=1))
        return corners_icf

    def compute_hi_res_image_at_scan_index(self, platform, raster_idx, rotation_matrix, sat_to_sun_distance_km,
                                           save_data,
                                           scenario, solar_disk_interpolator, utc_time,
                                           config: altius_config.AltiusL1ppInflightConfig):
        disk_centers = np.array(solar_disk_interpolator.solar_disk_center_pixels())
        print('Raster index ', raster_idx, ':    Computing image for ', utc_time)
        disk_radiances = solar_disk_interpolator.interpolate_disk(utc_time)
        img = SolarImage(scenario=scenario,
                         raster_index=raster_idx,
                         disk_center_pixels=disk_centers,
                         solar_disk_values=disk_radiances.radiance(),
                         satellite_position_itrs=platform.platform_ecef_positions[:, raster_idx],
                         time=utc_time,
                         satellite_to_sun_distance_km=sat_to_sun_distance_km,
                         icf_to_itrs_rotation_matrix=rotation_matrix,
                         vdpa=self,
                         config=config)
        if save_data:
            print('Saving hi-resolution image to netcdf')
            img.save_netcdf()
        return img


# TODO will mostly do away with this interface and move it over to SpsHighResolutionSceneInterface
#  where multiple images at a specific limb-darkened wavelength can be saved to netcdf

# noinspection PyUnresolvedReferences
class SolarImage:
    """
    <Description>
    Class to save the image and all auxillary data for each snapshot image of the sun at a specific raster scan position.
    Built to eventually interface with what is needed for QinetiQ SPS (appropriate frame TBD).
    Look vectors are not stored on file, as the 4096x4096 radiance image occupies about 50MB on disk, and adding another
    3x4096x4096 array for the vectors would add another 150MB, which is too high for when ~100 separate images are needed
    for the complete raster.

    The class provides read/write capability to the netcdf file, and can compute the ITRS look vectors when needed for
    computation.

    Parameters
    ----------
    scenario : SolarDataScenario
       Scenario defining the dataset with a certain amount of sunspots present.

    raster_index : int
        Index describing the position of the sun on the detector in the raster scan.

    disk_center_pixels : np.ndarray
        Two-element array of the centroid of the solar disk on the (typically) 4096 x 4096 HMI pixel grid.

    solar_disk_values : np.ndarray
        Two-dimensional array of the disk radiance values in DN (arbitrary)

    satellite_position_itrs : np.ndarray
        ITRS position of the satellite at the time of measurement.

    satellite_to_sun_distance_km : float
        Distance from the satellite to the sun in km at the time of measurement.

    time : datetime
        Time of the simulated measurement

    icf_to_itrs_rotation_matrix : RotationMatrix
        Rotation matrix defining the transformation from ICF to ITRS at this time of measurement and satellite
        orientation.  Typically determined by the :code:`Platform` class.
    """

    def __init__(self,
                 scenario: SolarDataScenario,
                 raster_index: int,
                 disk_center_pixels: np.ndarray,
                 solar_disk_values: np.ndarray,
                 satellite_position_itrs: np.ndarray,
                 satellite_to_sun_distance_km: float,
                 time: datetime,
                 icf_to_itrs_rotation_matrix: RotationMatrix,
                 vdpa: VirtualDetectorPixelArray,
                 config: altius_config.AltiusL1ppInflightConfig):

        self._vdpa = vdpa
        solar_scan_grid = vdpa._solar_scan_grid
        disk_radius_pixels = vdpa._solar_disk_radius_pixels
        self._scenario = scenario
        self._dataset = xr.Dataset(
            data_vars={
                'disk_radiance': (['xp', 'yp'], solar_disk_values),
                'scenario': scenario.value,
                'raster_index': raster_index,
                'solar_disk_pixel_center': (['pixel_grid'], disk_center_pixels),
                'satellite_position_itrs': (['itrs_xyz'], satellite_position_itrs),
                'satellite_to_sun_distance_km': satellite_to_sun_distance_km,
                'time': time,
                'icf_to_itrs_rotation_matrix': (['itrs_xyz', 'itrs_xyz'], icf_to_itrs_rotation_matrix.R),
                'itrs_to_icf_rotation_matrix': (['itrs_xyz', 'itrs_xyz'], icf_to_itrs_rotation_matrix.RInv),
                'disk_radius_apparent_angle': solar_scan_grid.solar_disc_radius_angle(),
                'disk_radius_pixels': disk_radius_pixels,
                'scan_rate_deg_per_sec': solar_scan_grid.scan_rate_degrees_per_sec(),
                'integration_time_sec': solar_scan_grid.integration_time_sec(),
                'fov_horizontal_deg': solar_scan_grid.fov_horz_deg(),
                'fov_vertical_deg': solar_scan_grid.fov_vert_deg(),
                'boresight_horizontal_portion': vdpa._boresight_horizontal_portion,
                'boresight_vertical_portion': vdpa._boresight_vertical_portion,
                'num_vertical_scans': solar_scan_grid.get_num_grid_positions(),
                'num_horizontal_scans': solar_scan_grid.get_num_grid_positions()
            }
        )
        self._directory = config.get_processed_solar_prnu_file_directory()

    def get_disk_radiance(self):
        return self._dataset['disk_radiance'].values

    def get_disk_centers(self):
        return self._dataset['solar_disk_pixel_center'].values

    def get_disk_radius_pixels(self):
        return self._dataset['disk_radius_pixels'].values

    # TODO more getters

    def _get_full_file_path(self):
        nv = self._vdpa._solar_scan_grid.num_scans_vertical_infield()
        nh = self._vdpa._solar_scan_grid.num_scans_horizontal_infield()
        grid_config = self._grid_config(nh, nv)
        return SolarImage._get_full_netcdf_path(self._dataset['raster_index'].values, self._directory,
                                                self._scenario, grid_config)

    @staticmethod
    def _grid_config(nh, nv):
        grid_config = str(nv) + 'x' + str(nh)
        return grid_config

    def hmi_disk_radiance(self):
        return self._dataset['disk_radiance'].values

    def make_altius_observation(self, num_pixels_horizontal, num_pixels_vertical,
                                wavelength, srf: SrfInterface, fast_scaling=False):
        """
        Converts the virtual pixels to actual pixels by summing the virtual pixel values
        into the coinciding actual ALTIUS detector pixels.  It is assumed that the pixel dimensions
        passed in map perfectly to the virtual detector space, not including the padding rows of
        raster scans.

        This function corrects for:

            * viewing angle: this is encapsulated in the horizontal_pixel_shift and vertical_pixel_shift
              parameters - this helps to determine the correct differential solid angles of pixel to perform the
              correct integration across the disk
            * limb darkening: employs the Pierce & Slaughter limb darkening model to remove the limb darkening
              at the current wavelength and apply the limb darkening at the requested wavelength
            * SRF: computes the total irradiance that would be observed from the solar disk using the SAO2010
              solar spectrum and with any spectral response function at the requested wavelength

        Parameters
        ----------
        num_pixels_horizontal : int
            Number of pixels across the horizontal dimension of the ALTIUS detector

        num_pixels_vertical : int
            Number of pixels across the vertical dimension of the ALTIUS detector

        srf : SrfInterface

        wavelength : float

        Returns
        -------
        image : np.ndarray
            2D array of shape (num_pixels_horizontal, num_pixels_vertical) of the aggregated
            radiance values, in units of photons/s/sr/cm^2/nm

        image_geometry : dict
            raster_bounding_box : tuple
                (minimum_x, minimum_y, width_x, width_y) in pixels of the raster box, mostly for
                debugging/troubleshooting
            altius_solar_center_horz : float
                the expected center of the solar disk to a sub-pixel level on the ALTIUS image
            altius_solar_center_vert : float
                the expected center of the solar disk to a sub-pixel level on the ALTIUS image
            altius_solar_radius : float
                sub-pixel resolution of the solar disk radius on the ALTIUS image
        """
        raster_idx = self.get_raster_index()
        uncorrected_disk = self.get_solar_disk_radiance_result()

        horz_virtual_shift, vert_virtual_shift = \
            self._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(raster_idx)

        corrected_result = uncorrected_disk.to_scaled_image_result(wavelength=wavelength,
                                                                   srf=srf,
                                                                   horizontal_pixel_shift=horz_virtual_shift,
                                                                   vertical_pixel_shift=vert_virtual_shift)

        self._vdpa.set_current_scanned_image(corrected_result, num_pixels_vertical,
                                             num_pixels_horizontal, raster_idx)
        return self._vdpa.to_altius_resolution_image(fast_scaling)

    def get_solar_disk_radiance_result(self) -> SolarDiskRadianceResult:
        disk_radiance = self.get_disk_radiance()
        vsize, hsize = np.shape(disk_radiance)
        center_pix = self.get_disk_centers()
        horz_centroid_pixels = center_pix[0]
        vert_centroid_pixels = center_pix[1]
        radius_pixels = self.get_disk_radius_pixels()
        solar_distance_km = self.get_satellite_to_sun_distance_km()
        time = self.get_time()
        apparent_disk_angle = self.get_disk_radius_apparent_angle()
        hmi_wavelength = 617.3

        solar_geo = SolarGeometry(hsize, vsize, vert_centroid_pixels, horz_centroid_pixels, 0.5,
                                  0.5)  # FOV doesn't matter here
        solar_geo.set_disk(radius_pixels, horz_centroid_pixels, vert_centroid_pixels, solar_distance_km)
        mu = solar_geo.mu()

        base_radiance_result = SolarDiskRadianceResult(radiance_disk=disk_radiance,
                                                       vertical_disk_pixel_centroid=vert_centroid_pixels,
                                                       horizontal_disk_pixel_centroid=horz_centroid_pixels,
                                                       average_radius_pixels=radius_pixels,
                                                       actual_solar_distance_km=solar_distance_km,
                                                       mu=mu, time=time, wavelength=hmi_wavelength,
                                                       solar_radius_apparent_angle_deg=apparent_disk_angle)
        return base_radiance_result

    def visualize_detector_image(self):
        """
        More for debugging purposes, to visualize that the sun is in the correct position.
        """
        radiance = self._dataset['disk_radiance'].values[::10, ::10]
        look_and_orientation_vectors = self.compute_solar_image_itrs_look_vectors()
        los_itrs = look_and_orientation_vectors['itrs_solar_disk_look_vectors']
        look_vectors = los_itrs[:, ::10, ::10]

        horz = look_and_orientation_vectors['detector_horz']
        vert = look_and_orientation_vectors['detector_vert']
        corner_look_vectors = look_and_orientation_vectors['corners_itrs']

        projected_coords_x = np.dot(horz, np.diagonal(look_vectors, axis1=1, axis2=2))
        projected_coords_y = np.dot(vert, np.diagonal(look_vectors, axis1=1, axis2=2))

        projected_corners_x = [np.dot(horz, cc) for cc in np.reshape(corner_look_vectors, (3, 4)).T]
        projected_corners_y = [np.dot(vert, cc) for cc in np.reshape(corner_look_vectors, (3, 4)).T]

        central_bounds = np.reshape(look_and_orientation_vectors['corner_raster_grid_bounds'], (3, 4)).T

        projected_center_bounds_x = [np.dot(horz, cb) for cb in central_bounds]
        projected_center_bounds_y = [np.dot(vert, cb) for cb in central_bounds]

        dx = np.mean(projected_coords_x[1:] - projected_coords_x[:-1])
        dy = np.mean(projected_coords_y[1:] - projected_coords_y[:-1])

        x = np.arange(np.min(projected_corners_x), np.max(projected_corners_x), dx)
        y = np.arange(np.min(projected_corners_y), np.max(projected_corners_y), dy)
        f = interp2d(projected_coords_x, projected_coords_y, np.nan_to_num(radiance), fill_value=0.0)

        xb = np.abs(np.flip(x) - np.max(projected_center_bounds_x)).argmin()
        wxb = np.abs(np.abs(np.flip(x) - np.min(projected_center_bounds_x)).argmin() -
                     np.abs(np.flip(x) - np.max(projected_center_bounds_x)).argmin())

        yb = np.abs(np.flip(y) - np.min(projected_center_bounds_y)).argmin()
        wyb = np.abs(np.abs(y - np.min(projected_center_bounds_y)).argmin() -
                     np.abs(y - np.max(projected_center_bounds_y)).argmin())

        image = f(x, y)

        fig, ax = plt.subplots(1)
        # flipped upside-down because imshow indexes elements with (0, 0) at the top left
        # flipped left-right because y-axis is positive to the left in detector coordinates
        im = ax.imshow(np.fliplr(np.flipud(image)))
        plt.colorbar(im)
        rect = pch.Rectangle((xb, yb), wxb, -wyb, edgecolor='r', facecolor='none')
        ax.add_patch(rect)

        plt.figure()
        plt.plot(np.arange(0, len(radiance)), radiance[:, int(len(radiance[:, 0]) / 2)])
        plt.show()

    def save_netcdf(self):
        """
        Saves the :code:`xarray.Dataset` member using the to_netcdf interface.  Uses compression to
        6 significant digits.  Yields a file about 56 MB.

        """
        dirname = os.path.dirname(self._get_full_file_path())
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        comp = dict(zlib=True, complevel=9, least_significant_digit=6)
        encoding = {var: comp for var in self._dataset.data_vars}
        with open(self._get_full_file_path(), 'wb') as f:
            # have to pass in the string filename!!
            self._dataset.to_netcdf(path=f.name, format='NETCDF4', engine='netcdf4', encoding=encoding)

    @classmethod
    def from_netcdf(cls, scenario: SolarDataScenario, num_vertical_scans: int,
                    num_horizontal_scans: int, raster_index: int):
        config = altius_config.AltiusL1ppInflightConfig()
        home_dir = config.get_processed_solar_prnu_file_directory()  # os.environ['SOLAR_PROCESSED_SCENE_DIR']

        grid_config = SolarImage._grid_config(num_horizontal_scans, num_vertical_scans)

        filepath = cls._get_full_netcdf_path(raster_index, home_dir, scenario, grid_config)

        if not os.path.exists(filepath):
            raise IOError('No file ' + str(filepath))

        with open(filepath, 'rb') as f:
            ds = xr.open_dataset(f.name, decode_cf=True, engine='netcdf4')

        # TODO this is a little ugly, but didn't feel like giving these classes their own _to_netcdf interface
        solar_scan_grid = RasterScanGrid(solar_radius_angular_size_deg=ds['disk_radius_apparent_angle'].values,
                                         scan_rate_deg_per_sec=ds['scan_rate_deg_per_sec'].values,
                                         time_per_image_sec=ds['integration_time_sec'].values,
                                         fov_horizontal_deg=ds['fov_horizontal_deg'].values,
                                         fov_vertical_deg=ds['fov_vertical_deg'].values,
                                         v_in=ds['num_vertical_scans'].values,
                                         h_in=ds['num_horizontal_scans'].values)

        vdpa = VirtualDetectorPixelArray(disk_radius_pixels=ds['disk_radius_pixels'].values,
                                         solar_scan_grid=solar_scan_grid,
                                         boresight_horz_portion=ds['boresight_horizontal_portion'].values,
                                         boresight_vert_portion=ds['boresight_vertical_portion'].values)

        return cls(scenario=scenario, raster_index=raster_index,
                   disk_center_pixels=ds['solar_disk_pixel_center'].values,
                   solar_disk_values=ds['disk_radiance'].values,
                   satellite_position_itrs=ds['satellite_position_itrs'].values,
                   satellite_to_sun_distance_km=ds['satellite_to_sun_distance_km'].values,
                   time=ds['time'].values,
                   icf_to_itrs_rotation_matrix=RotationMatrix(ds['icf_to_itrs_rotation_matrix'].values),
                   vdpa=vdpa)

    @classmethod
    def _get_full_netcdf_path(cls, grid_index, home_dir, scenario, grid_configuration):
        return os.path.join(home_dir, scenario.to_string_directory(), grid_configuration,
                            'solar_prnu_' + scenario.to_string_directory() + '_' +
                            str(grid_index) + '.nc')

    def _compute_icf_raster_grid_corners(self):

        solar_pixel_radius = self._dataset['disk_radius_pixels'].values
        horz_rel_to_boresight, vert_rel_to_boresight = self._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(
            self._dataset['raster_index'].values)

        grid_vert_step_pixels = self._vdpa._solar_scan_grid.grid_vertical_step_in_pixels(solar_pixel_radius)
        grid_horz_step_pixels = self._vdpa._solar_scan_grid.grid_horizontal_step_in_pixels(solar_pixel_radius)

        y_left = horz_rel_to_boresight - grid_horz_step_pixels / 2
        y_right = horz_rel_to_boresight + grid_horz_step_pixels / 2
        z_up = vert_rel_to_boresight + grid_vert_step_pixels / 2
        z_down = vert_rel_to_boresight - grid_vert_step_pixels / 2

        y = np.array([y_left, y_right, y_left, y_right])
        z = np.array([z_down, z_down, z_up, z_up])

        z = pixels_to_distance_km(z, solar_pixel_radius)
        y = pixels_to_distance_km(y, solar_pixel_radius)

        xi = self._dataset['satellite_to_sun_distance_km'].values
        x = xi * np.ones_like(z)

        raster_corners_icf = np.stack((x, y, z), axis=0)
        raster_corners_icf /= np.sqrt(np.sum(raster_corners_icf ** 2, axis=0))

        return raster_corners_icf

    def compute_icf_solar_center(self):
        horz_shift_rel_to_boresight, vert_shift_rel_to_boresight = self._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(
            self._dataset['raster_index'].values)

        xi = self._dataset['satellite_to_sun_distance_km'].values
        center_pix = self._dataset['solar_disk_pixel_center'].values
        y_centroid_pixels = center_pix[0]
        z_centroid_pixels = center_pix[1]

        z_virtual_pixels = vert_shift_rel_to_boresight + z_centroid_pixels  # array of pixels
        y_virtual_pixels = horz_shift_rel_to_boresight + y_centroid_pixels

        z = pixels_to_distance_km(z_virtual_pixels, self._dataset['disk_radius_pixels'].values)
        y = pixels_to_distance_km(y_virtual_pixels, self._dataset['disk_radius_pixels'].values)

        z = z.flatten()
        y = y.flatten()
        x = xi * np.ones_like(z)
        los_icf = np.stack((x, y, z), axis=0)
        los_icf /= np.sqrt(np.sum(los_icf ** 2, axis=0))

        return los_icf

    def compute_icf_los(self):
        """
        Takes the translated virtual detector grid of the HMI pixels and converts them to ICF look vectors

        Returns
        -------
        los_icf : np.ndarray
           3 x NLOS array where NLOS = size_horz x size_vert.  Line of sight in ICF to each HMI pixel
        """
        horz_shift_rel_to_boresight, vert_shift_rel_to_boresight = self._vdpa.get_solar_disk_detector_pixel_shift_from_boresight(
            self._dataset['raster_index'].values)

        xi = self._dataset['satellite_to_sun_distance_km'].values
        center_pix = self._dataset['solar_disk_pixel_center'].values
        y_centroid_pixels = center_pix[0]
        z_centroid_pixels = center_pix[1]

        h_size, v_size = self._dataset['disk_radiance'].values.shape
        solar_geometry = SolarGeometry(h_size, v_size, z_centroid_pixels, y_centroid_pixels)
        DY, DZ = solar_geometry.get_delta_pixels_relative_to_boresight()

        z_virtual_pixels = vert_shift_rel_to_boresight + DZ  # array of pixels
        y_virtual_pixels = horz_shift_rel_to_boresight + DY

        z = pixels_to_distance_km(z_virtual_pixels, self._dataset['disk_radius_pixels'].values)
        y = pixels_to_distance_km(y_virtual_pixels, self._dataset['disk_radius_pixels'].values)

        z = z.flatten()
        y = y.flatten()
        x = xi * np.ones_like(z)
        los_icf = np.stack((x, y, z), axis=0)
        los_icf /= np.sqrt(np.sum(los_icf ** 2, axis=0))

        return los_icf

    def compute_solar_image_itrs_look_vectors(self):
        """
        Computes all the necessary look and orientation vectors in ITRS frame.

        Returns
        -------
        los_itrs : dictionary
            boresight : np.ndarray
                ITRS vectors describing the boresight

            corners_itrs : np.ndarray
                3 x 2 x 2 array describing the ITRS vectors of each corner of the detector FOV

            detector_vert : np.ndarray
                ITRS vector for the vertical direction of the detector plane

            detector_horz : np.ndarray
                ITRS vector for the horizontal direction of the detector plane

            corner_raster_grid_bounds : np.ndarray
                ITRS vector for the corner bounds of the usable area for the raster grid, defined as the square
                with sides the same length as the solar radius centered inside the solar disk

            itrs_solar_disk_look_vectors : np.ndarray
                3 x 4096 x 4096 array describing the ITRS look vectors of each
        """
        los_itrs = {}
        # get the corner vectors of the detector window:

        corner_bound_look_vectors = self._vdpa.get_icf_corners(self._dataset['satellite_to_sun_distance_km'].values)
        R = self._dataset['icf_to_itrs_rotation_matrix'].values
        corners_itrs = R @ corner_bound_look_vectors
        los_itrs['corners_itrs'] = np.reshape(corners_itrs, (3, 2, 2))

        los_itrs['boresight'] = R @ np.array([1, 0, 0])
        los_itrs['detector_vert'] = R @ np.array([0, 0, 1])
        los_itrs['detector_horz'] = R @ np.array([0, 1, 0])

        los_disk_icf = self.compute_icf_los()
        itrs_solar_disk_look_vectors = R @ los_disk_icf
        size_horz, size_vert = np.shape(self._dataset['disk_radiance'])
        los_itrs['itrs_solar_disk_look_vectors'] = np.reshape(np.ravel(itrs_solar_disk_look_vectors),
                                                              (3, size_horz, size_vert))

        los_itrs['corner_raster_grid_bounds'] = R @ self._compute_icf_raster_grid_corners()

        return los_itrs

    def get_raster_index(self):
        return self._dataset['raster_index'].values

    def get_time(self):
        return self._dataset['time'].values

    def get_disk_radius_apparent_angle(self):
        return self._dataset['disk_radius_apparent_angle'].values

    def get_satellite_to_sun_distance_km(self):
        return self._dataset['satellite_to_sun_distance_km'].values
