from typing import Tuple

import netCDF4 as nc
import numpy as np
import xarray as xr
from sktime import mjd as mjd_conv

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.altius_scene_interfaces import AltiusLunarScene


class SpsToUsaskLunarPrnuSceneInterface:
    """
    Class representing the output of the SPS measurements of the moon for the PRNU lunar
    calibration.  It contains a set of images of the moon at a handful of locations on the
    detector for a single wavelength, in netcdf file format.

    In truth, this class may only represent the output from the SPS given to the calibration
    processes, in which case

    Main fields:

    - Images - units of DN
        - dim 0: image index
        - dim 1: image vertical dimension
        - dim 2: image horizontal dimension
    - detector resolution and size
    - center wavelength
    - detector pixel solid angles
    - detector boresight
    - other data tbd
    - lunar disk parameters:
        - lunar disk center, radius
        - etc.
    """

    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    XYZ = 'xyz'
    DETECTOR_POS_DIM = 'detector_positions_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    SCENES = 'scenes'
    TIMES = 'times'
    EXPECTED_LUNAR_CENTER = 'expected_lunar_center'
    EXPECTED_LUNAR_RADIUS = 'expected_lunar_radius'
    SAT_TO_MOON_DISTANCE = 'satellite_to_moon_distance'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths_nm'
    START_SNAPSHOT_BORESIGHT_POINTING = 'start_snapshot_pointing'
    END_SNAPSHOT_BORESIGHT_POINTING = 'end_snapshot_pointing'
    INTEGRATION_TIMES = 'integration_times'
    SUBSOLAR_SELEN_LONGITUDE = 'subsolar_selenographic_longitude'
    SUBSAT_SELEN_LONGITUDE = 'subsatellite_selenographic_longitude'
    SUBSAT_SELEN_LATITUDE = 'subsatellite_selenographic_latitude'
    LUNAR_PHASE_ANGLE_DEG = 'lunar_phase_angle'
    SUN_TO_MOON_DISTANCE_AU = 'sun_to_moon_distance_au'
    SATELLITE_POSITION_ECEF = 'satellite_position_ecef'
    DETECTOR_POSITIONS = 'detector_lunar_pixel_positions'

    def __init__(self):
        self._read_nc_file = None
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            calibration_wavelengths: np.ndarray,
                            detector_fov: Tuple[float, float],
                            detector_size_pixels: Tuple[int, int],
                            detector_boresight_pixels: Tuple[float, float],
                            grid_pixel_positions: np.ndarray):
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, len(calibration_wavelengths))
        self._boresight_vector_dim = self._write_nc_file.createDimension(self.XYZ, 3)
        self._exposure_detector_positions_dim = self._write_nc_file.createDimension(self.DETECTOR_POS_DIM,
                                                                                    len(grid_pixel_positions))

        # Variables:
        self._satellite_position = self._write_nc_file.createVariable(self.SATELLITE_POSITION_ECEF, float,
                                                                      (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                       self.XYZ))

        self._exposure_detector_positions = self._write_nc_file.createVariable(self.DETECTOR_POSITIONS, int,
                                                                               (self.DETECTOR_POS_DIM,
                                                                                self.PIXEL_LOCATION_SIZE))

        self._exposure_detector_positions[:] = grid_pixel_positions
        self._exposure_detector_positions.description = 'Approximate center pixel locations of the snapshots ' \
                                                        '- limited to a small set across the detector'
        self._exposure_detector_positions.units = 'Vertical pixel number, horizontal pixel number'

        self._snapshot_pointing_start = self._write_nc_file.createVariable(self.START_SNAPSHOT_BORESIGHT_POINTING,
                                                                           float,
                                                                           (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                            self.XYZ))
        self._snapshot_pointing_start.units = 'Arbitrary'
        self._snapshot_pointing_start.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                    'boresight at the beginning of each integration period'

        self._snapshot_pointing_end = self._write_nc_file.createVariable(self.END_SNAPSHOT_BORESIGHT_POINTING,
                                                                         float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                          self.XYZ))
        self._snapshot_pointing_end.units = 'Arbitrary'
        self._snapshot_pointing_end.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                  'boresight at the end of each integration period'

        self._integration_times = self._write_nc_file.createVariable(self.INTEGRATION_TIMES, float,
                                                                     (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._integration_times.units = 'seconds'
        self._integration_times.description = 'Exposure time of each snapshot of the sun'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.PIXEL_LOCATION_SIZE,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight_pixels[0], detector_boresight_pixels[1]]

        self._calibration_wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                                           (self.CAL_WL_DIM,))
        self._calibration_wavelengths[:] = calibration_wavelengths
        self._calibration_wavelengths.units = 'nm'
        self._calibration_wavelengths.description = 'Wavelength in nm for which the set of images is intended ' \
                                                    'to calibrate'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_SIZE,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._scenes = self._write_nc_file.createVariable(self.SCENES, float,
                                                          (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                           self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self._scenes.units = 'DN'
        self._scenes.description = 'Scene at detector resolution'

        self._times = self._write_nc_file.createVariable(self.TIMES, float,
                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._times.units = 'MJD'

        self._expected_lunar_center = self._write_nc_file.createVariable(self.EXPECTED_LUNAR_CENTER, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,
                                                                          self.PIXEL_LOCATION_SIZE))
        self._expected_lunar_center.description = 'Center of the lunar disk in vertical and horizontal pixels, ' \
                                                  'respectively'
        self._expected_lunar_center.units = 'Pixels'

        self._expected_lunar_radius = self._write_nc_file.createVariable(self.EXPECTED_LUNAR_RADIUS, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._expected_lunar_radius.units = 'Pixels'

        self._satellite_to_moon_distance = self._write_nc_file.createVariable(self.SAT_TO_MOON_DISTANCE, float,
                                                                              (self.DETECTOR_POS_DIM,
                                                                               self.CAL_WL_DIM,))
        self._satellite_to_moon_distance.units = 'km'

        # SUBSOLAR_SELEN_LONGITUDE = 'subsolar_selenographic_longitude'
        # SUBSAT_SELEN_LONGITUDE = 'subsatellite_selenographic_longitude'
        # SUBSAT_SELEN_LATITUDE = 'subsatellite_selenographic_latitude'
        # LUNAR_PHASE_ANGLE_DEG = 'lunar_phase_angle'
        # SUN_TO_MOON_DISTANCE_AU = 'sun_to_moon_distance_au'

        self._subsolar_selen_longitude = self._write_nc_file.createVariable(self.SUBSOLAR_SELEN_LONGITUDE, float,
                                                                            (self.DETECTOR_POS_DIM,
                                                                             self.CAL_WL_DIM,))
        self._subsolar_selen_longitude.units = 'Degrees'
        self._subsolar_selen_longitude.description = 'Selenographic longitude of the sun projected on to the moon'

        self._subsat_selen_latitude = self._write_nc_file.createVariable(self.SUBSAT_SELEN_LATITUDE, float,
                                                                         (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._subsat_selen_latitude.units = 'Degrees'
        self._subsat_selen_latitude.description = 'Selenographic latitude of the satellite projected onto the moon'

        self._subsat_selen_longitude = self._write_nc_file.createVariable(self.SUBSAT_SELEN_LONGITUDE, float,
                                                                          (self.DETECTOR_POS_DIM,
                                                                           self.CAL_WL_DIM,))
        self._subsat_selen_longitude.units = 'Degrees'
        self._subsat_selen_longitude.description = 'Selenographic longitude of the satellite projected onto the moon'

        self._lunar_phase_angle = self._write_nc_file.createVariable(self.LUNAR_PHASE_ANGLE_DEG, float,
                                                                     (self.DETECTOR_POS_DIM, self.CAL_WL_DIM,))
        self._lunar_phase_angle.units = 'Degrees'
        self._lunar_phase_angle.description = 'Phase angle in degrees of the moon'

        self._sun_to_moon_distance_au = self._write_nc_file.createVariable(self.SUN_TO_MOON_DISTANCE_AU, float,
                                                                           (self.DETECTOR_POS_DIM,
                                                                            self.CAL_WL_DIM,))
        self._sun_to_moon_distance_au.units = 'Astronomical Units'

    def get_channel(self, config: altius_config.AltiusL1ppInflightConfig):
        vals = self._read_nc_file.variables
        wl = vals[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS][0].values
        return config.get_channel(wl)

    def write_next_lunar_scene_to_netcdf(self,
                                         time: np.datetime64,
                                         scene: np.ndarray,
                                         wavelength: float,
                                         satellite_position: np.ndarray,
                                         expected_lunar_center: Tuple[float, float],
                                         expected_lunar_radius: float,
                                         sat_moon_distance_km: float,
                                         start_boresight_pointing: np.ndarray,
                                         end_boresight_pointing: np.ndarray,
                                         integration_time: float,
                                         subsolar_selen_longitude,
                                         subsat_selen_latitude,
                                         subsat_selen_longitude,
                                         sun_to_moon_distance_au,
                                         lunar_phase_angle_deg):
        # Determine the closest detector positions and wavelengths:
        wl_idx = np.argmin(np.abs(wavelength - self._calibration_wavelengths))
        pos_idx = np.argmin(
            np.sqrt(np.sum((np.asarray(expected_lunar_center) - self._exposure_detector_positions) ** 2, axis=1)))

        self._satellite_position[pos_idx, wl_idx, :] = satellite_position
        self._times[pos_idx, wl_idx] = mjd_conv.datetime64_to_mjd(time)
        self._scenes[pos_idx, wl_idx, :, :] = scene
        self._expected_lunar_center[pos_idx, wl_idx, :] = [expected_lunar_center[0], expected_lunar_center[1]]
        self._expected_lunar_radius[pos_idx, wl_idx] = expected_lunar_radius
        self._satellite_to_moon_distance[pos_idx, wl_idx] = sat_moon_distance_km
        self._snapshot_pointing_start[pos_idx, wl_idx, :] = start_boresight_pointing
        self._snapshot_pointing_end[pos_idx, wl_idx, :] = end_boresight_pointing
        self._integration_times[pos_idx, wl_idx] = integration_time
        self._subsolar_selen_longitude[pos_idx, wl_idx] = subsolar_selen_longitude
        self._subsat_selen_latitude[pos_idx, wl_idx] = subsat_selen_latitude
        self._subsat_selen_longitude[pos_idx, wl_idx] = subsat_selen_longitude
        self._sun_to_moon_distance_au[pos_idx, wl_idx] = sun_to_moon_distance_au
        self._lunar_phase_angle[pos_idx, wl_idx] = lunar_phase_angle_deg

    def close_write_nc(self):
        self._write_nc_file.close()

    def close_read_nc(self):
        if self._read_nc_file is not None:
            self._read_nc_file.close()

    def open_read_dataset(self, filename):
        self._read_nc_file = xr.open_dataset(filename)  # , chunks={SpsToUsaskSceneInterface.TIME_DIM: 4})

    def get_num_scenes(self):
        return self._read_nc_file.dims[SpsToUsaskLunarPrnuSceneInterface.DETECTOR_POS_DIM]

    def get_num_pixels_scene(self):
        num_vert = self._read_nc_file.dims[SpsToUsaskLunarPrnuSceneInterface.NUM_VERT_PIXELS]
        num_horz = self._read_nc_file.dims[SpsToUsaskLunarPrnuSceneInterface.NUM_HORZ_PIXELS]
        return num_vert, num_horz

    def get_wavelengths(self):
        return self._read_nc_file[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS].values

    def load_scene(self, wavelength, detector_position_index) -> AltiusLunarScene:
        vals = self._read_nc_file.variables
        k = detector_position_index

        wl_idx = np.argmin(np.abs(wavelength - vals[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS].values))

        expected_center = vals[SpsToUsaskLunarPrnuSceneInterface.EXPECTED_LUNAR_CENTER][k, wl_idx].values
        return AltiusLunarScene(vals[SpsToUsaskLunarPrnuSceneInterface.SCENES][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.DETECTOR_FOV_DEG].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.SAT_TO_MOON_DISTANCE][k, wl_idx].values,
                                expected_center[1],
                                expected_center[0],
                                vals[SpsToUsaskLunarPrnuSceneInterface.EXPECTED_LUNAR_RADIUS][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS][wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.DETECTOR_BORESIGHT].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.START_SNAPSHOT_BORESIGHT_POINTING][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.END_SNAPSHOT_BORESIGHT_POINTING][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.INTEGRATION_TIMES][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.SUN_TO_MOON_DISTANCE_AU][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.LUNAR_PHASE_ANGLE_DEG][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.SUBSAT_SELEN_LATITUDE][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.SUBSAT_SELEN_LONGITUDE][k, wl_idx].values,
                                vals[SpsToUsaskLunarPrnuSceneInterface.SUBSOLAR_SELEN_LONGITUDE][k, wl_idx].values)


class LunarPrnuSpotCheckOutput:
    """
        This class describes the interface to be provided to the QinetiQ SPS as a result of the
        output of the spot check PRNU calibration using the moon as a source.  This output contains
        the calibration result for all chosen calibration wavelengths in a channel (uv, vis, nir).
        The wavelengths are set as the discrete set of wavelengths where the ROLO model parameters
        are well known.

        For now, this consists of the following netcdf structure.

        Dimensions:

        - NUM_VERT_PIXELS
            - value: 'num_vert_pixels'
            - fixed number of vertical pixels on the detector
        - NUM_HORZ_PIXELS
            - value: 'num_horz_pixels'
            - fixed number of horizontal pixels on the detector
        - WAVELENGTH_DIM
            - value: 'wavelength_dimension'
            - Number of wavelengths corresponding to the exact ROLO database wavelengths for this channel
        - SCENE_DIM
            - value: 'scene_dimension'
            - Index corresponding to one of the handful of moon center positions on the detector for each spot check
        - PIXEL_LOCATION_SIZE
            - value: 'pixel_location_size'
            - (vertical, horizontal) pixel position on the detector

        Variables:

        - CALIBRATION_WAVELENGTHS
            - value: 'calibration_wavelengths'
            - size: [WAVELENGTH_DIM,]
            - wavelength, in nm, of each PRNU calibration
        - RATIO
            - value: 'lunar_scene_to_rolo_irradiance_ratio'
            - size: [WAVELENGTH_DIM, SCENE_DIM]
            - The resulting ratio of the measured irradiance compared to the expected irradiance from the ROLO model
        - RATIO_ERROR
            - value: 'lunar_scene_to_rolo_irradiance_ratio_error'
            - size: [WAVELENGTH_DIM, SCENE_DIM]
            - The error in the resulting ratio of the measured irradiance compared to the expected irradiance
              from the ROLO model

        - PIXEL_MASK
            - value: 'pixel_mask'
            - size: [WAVELENGTH_DIM, SCENE_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
            - Mask of the detector for each measurement where the moon illuminated the pixels.

        - EXPOSURE_CENTER_PIXEL_POSITIONS
            - value: 'exposure_center_pixel_positions'
            - size: [SCENE_DIM, PIXEL_LOCATION_SIZE]
            - Center pixel positions of where the moon is expected to be for each measurement
        """

    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    WAVELENGTH_DIM = 'wavelength_dimension'
    SCENE_DIM = 'scene_dimension'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'

    # Variables:
    RATIO = 'lunar_scene_to_rolo_irradiance_ratio'
    RATIO_ERROR = 'lunar_scene_to_rolo_irradiance_ratio_error'
    PIXEL_MASK = 'pixel_mask'
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths'
    EXPOSURE_CENTER_PIXEL_POSITIONS = 'exposure_center_pixel_positions'

    def __init__(self, filename, num_vert_pixels, num_horz_pixels,
                 wavelengths: np.ndarray, grid_pixel_positions: np.ndarray):
        """
        Parameters
        ----------
        filename: str

        num_vert_pixels: int

        num_horz_pixels: int

        wavelengths: np.ndarray
            Calibration wavelengths used for this channel

        grid_pixel_positions: np.ndarray
            Array [len(wavelengths), 2] of (vertical, horizontal) pixel positions where the moon
            is expected to be imaged on the detector
        """
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._wavelength_dim = self._write_nc_file.createDimension(self.WAVELENGTH_DIM,
                                                                   len(wavelengths))  # variable size (appending)
        self._scene_dim = self._write_nc_file.createDimension(self.SCENE_DIM, len(grid_pixel_positions))
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, num_vert_pixels)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, num_horz_pixels)
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)

        # Variables:
        self._exposure_detector_positions = self._write_nc_file.createVariable(self.EXPOSURE_CENTER_PIXEL_POSITIONS,
                                                                               int,
                                                                               (self.SCENE_DIM,
                                                                                self.PIXEL_LOCATION_SIZE))
        self._exposure_detector_positions[:] = grid_pixel_positions
        self._wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                               (self.WAVELENGTH_DIM,))
        self._wavelengths[:] = wavelengths
        self._irradiance_ratio = self._write_nc_file.createVariable(self.RATIO, float,
                                                                    (self.WAVELENGTH_DIM, self.SCENE_DIM))
        self._irradiance_ratio_error = self._write_nc_file.createVariable(self.RATIO_ERROR, float,
                                                                          (self.WAVELENGTH_DIM, self.SCENE_DIM))
        self._pixel_mask = self._write_nc_file.createVariable(self.PIXEL_MASK, int,
                                                              (self.WAVELENGTH_DIM, self.SCENE_DIM,
                                                               self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))

    def write_next_ratio_and_mask(self, irradiance_ratio: float, irradiance_ratio_error: float,
                                  pixel_mask: np.ndarray, expected_lunar_center: Tuple[float, float],
                                  wavelength: float):
        # Determine the closest detector positions and wavelengths:
        wl_idx = np.argmin(np.abs(wavelength - self._wavelengths))
        pos_idx = np.argmin(
            np.sqrt(np.sum((np.asarray(expected_lunar_center) - self._exposure_detector_positions) ** 2, axis=1)))

        self._irradiance_ratio[wl_idx, pos_idx] = irradiance_ratio
        self._irradiance_ratio_error[wl_idx, pos_idx] = irradiance_ratio_error

        self._pixel_mask[wl_idx, pos_idx, :, :] = pixel_mask

    def close(self):
        self._write_nc_file.close()