from typing import Tuple

import netCDF4 as nc
import numpy as np
from sktime import mjd as mjd_conv


class SpsHighResolutionUsaskToSpsSceneInterface:
    """
    Class for passing high-resolution (4096x4096) simulated images of the sun to the
    SPS.  TBD if QinetiQ wants this or just the detector resolution images -
    if they want to simulate satellite pointing drift, it's likely they'll want this instead
    of a pre-computed SpsSceneInterface at detector resolution

    Images of the sun are produced from the solar HMI database.  Originally, images were directly
    taken from the HMI database because they accurately show solar convection cells and how they evolve
    with time.  However, it was determined the HMI data contained a certain bias, so solar images were
    decided to be simulated.  Realistic simulation of the solar convection cells is achieved by
    adding Gaussian noise of the same SNR as the HMI images to a limb-darkened solar disk.

    The set of images in this file usually consist of ~100 solar images, which is about 4 GB.

    Dimensions:

    - TIME_DIM
        - value: 'time'
        - variable length dimension - depends on number of images passed in
    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels in the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels in the detector
    - PIXEL_LOCATION_DIM
        - value: 'pixel_location_dim'
        - vert, horz coordinates of any pixel
    - ICF_VECTOR_DIM
        - value: 'icf_vector_dim'
        - xyz of ICF vector
    - NUM_SPECTRUM_WAVELENGTHS
        - value: 'num_spectrum_wavelengths'
        - length of solar spectrum used

    Variables:

    - WAVELENGTHS
        - value: 'wavelengths'
        - size: [TIME_DIM,]
        - wavelength in nm of each simulation
    - HI_RES_SOLAR_DISKS
        - value: 'scenes'
        - size: [TIME_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - Simulated image of the sun at high spatial resolution, for each interpolated time.  In arbitrary units
    - TIMES
        - value: 'times'
        - size: [TIME_DIM,]
        - Time of each image measurement in MJD
    - IMAGE_SOLAR_CENTER
        - value: 'expected_solar_center'
        - size: [TIME_DIM, PIXEL_LOCATION_DIM]
        - Vertical, horizontal pixel location of where the solar disk is centered.  The simulations
          generally place the center in the middle of the array, [2047.5, 2047.5]
    - IMAGE_SOLAR_RADIUS
        - value: 'expected_solar_radius'
        - size: [TIME_DIM,]
        - Radius of the solar disk for each image.
    - DETECTOR_FOV_DEG
        - value: 'detector_fov_deg'
        - size: [PIXEL_LOCATION_DIM,]
        - Vertical and horizontal field of view in degrees on the detector
    - SOLAR_DISK_CENTER_ICF
        - value: 'solar_disk_center_icf'
        - size: [TIME_DIM, ICF_VECTOR_DIM]
        - Location of the center of the solar disk on the detector in ICF coordinates
    - SOLAR_SPECTRUM_WAVELENGTHS
        - value: 'solar_spectrum_wavelengths'
        - size: [NUM_SPECTRUM_WAVELENGTHS,]
        - wavelengths, in nm of the solar spectrum to be used to convolve with the SRF when scaling the solar disk
    - SOLAR_SPECTRUM
        - value: 'solar_spectrum'
        - size: [NUM_SPECTRUM_WAVELENGTHS,]
        - solar spectrum values in photons/s/cm2/nm
    """
    # Dimensions:
    TIME_DIM = 'time'
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horiz_pixels'
    PIXEL_LOCATION_DIM = 'pixel_location_dim'
    ICF_VECTOR_DIM = 'icf_vector_dim'
    NUM_SPECTRUM_WAVELENGTHS = 'num_spectrum_wavelengths'

    # Variable names:
    WAVELENGTHS = 'wavelengths'
    HI_RES_SOLAR_DISKS = 'scenes'
    TIMES = 'times'
    IMAGE_SOLAR_CENTER = 'image_solar_center'
    IMAGE_SOLAR_RADIUS = 'image_solar_radius'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    SOLAR_DISK_CENTER_ICF = 'solar_disk_center_icf'
    SOLAR_SPECTRUM_WAVELENGTHS = 'solar_spectrum_wavelengths'
    SOLAR_SPECTRUM = 'solar_spectrum'

    def __init__(self):
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            detector_fov: Tuple[float, float],
                            solar_spectrum_wavelengths: np.ndarray,
                            solar_spectrum: np.ndarray):
        """
        Create a new netcdf file to write images to.

        Parameters
        ----------
        filename: str
            Full path to the netcdf file to write.

        detector_fov: Tuple[float, float]
            Vertical, horizontal FOV of the detector in degrees.

        solar_spectrum_wavelengths: np.ndarray
            Wavelengths, in nm, where the solar spectrum is evaluated.

        solar_spectrum: np.ndarray
            Irradiance spectrum of the sun at solar_spectrum_wavelengths, in photons/s/cm2/nm
        """
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._time_dimension = self._write_nc_file.createDimension(self.TIME_DIM, None)  # variable size (appending)
        self._pixel_location_dim = self._write_nc_file.createDimension(self.PIXEL_LOCATION_DIM, 2)
        self._icf_dim = self._write_nc_file.createDimension(self.ICF_VECTOR_DIM, 3)
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, 4096)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, 4096)

        self._wavelengths = self._write_nc_file.createVariable(self.WAVELENGTHS, float, (self.TIME_DIM,))
        self._wavelengths.units = 'nm'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_DIM,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._num_spectrum_wavelengths = self._write_nc_file.createDimension(self.NUM_SPECTRUM_WAVELENGTHS,
                                                                             len(solar_spectrum_wavelengths))

        self._solar_spectrum_wavelengths = self._write_nc_file.createVariable(self.SOLAR_SPECTRUM_WAVELENGTHS,
                                                                              float,
                                                                              (self.NUM_SPECTRUM_WAVELENGTHS,))
        self._solar_spectrum_wavelengths.units = 'nm'
        self._solar_spectrum_wavelengths.description = 'Wavelengths (in nm) of the solar spectrum to be ' \
                                                       'used in scaling the solar disk'
        self._solar_spectrum_wavelengths[:] = solar_spectrum_wavelengths

        self._solar_spectrum = self._write_nc_file.createVariable(self.SOLAR_SPECTRUM,
                                                                  float,
                                                                  (self.NUM_SPECTRUM_WAVELENGTHS,))
        self._solar_spectrum.units = 'photons/s/cm2/nm'
        self._solar_spectrum.description = 'Total disk irradiance spectrum of the sun at 1 AU used for ' \
                                           'scaling the solar disk'
        self._solar_spectrum[:] = solar_spectrum

        self._solar_disks = self._write_nc_file.createVariable(self.HI_RES_SOLAR_DISKS, float,
                                                               (self.TIME_DIM,
                                                                self.NUM_VERT_PIXELS,
                                                                self.NUM_HORZ_PIXELS),
                                                               zlib=True, complevel=9,
                                                               least_significant_digit=6)
        self._solar_disks.units = 'arbitrary'
        self._solar_disks.description = 'High-resolution solar image with limb darkening applied at wavelength'

        self._times = self._write_nc_file.createVariable(self.TIMES, float, (self.TIME_DIM,))
        self._times.units = 'MJD (seconds since 1970-01-01)'

        self._image_solar_center = self._write_nc_file.createVariable(self.IMAGE_SOLAR_CENTER, float,
                                                                      (self.TIME_DIM, self.PIXEL_LOCATION_DIM))
        self._image_solar_center.description = 'Center of the solar disk in vertical ' \
                                               'and horizontal pixels, respectively'
        self._image_solar_center.units = 'HMI Pixels'
        self._image_disk_radius = self._write_nc_file.createVariable(self.IMAGE_SOLAR_RADIUS, float,
                                                                     (self.TIME_DIM,))
        self._image_disk_radius.units = 'HMI Pixels'

        self._solar_detector_location_icf = self._write_nc_file.createVariable(self.SOLAR_DISK_CENTER_ICF, float,
                                                                               (self.TIME_DIM, self.ICF_VECTOR_DIM))
        self._solar_detector_location_icf.description = 'Unit vector in ICF coordinates ' \
                                                        'pointing to the center of the solar disk on the detector'
        self._solar_detector_location_icf.units = 'ICF unit vector'

        self._current_index = 0

    def write_next_solar_disk_to_netcdf(self,
                                        time: np.datetime64,
                                        high_res_disk_image: np.ndarray,
                                        wavelength: float,
                                        disk_center: Tuple[float, float],
                                        disk_radius_pixels: float,
                                        icf_solar_center: np.ndarray):
        """
        Parameters
        ----------
        time: np.datetime64
            Time of the measurement over the course of the orbit.

        high_res_disk_image: np.ndarray
            Image of the sun (4096x4096), in arbitrary units

        wavelength: float
            Wavelength of the simulation in nm.

        disk_center: Tuple[float, float]
            Vertical, horizontal pixel center location of the disk.

        disk_radius_pixels: float
            Pixel radius of solar disk.

        icf_solar_center: np.ndarray
            ICF detector location of the solar disk center.
        """
        self._wavelengths[self._current_index] = wavelength
        self._times[self._current_index] = mjd_conv.datetime64_to_mjd(time)
        self._solar_disks[self._current_index, :, :] = high_res_disk_image
        self._image_solar_center[self._current_index, :] = [disk_center[0], disk_center[1]]
        self._image_disk_radius[self._current_index] = disk_radius_pixels
        self._solar_detector_location_icf[self._current_index, :] = icf_solar_center

        self._current_index += 1

    def close(self):
        self._write_nc_file.close()