from typing import Tuple

import netCDF4 as nc
import numpy as np
import xarray as xr
from sktime import mjd as mjd_conv

from altius_l1pp_inflight.scene.altius_scene_interfaces import AltiusSolarScene


class SpsToUsaskSolarPrnuSceneInterface:
    """
    This class represents the output from the SPS given to the calibration
    processes using images of the sun.  So far this includes the PRNU and SRF calibrations.
    It contains images of the sun for a single wavelength calibration.

    For a PRNU calibration, a single instance of the class will contain images of the sun
    at a single wavelength over all locations on the detector.  This usually consists of ~100
    images, typically at detector resolution, resulting in a ~1 GB file.

    The contents of the class are saved in netcdf format.

    Dimensions:

    - TIME_DIM
        - value: 'time'
        - variable length dimension - depends on number of images passed in
    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels in the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels in the detector
    - PIXEL_LOCATION_SIZE
        - value: 'pixel_location_size'
        - vert, horz coordinates of any pixel
    - CAL_WL_DIM
        - value: 'calibration_wavelength_nm'
        - length of the calibration wavelength (1)
    - BORESIGHT_VECTOR_DIM
        - value: 'boresight_vector_dim'
        - length of the vector for specifying look direction of the boresight (3)

    Variables:

    - DETECTOR_BORESIGHT
        - value: 'detector_boresight'
        - size: [PIXEL_LOCATION_SIZE,]
        - vertical, horizontal pixel location of the detector boresight
    - WAVELENGTHS
        - value: 'wavelengths'
        - size: [TIME_DIM,]
        - wavelength in nm of each measurement
    - SCENES
        - value: 'scenes'
        - size: [TIME_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - Image of the sun on the detector for each measurement.  In units of DN
    - TIMES
        - value: 'times'
        - size: [TIME_DIM,]
        - Time of each image measurement in MJD
    - EXPECTED_SOLAR_CENTER
        - value: 'expected_solar_center'
        - size: [TIME_DIM, PIXEL_LOCATION_SIZE]
        - Vertical, horizontal pixel location of where the solar disk was intended to be
          aimed on the detector surface.  The actual position of the sun may differ due
          to pointing errors.
    - EXPECTED_SOLAR_RADIUS
        - value: 'expected_solar_radius'
        - size: [TIME_DIM,]
        - Expected radius of the solar disk on the detector surface in pixels.  Can be directly computed
          based on the solar distance and constant solar radius.  This is only used to have a starting
          point for finding the actual solar radius of the image.
    - SAT_TO_SUN_DISTANCE
        - value: 'satellite_to_sun_distance'
        - size: [TIME_DIM, ]
        - Distance from the satellite to the sun at each measurement, in km.
    - DETECTOR_FOV_DEG
        - value: 'detector_fov_deg'
        - size: [PIXEL_LOCATION_SIZE,]
        - Vertical and horizontal field of view in degrees on the detector
    - CALIBRATION_WAVELENGTH
        - value: 'calibration_wavelength_nm'
        - size: [CAL_WL_DIM,]
        - The calibration wavelength in nm for the measurements
    - START_SNAPSHOT_BORESIGHT_POINTING
        - value: 'start_snapshot_pointing'
        - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
        - Vector of the boresight in (TBD coordinates) at the start of the integration for each measurement.
          Used to help correct for the blurring that occurs.
    - END_SNAPSHOT_BORESIGHT_POINTING
        - value: 'end_snapshot_pointing'
        - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
        - Vector of the boresight in (TBD coordinates) at the end of the integration for each measurement.
          Used to help correct for the blurring that occurs.
    - INTEGRATION_TIMES
        - value: 'integration_times'
        - size: [TIME_DIM,]
        - Integration time in seconds of each measurement.
    """

    # Dimensions:
    TIME_DIM = 'time'
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    PIXEL_LOCATION_SIZE = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    BORESIGHT_VECTOR_DIM = 'boresight_vector_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    WAVELENGTHS = 'wavelengths'
    SCENES = 'scenes'
    TIMES = 'times'
    EXPECTED_SOLAR_CENTER = 'expected_solar_center'
    EXPECTED_SOLAR_RADIUS = 'expected_solar_radius'
    SAT_TO_SUN_DISTANCE = 'satellite_to_sun_distance'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    CALIBRATION_WAVELENGTH = 'calibration_wavelength_nm'
    START_BORESIGHT_POINTING = 'start_snapshot_pointing'
    END_BORESIGHT_POINTING = 'end_snapshot_pointing'
    INTEGRATION_TIMES = 'integration_times'

    def __init__(self):
        self._read_nc_file = None
        self._write_nc_file = None

    def open_file_for_write(self,
                            filename: str,
                            calibration_wavelength: float,
                            detector_fov: Tuple[float, float],
                            detector_size_pixels: Tuple[int, int],
                            detector_boresight_pixels: Tuple[float, float]):
        """
        Initialize a new netcdf file for writing data.

        Parameters
        ----------
        filename : str
            Full path string to the netcdf file.

        calibration_wavelength : float
            Wavelength in nm for the calibration measurements

        detector_fov : Tuple[float, float]
            Vertical, horizontal FOV of the detector, in degrees

        detector_size_pixels : Tuple[int, int]
            Vertical, horizontal number of pixels for the detector

        detector_boresight_pixels : Tuple[float, float]
            Vertical, horizontal pixel location of the boresight of the detector.

        """

        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._time_dimension = self._write_nc_file.createDimension(self.TIME_DIM, None)  # variable size (appending)
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.PIXEL_LOCATION_SIZE, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, 1)
        self._boresight_vector_dim = self._write_nc_file.createDimension(self.BORESIGHT_VECTOR_DIM, 3)

        # Variables:
        self._snapshot_pointing_start = self._write_nc_file.createVariable(self.START_BORESIGHT_POINTING,
                                                                           float,
                                                                           (self.TIME_DIM, self.BORESIGHT_VECTOR_DIM))
        self._snapshot_pointing_start.units = 'Arbitrary'
        self._snapshot_pointing_start.description = 'ECEF absolute vector describing the pointing of the ' \
                                                    'instrument boresight at the beginning of each integration period'

        self._snapshot_pointing_end = self._write_nc_file.createVariable(self.END_BORESIGHT_POINTING,
                                                                         float,
                                                                         (self.TIME_DIM, self.BORESIGHT_VECTOR_DIM))
        self._snapshot_pointing_end.units = 'Arbitrary'
        self._snapshot_pointing_end.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                  'boresight at the end of each integration period'

        self._integration_times = self._write_nc_file.createVariable(self.INTEGRATION_TIMES, float, (self.TIME_DIM,))
        self._integration_times.units = 'seconds'
        self._integration_times.description = 'Exposure time of each snapshot of the sun'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.PIXEL_LOCATION_SIZE,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight_pixels[0], detector_boresight_pixels[1]]

        self._calibration_wavelength = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTH,
                                                                          float, (self.CAL_WL_DIM,))
        self._calibration_wavelength[:] = calibration_wavelength
        self._calibration_wavelength.units = 'nm'
        self._calibration_wavelength.description = 'Wavelength in nm for which the set of images is intended ' \
                                                   'to calibrate'

        self._wavelengths = self._write_nc_file.createVariable(self.WAVELENGTHS, float, (self.TIME_DIM,))
        self._wavelengths.units = 'nm'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.PIXEL_LOCATION_SIZE,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._scenes = self._write_nc_file.createVariable(self.SCENES, float,
                                                          (self.TIME_DIM,
                                                           self.NUM_VERT_PIXELS,
                                                           self.NUM_HORZ_PIXELS))
        self._scenes.units = 'DN'
        self._scenes.description = 'Scene at detector resolution'

        self._times = self._write_nc_file.createVariable(self.TIMES, float, (self.TIME_DIM,))
        self._times.units = 'MJD'

        self._expected_solar_center = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_CENTER, float,
                                                                         (self.TIME_DIM, self.PIXEL_LOCATION_SIZE))
        self._expected_solar_center.description = 'Center of the solar disk in vertical and horizontal pixels, ' \
                                                  'respectively'
        self._expected_solar_center.units = 'Pixels'

        self._expected_solar_radius = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_RADIUS, float,
                                                                         (self.TIME_DIM,))
        self._expected_solar_radius.units = 'Pixels'

        self._satellite_to_sun_distance = self._write_nc_file.createVariable(self.SAT_TO_SUN_DISTANCE, float,
                                                                             (self.TIME_DIM,))
        self._satellite_to_sun_distance.units = 'km'

        self._current_index = 0

    def write_next_solar_scene_to_netcdf(self,
                                         time: np.datetime64,
                                         scene: np.ndarray,
                                         wavelength: float,
                                         expected_solar_center: Tuple[float, float],
                                         expected_solar_radius: float,
                                         sat_sun_distance_km: float,
                                         start_boresight_pointing: np.ndarray,
                                         end_boresight_pointing: np.ndarray,
                                         integration_time: float):
        """
        Once a netcdf file is open for writing, calling this appends another detector
        image to the set of images in currently open netcdf.

        Parameters
        ----------
        time : np.datetime64
            Time at the start of the measurement.

        scene : np.ndarray
            The actual detector pixels with measured DN values of the sun.

        wavelength : float
            Wavelength in nm of the measurement, not necessarily the same as the calibration wavelength.

        expected_solar_center : Tuple[float, float]
            Vertical, horizontal pixel location of the location of the sun where it was intended to be.

        expected_solar_radius : float
            Pixel radius of the solar disk as expected by the detector FOV and solar distance.

        sat_sun_distance_km : float
            Distance from the satellite to the sun in km.

        start_boresight_pointing : np.ndarray
            xyz vector of the boresight in TBD coordinate system at the start of the integration time.

        end_boresight_pointing: np.ndarray
            xyz vector of the boresight in TBD coordinate system at the end of the integration time.

        integration_time : float
            Integration time of the measurement in seconds.
        """
        self._wavelengths[self._current_index] = wavelength
        self._times[self._current_index] = mjd_conv.datetime64_to_mjd(time)
        self._scenes[self._current_index, :, :] = scene
        self._expected_solar_center[self._current_index, :] = [expected_solar_center[0], expected_solar_center[1]]
        self._expected_solar_radius[self._current_index] = expected_solar_radius
        self._satellite_to_sun_distance[self._current_index] = sat_sun_distance_km
        self._snapshot_pointing_start[self._current_index, :] = start_boresight_pointing
        self._snapshot_pointing_end[self._current_index, :] = end_boresight_pointing
        self._integration_times[self._current_index] = integration_time

        self._current_index += 1

    def close_write_nc(self):
        self._write_nc_file.close()

    def close_read_nc(self):
        if self._read_nc_file is not None:
            self._read_nc_file.close()

    def open_read_dataset(self, filename):
        """
        Open this netcdf file for reading.

        Parameters
        ----------
        filename : str
            Path to the netcdf file.
        """
        self._read_nc_file = xr.open_dataset(filename)  # , chunks={SpsToUsaskSceneInterface.TIME_DIM: 4})

    def get_num_scenes(self):
        return self._read_nc_file.dims[SpsToUsaskSolarPrnuSceneInterface.TIME_DIM]

    def load_scene(self, scene_index) -> AltiusSolarScene:
        """
        Get an AltiusSolarScene instance for the given scene index from the current file.
        Must have called :code:`open_read_dataset` prior to this, and can use
        :code:`get_num_scenes` to figure out how many scenes are present.

        Parameters
        ----------
        scene_index : int
            Raster scene index of the solar scene wanted from the netcdf file.

        Returns
        -------
        val : AltiusSolarScene
            An instance of the AltiusSolarScene for this measurement.
        """
        vals = self._read_nc_file.variables
        expected_center = vals[SpsToUsaskSolarPrnuSceneInterface.EXPECTED_SOLAR_CENTER].values[scene_index]
        return AltiusSolarScene(vals[SpsToUsaskSolarPrnuSceneInterface.SCENES].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.DETECTOR_FOV_DEG].values,
                                vals[SpsToUsaskSolarPrnuSceneInterface.SAT_TO_SUN_DISTANCE].values[scene_index],
                                expected_center[1],
                                expected_center[0],
                                vals[SpsToUsaskSolarPrnuSceneInterface.EXPECTED_SOLAR_RADIUS].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.WAVELENGTHS].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.DETECTOR_BORESIGHT].values,
                                vals[SpsToUsaskSolarPrnuSceneInterface.START_BORESIGHT_POINTING].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.END_BORESIGHT_POINTING].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.INTEGRATION_TIMES].values[scene_index],
                                vals[SpsToUsaskSolarPrnuSceneInterface.TIMES].values[scene_index])


class SolarPrnuMapOutput:
    """
    This class describes the interface to be provided to the QinetiQ SPS as a result of the
    output of the PRNU calibration.

    For now, this consists of the following netcdf structure.

    Dimensions:

    - NUM_VERT_PIXELS
        - value: 'num_vert_pixels'
        - fixed number of vertical pixels on the detector
    - NUM_HORZ_PIXELS
        - value: 'num_horz_pixels'
        - fixed number of horizontal pixels on the detector
    - CAL_WL
        - value: 'cal_wl'
        - the length of the number of PRNU images, for each calibration wavelength

    Variables:

    - CALIBRATION_WAVELENGTHS
        - value: 'calibration_wavelengths'
        - size: [CAL_WL,]
        - wavelength, in nm, of each PRNU calibration
    - PRNU_MAP
        - value: 'prnu_map'
        - size: [CAL_WL, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - the output PRNU map for each calibration wavelength, on a unitary scale
    - PRNU_MAP_ERROR
        - value: 'prnu_map_error'
        - size: [CAL_WL, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
        - the error in each pixel in each PRNU map
    """
    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    CAL_WL = 'cal_wl'

    # Variables:
    CALIBRATION_WAVELENGTHS = 'calibration_wavelengths'
    PRNU_MAP = 'prnu_map'
    PRNU_MAP_ERROR = 'prnu_map_error'

    def __init__(self,
                 filename: str,
                 calibration_wavelength: float,
                 prnu_image: np.ndarray,
                 prnu_image_error: np.ndarray):
        """
        Parameters
        ----------
        filename: str
            Absolute file path of the output netcdf.

        calibration_wavelength: float
            Wavelength of the PRNU calibration in nm.

        prnu_image: np.ndarray
            The obtained PRNU image for the detector.  The PRNU image is obtained by correcting the
            solar images for limb darkening, stitching them together, and dividing by the ideal modelled
            and limb-darkening-corrected solar radiance.  The ideal image is a 2D array of 1.0's.

        prnu_image_error: np.ndarray
            The error in the PRNU for each detector pixel.
        """

        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        nv, nh = np.shape(prnu_image)

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, nv)
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, nh)
        self._cal_wl_dim = self._write_nc_file.createDimension(self.CAL_WL, 1)

        # Variables:
        self._calibration_wavelength = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTHS, float,
                                                                          (self.CAL_WL,))
        self._calibration_wavelength.units = 'nm'
        self._calibration_wavelength[:] = [calibration_wavelength]
        self._calibration_wavelength.description = 'Wavelength in nm for which the set of images is intended ' \
                                                   'to calibrate'

        self.prnu_map = self._write_nc_file.createVariable(self.PRNU_MAP, float,
                                                           (self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self.prnu_map.units = 'photons/s/cm2/sr/nm'
        self.prnu_map.description = 'PRNU Flat-field map after applying corrections and removing old PRNU'
        self.prnu_map[:, :] = prnu_image

        self.prnu_map_error = self._write_nc_file.createVariable(self.PRNU_MAP_ERROR, float,
                                                                 (self.NUM_VERT_PIXELS, self.NUM_HORZ_PIXELS))
        self.prnu_map_error.units = 'photons/s/cm2/sr/nm'
        self.prnu_map_error.description = 'Error in PRNU Flat-field map'
        self.prnu_map_error[:, :] = prnu_image_error

        self._write_nc_file.close()