from typing import Tuple

import netCDF4 as nc
import numpy as np
import xarray as xr
from sktime import mjd as mjd_conv

from altius_l1pp_inflight.scene.altius_scene_interfaces import AltiusSolarScene


class SpsSingleSolarImageInterface:
    """
        This class encompasses the raw output from the SPS of a solar
        image on the detector.

        The interface is very simple:

        Dimensions
        ----------
        vert_pixels:
            Vertical size in pixels of the ALTIUS image

        horz_pixels:
            Horizontal size in pixels of the ALTIUS image

        Variables
        ---------
        image: np.ndarray (vert_pixels x horz_pixels)
            Image of the sun for the full detector

        calibration_wavelength: float
            Central wavelength of the SRF band to be
            calibrated.  This wavelength is intended to be
            well-calibrated for retrievals

        measurement_wavelength: float
            One wavelength of the set of wavelengths
            surrounding the calibration wavelength where this
            measurement of the sun is acquired.

        distance_to_sun: float
            Distance, in km, from the sun to the satellite



    """
    def __init__(self, altius_image_dn: np.ndarray,
                 calibration_wavelength: float,
                 measurement_wavelength: float,
                 distance_to_sun: float,
                 start_pointing: np.ndarray,
                 end_pointing: np.ndarray,
                 integration_time: float,
                 time: np.datetime64,
                 detector_position: Tuple[float, float]):
        self._ds = xr.Dataset({'image': (['vert_pixels', 'horz_pixels'], altius_image_dn),
                               'calibration_wavelength': calibration_wavelength,
                               'measurement_wavelength': measurement_wavelength,
                               'distance_to_sun': distance_to_sun,
                               'start_pointing': (['xyz'], start_pointing),
                               'end_pointing': (['xyz'], end_pointing),
                               'integration_time': integration_time,
                               'time': time,
                               'detector_position': (['vh'], np.asarray(detector_position))},
                              coords={'xyz': ['x', 'y', 'z'],
                                      'vert_pixels': np.arange(0, len(altius_image_dn)),
                                      'horz_pixels': np.arange(0, len(altius_image_dn[0]))})
        self._ds['image'].attrs = {'unit': 'DN'}
        self._ds['calibration_wavelength'].attrs = {'unit': 'nm',
                                                    'description': 'Central wavelength of the SRF band to be '
                                                                   'calibrated.  This wavelength is intended to be'
                                                                   'well-calibrated for retrievals'}
        self._ds['measurement_wavelength'].attrs = {'unit': 'nm',
                                                    'description': 'One wavelength of the set of wavelengths '
                                                                   'surrounding the calibration wavelength where this'
                                                                   'measurement of the sun is acquired.'}
        self._ds['distance_to_sun'].attrs = {'unit': 'km',
                                             'description': 'Distance, in km, from the sun to the satellite'}
        self._ds['start_pointing'].attrs = {'unit': 'unitless',
                                            'description': 'Detector boresight pointing in TBD coordinates at beginning'
                                                           'of image acquisition'}
        self._ds['end_pointing'].attrs = {'unit': 'unitless',
                                          'description': 'Detector boresight pointing in TBD coordinates at end'
                                                         'of image acquisition'}
        self._ds['integration_time'].attrs = {'unit': 'seconds'}

        self._ds['time'].attrs = {'unit': 'datetime format'}

        self._ds['detector_position'].attrs = {'unit': 'Unitless',
                                               'description': 'Vertical and horizontal fractions of position along '
                                                              'the detector - 0.5, 0.5 is the middle'}

    def to_netcdf(self, filename: str):
        self._ds.to_netcdf(filename)

    def raw_data(self):
        return self._ds

    @staticmethod
    def from_netcdf(filename: str):
        ds = xr.open_dataset(filename)
        return SpsSingleSolarImageInterface(ds['image'].values,
                                            ds['calibration_wavelength'].values,
                                            ds['measurement_wavelength'].values,
                                            ds['distance_to_sun'].values,
                                            ds['start_pointing'],
                                            ds['end_pointing'],
                                            ds['integration_time'].values,
                                            ds['time'].values,
                                            tuple(ds['detector_position'].values))


class SpsToUsaskSolarSrfSceneInterface:
    """
        This class represents the output from the SPS given to the calibration
        processes using images of the sun.  So far this includes the PRNU and SRF calibrations.
        It contains images of the sun for a single wavelength calibration.

        For an SRF calibration, a single instance of the class will contain multiples wavelengths
        of the sun at multiple detector locations.  The number of wavelengths will depend on future studies to
        determine the number of wavelengths necessary for an accurate calibration.
        The measurements are intended for calibrating a single wavelength at the center
        of the measurement wavelengths.

        For a UV calibration with 8 wavelengths and 4 detector locations for the sun at detector
        resolution, we get 8 bytes * 513x513 pixels * 8 wavelengths * 30 measurements per wavelength * 4 locations
        = 2 GB, and for vis with about 4 calibration wavelengths, the size of the file is 8.6 GB (at native resolution)

        The contents of the class are saved in netcdf format.

        Dimensions:

        - NUM_VERT_PIXELS
            - value: 'num_vert_pixels'
            - fixed number of vertical pixels in the detector
        - NUM_HORZ_PIXELS
            - value: 'num_horz_pixels'
            - fixed number of horizontal pixels in the detector
        - VH
            - value: 'pixel_location_size'
            - number of coordinates to describe a pixel in vertical, horizontal (2)
        - CAL_WL_DIM
            - value: 'calibration_wavelength_nm'
            - number of calibration wavelengths
        - XYZ
            - value: 'xyz'
            - Length of a vector (3)
        - MEASUREMENT_WAVELENGTH_DIM
            - value: 'measurement_wavelength_dim'
            - Number of measurements for each calibration wavelength
        - DETECTOR_POS_PIXEL_DIM
            - value: 'detector_position_pixel_dim'
            - Total number of unique positions on the detector for the center of the sun

        Variables:

        - DETECTOR_BORESIGHT
            - value: 'detector_boresight'
            - size: [PIXEL_LOCATION_SIZE,]
            - vertical, horizontal pixel location of the detector boresight
        - WAVELENGTHS
            - value: 'wavelengths'
            - size: [TIME_DIM,]
            - wavelength in nm of each measurement
        - SCENES
            - value: 'scenes'
            - size: [TIME_DIM, NUM_VERT_PIXELS, NUM_HORZ_PIXELS]
            - Image of the sun on the detector for each measurement.  In units of DN
        - TIMES
            - value: 'times'
            - size: [TIME_DIM,]
            - Time of each image measurement in MJD
        - EXPECTED_SOLAR_CENTER
            - value: 'expected_solar_center'
            - size: [TIME_DIM, PIXEL_LOCATION_SIZE]
            - Vertical, horizontal pixel location of where the solar disk was intended to be
              aimed on the detector surface.  The actual position of the sun may differ due
              to pointing errors.
        - EXPECTED_SOLAR_RADIUS
            - value: 'expected_solar_radius'
            - size: [TIME_DIM,]
            - Expected radius of the solar disk on the detector surface in pixels.  Can be directly computed
              based on the solar distance and constant solar radius.  This is only used to have a starting
              point for finding the actual solar radius of the image.
        - SAT_TO_SUN_DISTANCE
            - value: 'satellite_to_sun_distance'
            - size: [TIME_DIM, ]
            - Distance from the satellite to the sun at each measurement, in km.
        - DETECTOR_FOV_DEG
            - value: 'detector_fov_deg'
            - size: [PIXEL_LOCATION_SIZE,]
            - Vertical and horizontal field of view in degrees on the detector
        - CALIBRATION_WAVELENGTH
            - value: 'calibration_wavelength_nm'
            - size: [CAL_WL_DIM,]
            - The calibration wavelength in nm for the measurements
        - START_SNAPSHOT_BORESIGHT_POINTING
            - value: 'start_snapshot_pointing'
            - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
            - Vector of the boresight in (TBD coordinates) at the start of the integration for each measurement.
              Used to help correct for the blurring that occurs.
        - END_SNAPSHOT_BORESIGHT_POINTING
            - value: 'end_snapshot_pointing'
            - size: [TIME_DIM, BORESIGHT_VECTOR_DIM]
            - Vector of the boresight in (TBD coordinates) at the end of the integration for each measurement.
              Used to help correct for the blurring that occurs.
        - INTEGRATION_TIMES
            - value: 'integration_times'
            - size: [TIME_DIM,]
            - Integration time in seconds of each measurement.
        """
    # Dimensions:
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    VH = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    XYZ = 'xyz'
    MEASUREMENT_WAVELENGTH_DIM = 'measurement_wavelength_dim'
    DETECTOR_POS_PIXEL_DIM = 'detector_position_pixel_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    WAVELENGTHS = 'wavelengths'
    SUMMED_CENTER_PIXELS = 'summed_center_pixels'
    TIMES = 'times'
    EXPECTED_SOLAR_CENTER = 'expected_solar_center'
    EXPECTED_SOLAR_RADIUS = 'expected_solar_radius'
    SAT_TO_SUN_DISTANCE = 'satellite_to_sun_distance'
    DETECTOR_FOV_DEG = 'detector_fov_deg'
    CALIBRATION_WAVELENGTH = 'calibration_wavelength_nm'
    START_BORESIGHT_POINTING = 'start_snapshot_pointing'
    END_BORESIGHT_POINTING = 'end_snapshot_pointing'
    INTEGRATION_TIMES = 'integration_times'
    DETECTOR_POS_PIXEL = 'detector_position_pixel'
    CENTER_LOBE_SRF_SUM = 'center_lobe_srf_sum'

    def __init__(self):
        self._write_nc_file = None

    def open_file_for_write(self, filename: str,
                            calibration_wavelengths: np.ndarray,
                            detector_pixel_positions: np.ndarray,
                            detector_size_pixels: Tuple[int, int],
                            detector_boresight_pixels: Tuple[float, float],
                            detector_fov: Tuple[float, float],
                            num_wavelengths_per_measurement: int):
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.VH, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, len(calibration_wavelengths))
        self._boresight_vector_dim = self._write_nc_file.createDimension(self.XYZ, 3)
        self._measurement_wavelength_dim = self._write_nc_file.createDimension(self.MEASUREMENT_WAVELENGTH_DIM,
                                                                               num_wavelengths_per_measurement)
        self._detector_pixel_position = self._write_nc_file.createDimension(
            self.DETECTOR_POS_PIXEL_DIM, len(detector_pixel_positions))

        # Variables:
        self._detector_pixel_positions = self._write_nc_file.createVariable(self.DETECTOR_POS_PIXEL,
                                                                            float,
                                                                            (self.DETECTOR_POS_PIXEL_DIM, self.VH))
        self._detector_pixel_positions[:] = detector_pixel_positions
        self._detector_pixel_positions.units = 'Pixels'
        self._detector_pixel_positions.description = '(vert, horz) pixel position of the center of the sun'

        self._central_srf_lobe_sum = self._write_nc_file.createVariable(self.CENTER_LOBE_SRF_SUM,
                                                                        float,
                                                                        (self.DETECTOR_POS_PIXEL_DIM,
                                                                         self.CAL_WL_DIM,
                                                                         self.MEASUREMENT_WAVELENGTH_DIM))
        self._central_srf_lobe_sum.units = 'arbitrary'
        self._central_srf_lobe_sum.description = 'estimate of the integral under the central lobe of each SRF ' \
                                                 'at each measurement wavelength'

        self._snapshot_pointing_start = self._write_nc_file.createVariable(self.START_BORESIGHT_POINTING,
                                                                           float,
                                                                           (self.DETECTOR_POS_PIXEL_DIM,
                                                                            self.CAL_WL_DIM,
                                                                            self.MEASUREMENT_WAVELENGTH_DIM,
                                                                            self.XYZ))
        self._snapshot_pointing_start.units = 'Arbitrary'
        self._snapshot_pointing_start.description = 'ECEF absolute vector describing the pointing of the ' \
                                                    'instrument boresight at the beginning of each integration period'

        self._snapshot_pointing_end = self._write_nc_file.createVariable(self.END_BORESIGHT_POINTING,
                                                                         float,
                                                                         (self.DETECTOR_POS_PIXEL_DIM,
                                                                          self.CAL_WL_DIM,
                                                                          self.MEASUREMENT_WAVELENGTH_DIM,
                                                                          self.XYZ))
        self._snapshot_pointing_end.units = 'Arbitrary'
        self._snapshot_pointing_end.description = 'ECEF absolute vector describing the pointing of the instrument ' \
                                                  'boresight at the end of each integration period'

        self._integration_times = self._write_nc_file.createVariable(self.INTEGRATION_TIMES, float,
                                                                     (self.DETECTOR_POS_PIXEL_DIM,
                                                                      self.CAL_WL_DIM,
                                                                      self.MEASUREMENT_WAVELENGTH_DIM))
        self._integration_times.units = 'seconds'
        self._integration_times.description = 'Exposure time of each snapshot of the sun'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.VH,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight_pixels[0], detector_boresight_pixels[1]]

        self._calibration_wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTH,
                                                                           float, (self.CAL_WL_DIM,))
        self._calibration_wavelengths[:] = calibration_wavelengths
        self._calibration_wavelengths.units = 'nm'
        self._calibration_wavelengths.description = 'Wavelength in nm for which the set of images is intended ' \
                                                    'to calibrate'

        self._measurement_wavelengths = self._write_nc_file.createVariable(self.WAVELENGTHS, float,
                                                                           (self.DETECTOR_POS_PIXEL_DIM,
                                                                            self.CAL_WL_DIM,
                                                                            self.MEASUREMENT_WAVELENGTH_DIM))
        self._measurement_wavelengths.units = 'nm'
        self._measurement_wavelengths.description = 'Set of wavelengths over which each calibration measurement is ' \
                                                    'performed.'

        self._detector_fov = self._write_nc_file.createVariable(self.DETECTOR_FOV_DEG, float,
                                                                (self.VH,))
        self._detector_fov.units = 'deg'
        self._detector_fov.description = 'Vertical and horizontal angular width of detector, respectively'
        self._detector_fov[:] = [detector_fov[0], detector_fov[1]]

        self._summed_center_pixels = self._write_nc_file.createVariable(self.SUMMED_CENTER_PIXELS, float,
                                                                        (self.DETECTOR_POS_PIXEL_DIM,
                                                                         self.CAL_WL_DIM,
                                                                         self.MEASUREMENT_WAVELENGTH_DIM))
        self._summed_center_pixels.units = 'DN'
        self._summed_center_pixels.description = 'Summed center pixels of the sun at detector resolution in DN'

        self._times = self._write_nc_file.createVariable(self.TIMES, float,
                                                         (self.DETECTOR_POS_PIXEL_DIM,
                                                          self.CAL_WL_DIM,
                                                          self.MEASUREMENT_WAVELENGTH_DIM))
        self._times.units = 'MJD'

        self._expected_solar_center = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_CENTER, float,
                                                                         (self.DETECTOR_POS_PIXEL_DIM,
                                                                          self.CAL_WL_DIM,
                                                                          self.MEASUREMENT_WAVELENGTH_DIM,
                                                                          self.VH))
        self._expected_solar_center.description = 'Center of the solar disk in vertical and horizontal pixels, ' \
                                                  'respectively'
        self._expected_solar_center.units = 'Pixels'

        self._expected_solar_radius = self._write_nc_file.createVariable(self.EXPECTED_SOLAR_RADIUS, float,
                                                                         (self.DETECTOR_POS_PIXEL_DIM,
                                                                          self.CAL_WL_DIM,
                                                                          self.MEASUREMENT_WAVELENGTH_DIM))
        self._expected_solar_radius.units = 'Pixels'

        self._satellite_to_sun_distance = self._write_nc_file.createVariable(self.SAT_TO_SUN_DISTANCE, float,
                                                                             (self.DETECTOR_POS_PIXEL_DIM,
                                                                              self.CAL_WL_DIM,
                                                                              self.MEASUREMENT_WAVELENGTH_DIM))
        self._satellite_to_sun_distance.units = 'km'

    def write_current_measurement_wavelengths(self, detector_position: Tuple[int, int],
                                              calibration_wavelength,
                                              measurement_wavelengths: np.ndarray):
        self._current_meas_wl = measurement_wavelengths
        self._set_current_write_indices(detector_position, calibration_wavelength)
        self._measurement_wavelengths[self._current_pos_idx, self._cal_wl_idx, :] = measurement_wavelengths

    def _set_current_write_indices(self, detector_position: Tuple[int, int],
                                   calibration_wavelength: float):
        self._cal_wl_idx = np.argmin(np.abs(calibration_wavelength - self._calibration_wavelengths))
        self._current_pos_idx = np.argmin(
            np.abs(np.sqrt(np.sum((np.asarray(detector_position) - self._detector_pixel_positions) ** 2))))

    def close_write(self):
        self._write_nc_file.close()

    def close_read(self):
        self._read_nc_file.close()

    def write_next_solar_scene_to_netcdf(self,
                                         time: np.datetime64,
                                         summed_center_pixel_val: float,
                                         measurement_wavelength: float,
                                         expected_solar_center: Tuple[float, float],
                                         expected_solar_radius: float,
                                         sat_sun_distance_km: float,
                                         start_boresight_point: np.ndarray,
                                         end_boresight_point: np.ndarray,
                                         integration_time: float,
                                         central_srf_lobe_sum: float):
        """
        Note:
        It is required to call write_current_measurement_wavelengths() prior to this call.

        Once a netcdf file is open for writing, calling this appends another detector
        image to the set of images in currently open netcdf.

        Parameters
        ----------
        time : np.datetime64
            Time at the start of the measurement.

        summed_center_pixel_val : float


        expected_solar_center : Tuple[float, float]
            Vertical, horizontal pixel location of the location of the sun where it was intended to be.

        expected_solar_radius : float
            Pixel radius of the solar disk as expected by the detector FOV and solar distance.

        sat_sun_distance_km : float
            Distance from the satellite to the sun in km.

        start_boresight_point : np.ndarray
            xyz vector of the boresight in TBD coordinate system at the start of the integration time.

        end_boresight_point: np.ndarray
            xyz vector of the boresight in TBD coordinate system at the end of the integration time.

        integration_time : float
            Integration time of the measurement in seconds.
        """
        meas_wl_idx = np.argmin(np.abs(measurement_wavelength - self._current_meas_wl))

        self._times[self._current_pos_idx, self._cal_wl_idx] = mjd_conv.datetime64_to_mjd(time)
        self._summed_center_pixels[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx] = summed_center_pixel_val
        self._expected_solar_center[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx, :] = expected_solar_center
        self._expected_solar_radius[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx] = expected_solar_radius
        self._satellite_to_sun_distance[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx] = sat_sun_distance_km
        self._snapshot_pointing_start[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx, :] = start_boresight_point
        self._snapshot_pointing_end[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx, :] = end_boresight_point
        self._integration_times[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx] = integration_time
        self._central_srf_lobe_sum[self._current_pos_idx, self._cal_wl_idx, meas_wl_idx] = central_srf_lobe_sum

    def open_read_dataset(self, filename):
        """
        Open this netcdf file for reading.

        Parameters
        ----------
        filename : str
            Path to the netcdf file.
        """
        self._read_nc_file = xr.open_dataset(filename)  # , chunks={SpsToUsaskSceneInterface.TIME_DIM: 4})

    # def get_num_scenes(self):
    #     return self._read_nc_file.dims[SpsToUsaskSolarSrfSceneInterface.]

    def _set_current_read_indices(self, detector_position: Tuple[int, int],
                                  calibration_wavelength: float):
        cal_wl = self._read_nc_file[self.CALIBRATION_WAVELENGTH].values
        det_pos = self._read_nc_file[self.DETECTOR_POS_PIXEL].values
        self._cal_wl_idx = np.argmin(np.abs(calibration_wavelength - cal_wl))
        self._current_pos_idx = np.argmin(
            np.abs(np.sqrt(np.sum((np.asarray(detector_position) - det_pos) ** 2))))

    def load_scene(self,
                   detector_position: Tuple[int, int],
                   calibration_wavelength: float,
                   measurement_wavelength: float) -> AltiusSolarScene:
        """
        Get an AltiusSolarScene instance for the given scene index from the current file.
        Must have called :code:`open_read_dataset` prior to this, and can use
        :code:`get_num_scenes` to figure out how many scenes are present.

        Parameters
        ----------

        Returns
        -------
        val : AltiusSolarScene
            An instance of the AltiusSolarScene for this measurement.
        """
        self._set_current_read_indices(detector_position, calibration_wavelength)
        vals = self._read_nc_file.variables
        a = self._current_pos_idx
        b = self._cal_wl_idx
        meas_wl_idx = np.argmin(np.abs(measurement_wavelength -
                                       vals[SpsToUsaskSolarSrfSceneInterface.WAVELENGTHS].values[a, b]))
        c = meas_wl_idx
        expected_center = vals[SpsToUsaskSolarSrfSceneInterface.EXPECTED_SOLAR_CENTER].values[a, b, c]
        return AltiusSolarScene(
            vals[SpsToUsaskSolarSrfSceneInterface.SUMMED_CENTER_PIXELS].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.DETECTOR_FOV_DEG].values,
            vals[SpsToUsaskSolarSrfSceneInterface.SAT_TO_SUN_DISTANCE].values[a, b, c],
            expected_center[1],
            expected_center[0],
            vals[SpsToUsaskSolarSrfSceneInterface.EXPECTED_SOLAR_RADIUS].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.WAVELENGTHS].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.DETECTOR_BORESIGHT].values,
            vals[SpsToUsaskSolarSrfSceneInterface.START_BORESIGHT_POINTING].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.END_BORESIGHT_POINTING].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.INTEGRATION_TIMES].values[a, b, c],
            vals[SpsToUsaskSolarSrfSceneInterface.TIMES].values[a, b, c])


class SolarSrfCalibrationOutput:
    """
        Netcdf format holding the results of the SRF calibration using the sun as a source.  The core results
        are FWHM of the SRF central lobe and wavelength shift of the SRF peak at each calibration wavelength
        for each radial position on the detector.  The calibration currently assumes radial symmetry of the SRF
        around the boresight of the detector.
    """
    NUM_VERT_PIXELS = 'num_vert_pixels'
    NUM_HORZ_PIXELS = 'num_horz_pixels'
    VH = 'pixel_location_size'
    CAL_WL_DIM = 'calibration_wavelength_dimension'
    XYZ = 'xyz'
    DETECTOR_POS_PIXEL_DIM = 'detector_position_pixel_dim'

    # Variable names:
    DETECTOR_BORESIGHT = 'detector_boresight'
    FWHM = 'fwhm'
    FWHM_ERROR = 'fwhm_error'
    SHIFT = 'shift'
    SHIFT_ERROR = 'shift_error'
    TIMES = 'times'
    CALIBRATION_WAVELENGTH = 'calibration_wavelength_nm'
    DETECTOR_POS_PIXEL = 'detector_position_pixel'

    def __init__(self, filename: str, calibration_wavelengths: np.ndarray,
                 detector_position_pixels: np.ndarray,
                 detector_boresight: Tuple[float, float],
                 detector_size_pixels: Tuple[int, int]):
        self._write_nc_file = nc.Dataset(filename, mode='w', format='NETCDF4')

        # Dimensions:
        self._num_vert_pixels = self._write_nc_file.createDimension(self.NUM_VERT_PIXELS, detector_size_pixels[0])
        self._num_horz_pixels = self._write_nc_file.createDimension(self.NUM_HORZ_PIXELS, detector_size_pixels[1])
        self._pixel_location_size = self._write_nc_file.createDimension(self.VH, 2)
        self._calibration_wl_dim = self._write_nc_file.createDimension(self.CAL_WL_DIM, len(calibration_wavelengths))
        self._detector_position_pixel_dim = self._write_nc_file.createDimension(
            self.DETECTOR_POS_PIXEL_DIM, len(detector_position_pixels))

        # Variables:
        self._detector_pixel_positions = self._write_nc_file.createVariable(self.DETECTOR_POS_PIXEL,
                                                                            float,
                                                                            (self.DETECTOR_POS_PIXEL_DIM, self.VH))
        self._detector_pixel_positions[:] = detector_position_pixels
        self._detector_pixel_positions.units = 'Pixels'
        self._detector_pixel_positions.description = 'Pixel position of the center of the sun on the detector'

        self._detector_boresight = self._write_nc_file.createVariable(self.DETECTOR_BORESIGHT, float,
                                                                      (self.VH,))
        self._detector_boresight.units = 'Pixels'
        self._detector_boresight[:] = [detector_boresight[0], detector_boresight[1]]

        self._calibration_wavelengths = self._write_nc_file.createVariable(self.CALIBRATION_WAVELENGTH,
                                                                           float, (self.CAL_WL_DIM,))
        self._calibration_wavelengths[:] = calibration_wavelengths
        self._calibration_wavelengths.units = 'nm'
        self._calibration_wavelengths.description = 'Wavelength in nm for which the set of images is intended ' \
                                                    'to calibrate'

        self._fwhm = self._write_nc_file.createVariable(self.FWHM, float,
                                                        (self.DETECTOR_POS_PIXEL_DIM, self.CAL_WL_DIM))
        self._fwhm.units = 'nm'
        self._fwhm.description = 'Calibrated FWHM (in nm) of each calibration wavelength at each detector position'

        self._fwhm_error = self._write_nc_file.createVariable(self.FWHM_ERROR, float,
                                                              (self.DETECTOR_POS_PIXEL_DIM, self.CAL_WL_DIM))
        self._fwhm_error.units = 'nm'
        self._fwhm_error.description = 'Error in fit FWHM for calibration wavelength central lobe'

        self._shift = self._write_nc_file.createVariable(self.SHIFT, float,
                                                         (self.DETECTOR_POS_PIXEL_DIM, self.CAL_WL_DIM))
        self._shift.units = 'nm'
        self._shift.description = 'Wavelength shift undergone by the SRF central lobe at the calibration wavelength'

        self._shift_error = self._write_nc_file.createVariable(self.SHIFT_ERROR, float,
                                                               (self.DETECTOR_POS_PIXEL_DIM, self.CAL_WL_DIM))
        self._shift_error.units = 'nm'
        self._shift_error.desciption = 'Error in wavelength shift undergone by the SRF central lobe ' \
                                       'at the calibration wavelength'

        self._times = self._write_nc_file.createVariable(self.TIMES, float,
                                                         (self.DETECTOR_POS_PIXEL_DIM, self.CAL_WL_DIM))
        self._times.units = 'MJD'
        self._times.description = 'Average time of measurement for each calibration wavelength at each ' \
                                  'detector position'

    def write_current_calibration_results(self,
                                          fwhm: float,
                                          wavelength_shift: float,
                                          time: np.datetime64,
                                          detector_position: Tuple[int, int],
                                          calibration_wavelength):
        cal_wl_idx = np.argmin(np.abs(calibration_wavelength - self._calibration_wavelengths))
        current_pos_idx = np.argmin(
            np.sum((np.asarray(
                detector_position) - self._detector_pixel_positions) ** 2))

        self._times[current_pos_idx, cal_wl_idx] = mjd_conv.datetime64_to_mjd(time)
        self._fwhm[current_pos_idx, cal_wl_idx] = fwhm
        self._shift[current_pos_idx, cal_wl_idx] = wavelength_shift

        # TODO real error will eventually be incorporated
        self._fwhm_error[current_pos_idx, cal_wl_idx] = 0.01 * fwhm
        self._shift_error[current_pos_idx, cal_wl_idx] = 0.01 * wavelength_shift
