import logging

import numpy as np
import sasktran
import xarray as xr
from scipy.interpolate import RectBivariateSpline
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf


def gauss_blur(mu, variance):
    """ For now, the Gaussian weights used in the convolution.  But will eventually be a
    more general class or custom data """
    blur = 1 / np.sqrt(2 * np.pi * variance) * np.exp(-.5 * mu ** 2 / variance)
    return blur / sum(blur)


class BlurredSolarSpectrum(RectBivariateSpline):
    """ Class defining the Solar Spectrum as a function of wavelength and PSF blur std. dev."""

    DETECTOR_LEVEL = 10000
    window_size = 6.0  # Must make an integer when multiplied by 2000. Should not be less then 5 sigma
    num_window = int(window_size * 2000)  # this gives 0.001 nm spacing, hence the interpolation later on
    spacing = 2 * window_size / num_window

    high_res_spectra = None
    noisy_high_res_spectra = None

    # def __init__(self, ideal_spectra, noisy_spectra, sigma, start_wavelen=280, end_wavelen=380):
    def __init__(self, ideal_spectra: xr.DataArray,
                 noisy_spectra: xr.DataArray,
                 fwhm_delta: np.ndarray,
                 srf: GaussianSrf,
                 start_wavelen=280, end_wavelen=380):
        """
        Parameters
        ----------
        ideal_spectra  : xarray.DataArray
            1D Radiance spectrum as function of wavelength

        noisy_spectra  : xarray.DataArray
            1D Radiance spectrum with noise in the measurement

        sigma  : np.array
            Range of standard deviations of Gaussian blurring PSF to apply to
            the spectrum

        """
        self.min_wavelength = start_wavelen
        self.max_wavelength = end_wavelen
        self.high_res_spectra = ideal_spectra
        self.noisy_high_res_spectra = noisy_spectra
        # mu = np.linspace(-self.window_size, self.window_size, self.num_window + 1)
        # psf_array = np.array([gauss_blur(mu, sig ** 2.0) for sig in sigma])
        spectra_wavelen = self._convolution_wavelengths(window_size=self.window_size)
        convolution_spectra = self.high_res_spectra.interp(wavelength=spectra_wavelen).values
        # blurred2d = np.array([np.convolve(row, convolution_spectra) for row in psf_array])
        downsample = 10
        convolution_wl = np.arange(self.min_wavelength, self.max_wavelength, self.spacing * downsample)
        blurred2d = []
        for delta in fwhm_delta:
            new_srf = GaussianSrf(srf.grid_fwhm_wavelengths(), srf.grid_fwhm() + delta)
            conv_spectrum = new_srf.convolve_spectrum_with_srf(convolution_wl,
                                                               spectra_wavelen,
                                                               convolution_spectra)
            blurred2d.append(conv_spectrum.flatten())

        blurred2d = np.asarray(blurred2d)

        # downsampled_signal = blurred2d[:, self.num_window:-self.num_window:downsample]
        super().__init__(fwhm_delta, convolution_wl, blurred2d)

    def _convolution_wavelengths(self, window_size, shift=0.0):
        return np.arange(self.min_wavelength + window_size + shift,
                         self.max_wavelength - window_size + shift, 0.004)  # self.spacing)

    def get_measured_spectrum(self, actual_psf, psf_window_size, wavelen_shift, num_samples, snr=None):
        return self.compute_simulated_spectrum(actual_psf, psf_window_size, wavelen_shift, num_samples, snr)

    def compute_simulated_spectrum(self, actual_psf, psf_window_size, wavelen_shift, num_samples, snr=None) \
            -> (xr.DataArray, float):
        """Convenience method for generating a custom blurred and shifted test spectrum, with noise"""

        spectrum = self.high_res_spectra

        convolution_wavelengths = self._convolution_wavelengths(window_size=psf_window_size, shift=wavelen_shift)
        assumed_wavelengths = self._convolution_wavelengths(window_size=psf_window_size)  # no shift
        actual_blurred_spectrum = np.convolve(actual_psf,
                                              spectrum.interp(wavelength=convolution_wavelengths).values,
                                              mode='same')
        actual_blurred_spectrum = xr.DataArray(actual_blurred_spectrum, dims={'wavelength'},
                                               coords={'wavelength': assumed_wavelengths})

        # sample the spectrum at these wavelengths
        sample_wavelengths = np.linspace(self.min_wavelength + 0.5 * psf_window_size,
                                         self.max_wavelength - 0.5 * psf_window_size, num_samples)
        measured_spectrum = actual_blurred_spectrum.interp(wavelength=sample_wavelengths)

        actual_snr = snr

        if snr is not None:
            # spectrum = self.noisy_high_res_spectra

            factor = self.DETECTOR_LEVEL / np.mean(measured_spectrum.values)
            noisy_signal = np.random.poisson(factor * measured_spectrum.values, len(measured_spectrum))
            actual_snr = np.mean(factor * measured_spectrum.values) / np.mean(
                np.abs(noisy_signal - factor * measured_spectrum.values))

            noise_reduction_factor = snr / actual_snr
            num_measurements = int(np.ceil(noise_reduction_factor))

            # TODO handle the case for 0<n<1, where we need to amplify the noise (might not be necessary for DN=10000)
            if noise_reduction_factor >= 1:
                logging.info('Determined number of measurements per wavelength %d to achieve SNR=%.3f',
                             int(num_measurements), snr)
                measured_with_noise = 0
                for _ in np.arange(0, num_measurements):
                    measured_with_noise = np.random.poisson(factor * measured_spectrum,
                                                            len(measured_spectrum)) + measured_with_noise

                measured_with_noise = xr.DataArray(measured_with_noise / num_measurements / factor,
                                                   dims={'wavelength'},
                                                   coords={'wavelength': sample_wavelengths})

            else:
                logging.info('Simulating excessively noisy measurement with SNR=%.3f', snr)
                # just use the sqrt(A) approximation to figure out poisson mean
                p = snr ** 2
                factor = p / np.mean(measured_spectrum).values
                measured_with_noise = np.random.poisson(factor * measured_spectrum, len(measured_spectrum))
                measured_with_noise = xr.DataArray(measured_with_noise / factor,
                                                   dims={'wavelength'}, coords={'wavelength': sample_wavelengths})

            actual_snr = np.mean(measured_spectrum).values / np.mean(
                np.abs(measured_with_noise.values - measured_spectrum.values))
            measured_spectrum = measured_with_noise

        return measured_spectrum, actual_snr


class SAO2010(BlurredSolarSpectrum):
    """
        Implements the Solar Spectrum Class with the SAO2010 data.
        Class which takes in a Gaussian SRF model, which can vary FWHM in wavelength,
        and computes the solar spectrum by applying :code:`fwhm_delta` to the SRF
        to obtain a wide range of solar spectra where the FWHM vary, and are either scaled
        up or down in line width.  The underlying class interpolates the data to obtain
        quick lookups as a function of FWHM delta.

        This model assumes any shift in width  to the FWHM of an individual SRF is
        applied uniformly to all SRFs in the wavelength range.

        :param fwhm_delta:
            Array of deltas to the FWHM of the SRFs, in nm.
        :param gaussian_srf:
            Gaussian SRF model, which assumes Gaussian line shapes but with potentially
            varying FWHMs.
        :param start_wavelen:
            Start wavelength for the spectrum, in nm.
        :param end_wavelen:
            End wavelength for the spectrum, in nm.
        :param solar_distance_AU:
            Average distance from the satellite to the sun, in AU.
        :param scaling_factor:
            Scaling factor for the irradiance, accounting for the fact that we may be
            modelling only the center pixels of the sun.  Is a wavelength-dependent xarray.
    """
    DETECTOR_LEVEL = 10000

    def __init__(self, fwhm_delta: np.ndarray, gaussian_srf: GaussianSrf,
                 start_wavelen=280, end_wavelen=380,
                 solar_distance_AU=1.0, scaling_factor: xr.DataArray = None):

        solar_spectrum = sasktran.solarspectrum.SolarSpectrum(name='SAO2010')

        # resolutions = solar_spectrum.resolution([start_wavelen, end_wavelen])
        wavelen_resolution = 0.001  # np.max(resolutions)
        n_hires = int((end_wavelen - start_wavelen) / wavelen_resolution)
        wavelen = np.linspace(start_wavelen, end_wavelen, n_hires)

        sao = solar_spectrum.irradiance(wavelen, solardistance=solar_distance_AU)
        if scaling_factor is not None:
            sf = scaling_factor.interp(wavelength=wavelen)
            sao *= sf.values
        solar = np.array([(nm, sun) for (nm, sun) in zip(wavelen, sao)])
        solar = xr.DataArray(solar[:, 1], dims={'wavelength'}, coords={'wavelength': solar[:, 0]})
        self.num_measurements = 1
        measured_high_res = solar
        self.SNR = None

        self.wavelengths = wavelen

        window_size = BlurredSolarSpectrum.window_size
        super().__init__(solar, measured_high_res, fwhm_delta, gaussian_srf,
                         start_wavelen, end_wavelen)
        logging.info('Blurred Solar Spectra (SAO2010) loaded.')
