import os

import sktime.mjd as mjd_conv
import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from sasktran import SolarSpectrum

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.interfaces.solar_prnu import SpsToUsaskSolarPrnuSceneInterface
from altius_l1pp_inflight.scene.disk_geometry import D_1AU
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDataScenario, HmiFileManager, \
    SolarDiskRadianceResult
from altius_l1pp_inflight.scene.generation.solar.interpolate_solar_disk import SimulatedSolarDiskInterpolator
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import VirtualDetectorPixelArray
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import RasterScanGrid
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf
from altius_l1pp_inflight.srf.blurred_spectra import SAO2010
from altius_l1pp_inflight.srf.solar.srf_util import compute_irradiance_center_pixel_scaling_factor
from altius_l1pp_inflight.srf.wavelen_registration import fwhm_2_sigma, fit_measured_spectrum_window_least_squares

"""
Top level class to determine the wavelength registration for an ALTIUS image.
It is assumed that the ALTIUS image has been corrected for all major effects,
though the PRNU correction will have the most recent update.

A group of pixels in the center of the solar disk is used as the measurement for
modelling the solar irradiance.  This is repeated for various locations of the
sun on the detector.

The observed solar spectrum is "corrected" assuming perfect knowledge of the SRF.
The contribution attributed to the side band of the SRF at each operational wavelength
to an ideal solar spectrum is determined, then subtracted from the measured spectrum.
To restore the summation to the proper scale, the result is divided by the area under
the central lobe at each operational wavelength.

From here, a fitting procedure where a Gaussian SRF is adjusted to fit the corrected
spectrum at each measurement wavelength ensues.

This is then repeated for a number of constant wavelength shifts from roughly -1 to +1 nm.
The ideal wavelength shift-FWHM combo is then determined by TBD
"""

# Set the font dictionaries (for plot title and axis titles)
title_font = {'name': 'Arial', 'size': '20', 'color': 'black', 'weight': 'normal',
              'verticalalignment': 'bottom'}  # Bottom vertical alignment for more space
axis_font = {'name': 'Arial', 'size': '18'}
# Set the font properties (for use in legend)
font_path = 'C:\\Windows\\Fonts\\Arial.ttf'
font_prop = font_manager.FontProperties(fname=font_path, size=16)

recompute = True

# TODO simulate several high-res HMI solar disks with granulation noise,
#  each one representing a wavelength measurement, then bin down to
#  ALTIUS resolution and save to pickle file to be used for simulation
#       will need to do this for a couple wavelength shifts and perturbations
#       in the SRF data
#  Select a few center pixels only?  Will just compute the ratio of their
#  total irradiance to the irradiance of the entire disk, and use that ratio
#  to scale the model when running the fit?

srf_wavelens = np.arange(380, 820, step=0.04)
altius_srf = GaussianSrf(srf_wavelens, 2.0 * np.ones_like(srf_wavelens))

config_filename = "globalkey_3x3_grid.yaml"
config = altius_config.AltiusL1ppInflightConfig(config_filename)

home_dir = config.get_processed_wavelength_reg_file_directory()
output_directory = os.path.join(home_dir, 'altius_observation', 'center_detector')

nc_file = 'altius_wavelength_vis_srf_calibration.nc'
nc_file = os.path.join(output_directory, nc_file)


def run_least_squares_stretch_and_shift_fit():
    """
        One of the two main ways we can calibrate the SRF at a chosen wavelength.
        This fit takes a range of wavelengths around some target wavelength and applies
        a least squares fit to the spectrum to obtain an FWHM and wavelength shift
    """
    pass


def compute_altius_images_and_save():
    scene_interface = SpsToUsaskSolarPrnuSceneInterface()

    pixel_dim = config.get_detector_pixel_dimensions('vis', 'native')
    num_pixels_vertical = pixel_dim[0]
    num_pixels_horizontal = pixel_dim[1]
    raster_index = 4  # places the sun at the center
    num_vert_raster = config.get_num_vertical_raster_scans()
    num_horz_raster = config.get_num_horizontal_raster_scans()
    detector_angular_width = config.get_detector_fov('vis')
    maxDN = config.get_max_DN()
    max_radiance = config.get_max_photons_per_sec_per_nm_per_cm2()  # photons/s/cm2/sr/nm
    fov = config.get_detector_fov('vis')
    factor = maxDN / max_radiance
    file_manager = HmiFileManager(config)
    results = file_manager.load_scenario_files_as_images(SolarDataScenario.NO_SUNSPOTS)
    all_scaled_images = []
    for res in results:
        if not isinstance(res, SolarDiskRadianceResult):
            raise AssertionError('Object is not of type HmiSolarDiskRadianceResult')
        all_scaled_images.append(res)
    all_scaled_images = np.array(all_scaled_images)
    solar_disk_interpolator = SimulatedSolarDiskInterpolator(all_scaled_images)
    solar_pixel_radius = results[0].average_solar_radius_in_pixels()
    solar_angular_radius = results[0].angular_radius_degrees()
    solar_scan_grid = RasterScanGrid(solar_angular_radius,
                                     0.1, 0.1, detector_angular_width[0],
                                     detector_angular_width[1],
                                     num_vert_raster, num_horz_raster)
    vdpa = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                     solar_scan_grid=solar_scan_grid,
                                     boresight_horz_portion=1.0,
                                     boresight_vert_portion=1.0)

    WL = np.linspace(500, 510, 10)
    calibration_wavelength = 505

    t0 = solar_disk_interpolator.grid_mjd_times()[0]
    tend = solar_disk_interpolator.grid_mjd_times()[-1]
    TIMES = np.linspace(t0, tend, len(WL))

    boresight = (num_pixels_vertical / 2 - 0.5, num_pixels_horizontal / 2 - 0.5)
    scene_interface.open_file_for_write(nc_file,
                                        calibration_wavelength,
                                        (fov[0], fov[1]),
                                        (num_pixels_vertical, num_pixels_horizontal),
                                        boresight)

    for wavelength, time in zip(WL, TIMES):
        datetime_val = mjd_conv.mjd_to_datetime(time)
        # filepath = os.path.join(output_directory, 'altius_obs_' +
        #                         '{:.2f}'.format(wavelength).replace('.', '_') + '__' +
        #                         datetime_val.strftime('%m_%d_%Y__%H%M%S') + '.pickle')

        print('Computing ALTIUS +scene for ' + '{:.2f}'.format(wavelength) +
              ' nm and ' + datetime_val.strftime('%m_%d_%Y, %H:%M:%S'))
        # +srf, actual_wl = altius_srf.get_srf(wavelength, wavelength_grid, normalize=False)

        solar_image = solar_disk_interpolator.interpolate_disk(datetime_val)

        # photons/s/cm2/nm:
        scaled_image = solar_image.to_scaled_image_result(wavelength=wavelength,
                                                          srf=altius_srf,
                                                          horizontal_pixel_shift=0,
                                                          vertical_pixel_shift=0)
        vdpa.set_current_scanned_image(scaled_image, num_pixels_vertical,
                                       num_pixels_horizontal, raster_index)
        image, image_geometry = vdpa.to_altius_resolution_image(True)

        image_dn = convert_to_DN_and_add_noise(image, factor)

        expected_altius_solar_center_horz = image_geometry['altius_solar_center_horz']
        expected_altius_solar_center_vert = image_geometry['altius_solar_center_vert']
        expected_altius_solar_radius = image_geometry['altius_solar_radius']
        sat_to_sun_distance_km = solar_image.actual_solar_distance()

        scene_interface.write_next_solar_scene_to_netcdf(mjd_conv.mjd_to_datetime64(time),
                                                         image_dn,
                                                         wavelength,
                                                         (expected_altius_solar_center_vert,
                                                          expected_altius_solar_center_horz),
                                                         expected_altius_solar_radius,
                                                         sat_to_sun_distance_km,
                                                         np.array([1, 0, 0]),
                                                         np.array([1, 0, 0]),
                                                         0.1)

    scene_interface.close_write_nc()


if recompute:
    compute_altius_images_and_save()

scene_interface = SpsToUsaskSolarPrnuSceneInterface()
scene_interface.open_read_dataset(nc_file)

n_images = scene_interface.get_num_scenes()

# files = [os.path.join(output_directory, f) for f in os.listdir(output_directory) if os.path.isfile(os.path.join(output_directory, f))]
wavelengths = np.zeros((n_images,))
observed_center_irradiance = np.zeros_like(wavelengths)
observed_total_irradiance = np.zeros_like(wavelengths)

maxDN = config.get_max_DN()
max_photons_s_cm2 = config.get_max_photons_per_sec_per_nm_per_cm2()
dn_to_photons_factor = max_photons_s_cm2 / maxDN

average_solar_distance = 0
solid_angle_solar_disk = 0
solid_angle_center_pixels = 0

pixel_dim = config.get_detector_pixel_dimensions('vis', 'native')
num_pixels_vertical = pixel_dim[0]
num_pixels_horizontal = pixel_dim[1]

h = np.arange(0, num_pixels_horizontal) + 0.5
v = np.arange(0, num_pixels_vertical) + 0.5

hh, vv = np.meshgrid(h, v)
center_radius = 10

average_radius_pixels = 0
average_center_vert = 0
average_center_horz = 0
avg_boresight_vertical = 0
avg_boresight_horizontal = 0

for idx in range(n_images):
    scene = scene_interface.load_scene(idx)
    image = scene.get_altius_image()
    vc = scene.get_expected_altius_solar_center_vert()
    hc = scene.get_expected_altius_solar_center_horz()
    average_solar_distance += scene.get_satellite_to_moon_distance_km() / n_images

    disk_radius = scene.get_expected_altius_solar_radius()
    average_radius_pixels += disk_radius / n_images

    # TODO get these from the center-finding algorithm
    horz_center = scene.get_expected_altius_solar_center_horz()
    vert_center = scene.get_expected_altius_solar_center_vert()

    irrad, _ = scene.compute_irradiance(horz_center, vert_center, dn_to_photons_factor,
                                        center_radius)  # photons/cm2/s/sr/nm

    wavelengths[idx] = scene.get_wavelength()
    observed_center_irradiance[idx] = irrad
    observed_total_irradiance[idx] = np.nansum(image) * dn_to_photons_factor

    boresight = scene.get_boresight()
    avg_boresight_horizontal += boresight[1] / n_images
    avg_boresight_vertical += boresight[0] / n_images
    average_center_vert += vc / n_images
    average_center_horz += hc / n_images

wavelengths = wavelengths.flatten()
observed_center_irradiance = xr.DataArray(observed_center_irradiance.flatten(), dims={'wavelength'},
                                          coords={'wavelength': wavelengths})

# TODO add noise to scenes

solar_spectrum = SolarSpectrum()
hr_wavelengths = np.arange(400, 800, step=0.02)

irradiance_scaling_factor = compute_irradiance_center_pixel_scaling_factor(hr_wavelengths,
                                                                           average_radius_pixels,
                                                                           center_radius,
                                                                           average_solar_distance,
                                                                           average_center_horz,
                                                                           average_center_vert,
                                                                           num_pixels_horizontal,
                                                                           num_pixels_vertical,
                                                                           avg_boresight_horizontal,
                                                                           avg_boresight_vertical)

hr_solar_irradiance = solar_spectrum.irradiance(hr_wavelengths,
                                                solardistance=average_solar_distance / D_1AU)
irradiance_2nm = solar_spectrum.irradiance(hr_wavelengths,
                                           solardistance=average_solar_distance / D_1AU,
                                           fwhm=2.0)

wavelengths = wavelengths.flatten()

plt.figure()
plt.plot(hr_wavelengths, hr_solar_irradiance * irradiance_scaling_factor, alpha=0.3, linewidth=0.8)
plt.plot(wavelengths, observed_center_irradiance)
# plt.plot(wavelengths, observed_side_lobes_removed)
plt.plot(hr_wavelengths, irradiance_2nm * irradiance_scaling_factor)
plt.legend(['HiRes SAO2010', 'Observed Spectrum', 'SAO2010 2 nm Gaussian'])
plt.xlabel('Wavelength (nm)')
plt.ylabel('Irradiance (photons/cm2/sec/nm)')
plt.title('Scaled using Brute Force ')
plt.show()

scaling_factor = xr.DataArray(irradiance_scaling_factor, dims={'wavelength'}, coords={'wavelength': hr_wavelengths})
# perform the least squares fit - all wavelengths are intended for the single analysis wavelength at the center of the window
params = np.r_[fwhm_2_sigma(2.5), 0.0]
sigma = fwhm_2_sigma(np.arange(0.5, 5, step=0.1))
# Important: scaling factor has to be across larger wavelength than fit window
FIT_WINDOW = 10
solar = SAO2010(sigma, start_wavelen=wavelengths[0] - FIT_WINDOW, end_wavelen=wavelengths[-1] + FIT_WINDOW,
                solar_distance_AU=average_solar_distance, scaling_factor=scaling_factor)
fit_results = fit_measured_spectrum_window_least_squares(wavelengths,
                                                         observed_center_irradiance.values,
                                                         params,
                                                         solar)
