import os

import click
import numpy as np
import xarray as xr
from sasktran import SolarSpectrum
from sktime.mjd import mjd_to_datetime64

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.instrument.srf.gaussian_srf import GaussianSrf
from altius_l1pp_inflight.instrument.srf.sinc_squared_srf import AltiusSincSquaredSrf
from altius_l1pp_inflight.instrument.srf.uv_srf import AltiusUVSrf
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry, D_1AU
from altius_l1pp_inflight.scene.generation.solar import calibration_util
from altius_l1pp_inflight.scene.generation.solar.calibration_util import get_solar_srf_calibration_dir, \
    get_solar_srf_simulated_observation_dir
from altius_l1pp_inflight.scene.generation.solar.extract_center_solar_pixels import \
    convert_altius_images_to_srf_input_format
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDataScenario
from altius_l1pp_inflight.scene.interfaces.solar_srf import SpsToUsaskSolarSrfSceneInterface, SolarSrfCalibrationOutput
from altius_l1pp_inflight.srf.blurred_spectra import SAO2010
from altius_l1pp_inflight.srf.solar.srf_util import compute_irradiance_center_pixel_scaling_factor
from altius_l1pp_inflight.srf.wavelen_registration import fit_measured_spectrum_window_least_squares


@click.command()
@click.option('-f', '--config_file', type=str,
              help='Name of configuration yaml file stored in ALTIUS_L1PP_DATABASE/config', default=None)
@click.option('-i', '--input_directory', type=str,
              help='Input directory where solar scenes are saved in the ALTIUS_L1PP_DATABASE directory.  If not'
                   'specified, will be be determined from the config file', default=None)
@click.option('-o', '--output_filename', type=str,
              help='Unique output filename where the results of the SRF calibration are stored '
                   'within the ALTIUS_L1PP_DATABASE directory, determined from config file')
def cli(config_file: str, output_filename: str, input_directory: str = None):
    config = altius_config.AltiusL1ppInflightConfig(config_file)
    run_srf_calibration(config, input_directory, output_filename)


def run_srf_calibration(config: altius_config.AltiusL1ppInflightConfig,
                        output_filename: str, input_directory: str = None):
    """
    Given a directory containing all the images for a given calibration sequence,
    all defined by the format :ref:`singlesolarimageinterface`, performs a calibration on
    the SRF.  A full measurement set consists of all measurement wavelengths, calibration wavelengths,
    and detector positions as defined in the configuration file.

    A calibration wavelength is defined as the central wavelength where we intend to calibrate
    the SRF.  A measurement wavelength is one of the wavelengths surrounding a calibration wavelength
    where we need a sample of the spectrum in order to do a spectral fitting.  Therefore, a measurement
    of the sun is made at this wavelength.  Lastly, a detector position is defined as the position of the
    center of the sun on the detector (in pixel coordinates) where we intend to get a calibration
    at that calibration wavelength.  The motivation behind this is due to the SRF varying with
    position on the detector.

    A full dataset is necessary for the given channel, otherwise the calibration will end prematurely.
    A full dataset consists of all measurement wavelengths for each calibration wavelength and each
    detector position as defined in the configuration file.

    Since the SRF is position-dependent, it is inaccurate to take the entire disk of the sun to compute
    its irradiance.  Instead, only the central pixels of a given radius (as defined in the configuration file)
    are used.  Therefore, for each image, the centre is determined, and the sum of the pixels in DN is converted to
    an irradiance in photons/s/cm2/nm.  Obtaining this for all measurement wavelengths for a single calibration
    wavelength and detector position gives solar irradiance spectrum that is scaled from the full solar disk.

    To compare to a model solar irradiance spectrum, it is necessary to scale the model by the expected amount of
    center pixel irradiance divided by the total irradiance of the sun.  This is accomplished using a model where the
    limb darkening of a noiseless solar disk is simulated at the same resolution as the measurement, and the same
    central pixels are summed in order to obtain a scaling factor for each measurement wavelength.  This is accomplished
    by :ref:`centerscalingfactor`

    The current algorithm uses different models for the AOTF (VIS and NIR) and the FP (UV).  In the UV, a TBD lineshape
    function is employed to be convolved with the SAO2010 solar irradiance spectrum.  This comprises the UV solar model.
    In the VIS and NIR, a TBD modified sinc-squared function is convolved with the solar spectrum to yield the solar model.

    The fitting procedure involves searching the parameter space to minimize the cost function between the
    Gaussian SRF model convolved with the SAO2010 spectrum and the measured irradiance.

    The cost function is

    .. math::
            \sum (log(I_{model}(\lambda + \Delta\lambda, W)) - log(I_{meas}(\lambda)))^2

    where :math:`I_{meas}(\lambda)` is the measured irradiance for wavelength :math:`\lambda`, and
    :math:`I_{model}(\lambda + \Delta\lambda, W)` is the modelled irradiance for a given FWHM value :math:`W`,
    and :math:`\Delta\lambda` is the wavelength shift.  It is TBD if more non-linear terms will be needed
    in calibrating the wavelength grid.  It is also TBD what parameters will be used to describe the lineshape,
    as the lineshape models are TBD.  Presently, the UV

    The cost function is minimized by using a Levenberg-Marquardt stepping approach using the
    :code:`scipy.optimize.least_squares` routine.

    Parameters
    ----------
    config: AltiusL1ppInflightConfig
        Configuration read in from file.

    output_filename: str
        Name of the netcdf file (including .nc extension) summarizing the calibration results.  Results are saved in a single
        netcdf of format :ref:`srfcalibrationoutput`
    input_directory: str, Optional
        Directory where solar files defined by :ref:`singlesolarimageinterface` are placed for a single channel calibration.
        If not specified, it assumes the files are stored at $ALTIUS_L1PP_INFLIGHT_DATABASE$/l1pp_datafiles/srf_registration_processed/observation/
        <scenario>/<channel>/raw, where <scenario> and <channel> are taken from the configuration file

    Returns
    -------
        Output of the calibration script is a netcdf file as defined by :ref:`srfcalibrationoutput`.
        The output file is written to $ALTIUS_L1PP_INFLIGHT$/l1pp_datafiles/srf_registration_processed/calibration/<scenario>/<channel>/
        where channel and scenario are defined in the configuration file.  The netcdf file contains calibration
        information (e.g. FWHM and wavelength shift) for each calibration wavelength and detector position.
    """
    bypass_mode = True

    scenario = SolarDataScenario(config.get_solar_scenario())
    if input_directory is None:
        input_directory = get_solar_srf_simulated_observation_dir(config, scenario.to_string_directory())
    output_directory = get_solar_srf_calibration_dir(config, scenario.to_string_directory())

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    if not bypass_mode:
        measurement_file = 'intermediate_solar_center_pixels.nc'
        convert_altius_images_to_srf_input_format(input_directory,
                                                  config.get_config_source_filename(),
                                                  measurement_file)

    else:
        measurement_file = 'bypass_srf_center_pixels.nc'

    processed_meas_dir = calibration_util.get_solar_srf_intermediate_observation_dir(config,
                                                                                     scenario.to_string_directory())
    nc_in_file = os.path.join(processed_meas_dir, measurement_file)
    nc_out_file = os.path.join(output_directory, output_filename)

    sps_to_usask = SpsToUsaskSolarSrfSceneInterface()
    sps_to_usask.open_read_dataset(nc_in_file)

    all_wavelengths = config.get_srf_measurement_wavelengths(config.get_srf_calibration_channel())
    calibration_wavelengths = config.get_srf_calibration_wavelengths(config.get_srf_calibration_channel())

    detector_positions = config.get_solar_srf_detector_positions()
    center_radius = config.get_srf_center_pixel_radius()

    max_dn = config.get_max_DN()
    max_radiance = config.get_max_photons_per_sec_per_nm_per_cm2()

    dn_to_photons_factor = max_radiance / max_dn

    channel = config.get_srf_calibration_channel()
    mode = config.get_srf_detector_mode()
    pixel_dim = config.get_detector_pixel_dimensions(channel, mode)
    fov = config.get_detector_fov(channel)
    num_pixels_vertical = pixel_dim[0]
    num_pixels_horizontal = pixel_dim[1]
    # h = np.arange(0, num_pixels_horizontal) + 0.5
    # v = np.arange(0, num_pixels_vertical) + 0.5
    # # vv, hh = np.meshgrid(h, v)

    bv, bh = config.get_pixel_boresights(channel, mode)
    solar_spectrum = SolarSpectrum()

    calibration_output = SolarSrfCalibrationOutput(nc_out_file, calibration_wavelengths,
                                                   detector_positions,
                                                   (bv, bh),
                                                   (num_pixels_vertical, num_pixels_horizontal))

    fit_method = 'scale_srf_fwhm'  # 'constant_gaussian'  # 'scale_srf_fwhm'

    if channel == 'uv':
        srf_dir = os.path.join(config.get_srf_directory(), 'fpi')
        altius_srf = AltiusUVSrf(srf_dir, fwhm_method_using_area=True, normalize=False)

        if fit_method == 'scale_srf_fwhm':
            fwhms = []
            for wl in altius_srf.tunable_wavelengths():
                fit_fwhm_deltas, _ = altius_srf.compute_srf_approximate_fwhm(wl, altius_srf.get_wavelength_grid())
                fwhms.append(fit_fwhm_deltas)

            gaussian_srf = GaussianSrf(altius_srf.tunable_wavelengths(), np.asarray(fwhms))

            import matplotlib.pyplot as plt
            plt.figure()
            hires_wl = np.arange(250, 400, step=0.001)
            sao = SolarSpectrum()
            wavelen = np.arange(300, 340, step=0.1)
            plt.plot(hires_wl, sao.irradiance(hires_wl), alpha=0.5, linewidth=0.5)
            plt.plot(wavelen, sao.irradiance(wavelen, fwhm=3.0))
            gaus_conv_spect = gaussian_srf.convolve_spectrum_with_srf(wavelen, hires_wl,
                                                                      sao.irradiance(hires_wl)).squeeze()
            plt.plot(wavelen, gaus_conv_spect)
            conv_spect = altius_srf.convolve_spectrum_with_srf(wavelen, hires_wl, sao.irradiance(hires_wl)).squeeze()
            plt.plot(wavelen, conv_spect)
            plt.legend(['Full resolution', 'FWHM = 3.0 nm', 'Varying Gaussian', 'UV SRF'])
            plt.xlabel('Wavelength (nm)')
            plt.ylabel('Irradiance')
            plt.title('Comparison of UV SRF')

        elif fit_method == 'constant_gaussian':
            gaussian_srf = GaussianSrf(np.arange(200, 400), 2.5 * np.ones_like(np.arange(200, 400)))
        else:
            gaussian_srf = None
    else:
        grid_wavelengths = np.arange(400, 1200, step=0.5)
        grid_scale = 1.5 * np.ones_like(grid_wavelengths)
        altius_srf = AltiusSincSquaredSrf(grid_wavelengths, grid_scale)

    obs_irrad = []

    for detector_position in detector_positions:

        fit_fwhm_deltas = []
        shifts = []
        scales = []
        first_through = True
        labels = []
        for measurement_wavelengths, calibration_wavelength in zip(all_wavelengths, calibration_wavelengths):

            average_solar_distance = 0

            avg_time = 0
            n_images = len(measurement_wavelengths)

            wavelengths = np.zeros((n_images,))
            observed_irradiance = np.zeros_like(wavelengths)
            solar_geometry = SolarGeometry(num_pixels_vertical, num_pixels_horizontal, bv, bh, fov[0], fov[1])

            for idx, wl in enumerate(measurement_wavelengths):
                scene = sps_to_usask.load_scene(detector_position, calibration_wavelength, wl)
                average_solar_distance += scene.get_satellite_to_sun_distance_km() / n_images

                solar_geometry.set_disk(scene.get_expected_altius_solar_radius(),
                                        scene.get_expected_altius_solar_center_horz(),
                                        scene.get_expected_altius_solar_center_vert(),
                                        scene.get_satellite_to_sun_distance_km())

                # TODO apply centering to find the vertical and horizontal center components

                wavelengths[idx] = scene.get_wavelength()
                observed_irradiance[idx] = scene.get_altius_image() * dn_to_photons_factor

                avg_time += scene.get_time() / n_images

            wavelengths = wavelengths.flatten()
            observed_irradiance = xr.DataArray(observed_irradiance.flatten(), dims={'wavelength'},
                                               coords={'wavelength': wavelengths})

            obs_irrad.append(observed_irradiance)

            w0 = 250  # np.min(measurement_wavelengths)
            w1 = 380  # np.max(measurement_wavelengths)
            hr_wavelengths = np.arange(w0 - 20, w1 + 20, step=0.01)

            irradiance_scaling_factor = compute_irradiance_center_pixel_scaling_factor(hr_wavelengths,
                                                                                       center_radius,
                                                                                       solar_geometry)

            hr_solar_irradiance = solar_spectrum.irradiance(hr_wavelengths,
                                                            solardistance=average_solar_distance / D_1AU)
            # hr_solar_irradiance /= (hr_wavelengths[1] - hr_wavelengths[0]) # photons/cm2/sec/nm
            # irradiance_3nm = solar_spectrum.irradiance(hr_wavelengths,
            #                                            solardistance=average_solar_distance / D_1AU,
            #                                            fwhm=3.0)

            hr_solar_irradiance *= irradiance_scaling_factor
            # irradiance_3nm *= irradiance_scaling_factor

            wavelengths = wavelengths.flatten()

            if channel == 'uv':
                #     observed_side_lobes_removed, central_lobe_portion_corr, central_lobe_fwhm_corr, srf_central_lobe_sum, srf_side_lobe_sum = \
                #         altius_srf.convolve_corrected_srf(hr_wavelengths, hr_solar_irradiance, wavelengths,
                #                                           observed_irradiance)  # photons/cm2/sec/sr
                #
                #     fit_delta_fwhm, rel_percent_diff, fit_spectrum = run_fwhm_fit(wavelengths, observed_side_lobes_removed,
                #                                                             irradiance_scaling_factor, hr_wavelengths,
                #                                                             average_solar_distance, altius_srf)
                #
                #     fit_shift = 0
                # else:
                scaling_factor = xr.DataArray(irradiance_scaling_factor, dims={'wavelength'},
                                              coords={'wavelength': hr_wavelengths})
                # perform the least squares fit - all wavelengths are intended for the single analysis wavelength at the center of the window
                params = np.r_[-0.2, 0.0, 1.0]  # fit_fwhm_deltas, shift, scale
                # sigma = fwhm_2_sigma(np.arange(0.5, 5, step=0.1))
                fwhm_delta = np.arange(-1, 1, step=0.1)
                # Important: scaling factor has to be across larger wavelength than fit window
                FIT_WINDOW = config.get_wavelength_fitting_band()
                solar = SAO2010(fwhm_delta,
                                gaussian_srf=gaussian_srf,
                                start_wavelen=calibration_wavelength - 2.5 * FIT_WINDOW,
                                end_wavelen=calibration_wavelength + 2.5 * FIT_WINDOW,
                                solar_distance_AU=average_solar_distance / D_1AU,
                                scaling_factor=scaling_factor)
                fit_results = fit_measured_spectrum_window_least_squares(wavelengths,
                                                                         observed_irradiance.values,
                                                                         params,
                                                                         solar)
                fit_delta_fwhm = fit_results.x[0]  # TODO add to model initial FWHM
                fit_shift = fit_results.x[1]
                fit_scale = fit_results.x[2]

            fit_fwhm_deltas.append(fit_delta_fwhm)
            shifts.append(fit_shift)
            scales.append(fit_scale)

            calibration_output.write_current_calibration_results(fwhm=fit_delta_fwhm, wavelength_shift=fit_shift,
                                                                 time=mjd_to_datetime64(avg_time),
                                                                 detector_position=detector_position,
                                                                 calibration_wavelength=calibration_wavelength)
            # sim_spectrum = SimulatedUVSolarSpectrum(altius_srf, None,
            #                                         solar_distance_AU=average_solar_distance / D_1AU)  # trying to fit this with a shifted and blurred solar spectrum with only Gaussian

        plt.figure()
        irradiance_3nm = solar_spectrum.irradiance(hr_wavelengths,
                                                   solardistance=average_solar_distance / D_1AU,
                                                   fwhm=3.0)
        irradiance_3nm *= irradiance_scaling_factor
        plt.plot(hr_wavelengths, irradiance_3nm, linestyle='dashed')
        labels.append('3 nm Gaussian-convolved Spectrum')

        for oi in obs_irrad:
            plt.plot(oi.wavelength, oi.values, linewidth=3.0)

        plt.title('Simulated Spectra Compared with Measured Spectra')

        plt.figure()
        fit_fwhms = fit_fwhm_deltas + gaussian_srf.get_fwhm(calibration_wavelengths)

        plt.plot(calibration_wavelengths, fit_fwhms)
        actual_fwhm = altius_srf.get_fwhm(calibration_wavelengths)  # 3.0*np.ones_like(calibration_wavelengths)  #
        plt.plot(calibration_wavelengths, actual_fwhm)
        plt.title('FWHM Retrieval for a Constant Width Gaussian')
        # plt.plot(calibration_wavelengths, actual_fwhm)
        plt.plot(calibration_wavelengths, 1.05 * actual_fwhm, color='r', linestyle='dashed')
        plt.plot(calibration_wavelengths, 0.95 * actual_fwhm, color='r', linestyle='dashed')
        plt.ylim(2.8, 3.2)
        plt.legend(['Calibrated FWHM', 'Measured FWHM'])

        plt.figure()
        plt.plot(calibration_wavelengths, shifts)
        plt.xlabel('Calibration Wavelength, nm')
        plt.ylabel('Peak Shift, nm')
        plt.title('Central Wavelength Shift Retrieval')
        plt.show()

        # plt.figure()
        # plt.plot(calibration_wavelengths, scales)
        # plt.xlabel('Calibration Wavelength, nm')
        # plt.ylabel('Scaling Factor, unitless')
        # plt.title('Scaling Factor Retrieval')
        # print('Calculated shift: ', shifts)
        # print('Calculated scales: ', scales)


if __name__ == '__main__':
    run_srf_calibration(altius_config.AltiusL1ppInflightConfig(), 'out_srf_cal_file.nc')
