import numpy as np

# TODO can improve on the structure of this function
from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977


def compute_irradiance_center_pixel_scaling_factor(wavelengths: np.ndarray,
                                                   center_radius_pixels: float,
                                                   solar_geometry: SolarGeometry):
    """
        Determines the ratio of center pixel irradiance to the total solar irradiance.  The center
        pixels are defined by the circle of pixels of radius :code:`center_radius_pixels` surrounding
        the solar disk center on the detector.  This is computed at all :code:`wavelengths` by discretely
        computing it at key wavelengths and interpolating to the array :code:`wavelengths`.

        Passes in the dimensions of the detector to mimic the measurement - we want the exact
        same set of pixels in the center to be used as in the measurement.  The function creates a detector
        array and a solar disk centered at the same location and with the same radius as defined in
        :code:`solar_geometry`.  The solar disk sums to unity irradiance, and has the wavelength-dependent
        limb darkening applied to it.  The ratio is then simply the sum of the pixels making a small circle of
        radius :code:`center_radius_pixels` in the center.

        Parameters
        ----------
        wavelengths: np.ndarray
            Array of wavelengths, in nm, where the ratio needs to be calculated.  The actual calculations/simulations
            of the disk are performed at intervals of 2 nm within the range of wavelengths, then interpolated.

        center_radius_pixels: float
            Radius of pixels to be used in the center for computing the ratio.

        solar_geometry: SolarGeometry
            Solar geometry object defining the physical distance, detector position and apparent
            radius of the solar disk, in addition to auxilliary detector parameters.

        Returns
        -------
        irradiance_scaling_factor: np.ndarray
            Array of scaling factors at each of :code:`wavelengths`.

    """
    sample_hr_wl = np.arange(wavelengths[0] - 10, wavelengths[-1] + 10, step=2)
    irradiance_scaling_factor = np.zeros_like(sample_hr_wl)

    mu = solar_geometry.mu(include_padding=False)
    mask = solar_geometry.center_pixel_mask(center_radius_pixels)

    # Correction for off-axis viewing
    delta_omega_view = solar_geometry.delta_omega_view(include_padding=False)
    delta_omega_center = solar_geometry.delta_omega_center(include_padding=False)

    for idx, wl in enumerate(sample_hr_wl):
        irradiance_scaling_factor[idx] = _compute_irradiance_center_pixel_scaling_factor_wl(wl, mu,
                                                                                            delta_omega_center,
                                                                                            delta_omega_view,
                                                                                            mask)
    irradiance_scaling_factor = np.interp(wavelengths, sample_hr_wl, irradiance_scaling_factor)
    return irradiance_scaling_factor


def _compute_irradiance_center_pixel_scaling_factor_wl(wavelength: float,
                                                       mu: np.ndarray,
                                                       delta_omega_center: np.ndarray,
                                                       delta_omega_view: np.ndarray,
                                                       mask: np.ndarray):
    limb_darkening_model = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1, 50)

    ld_profile = limb_darkening_model.limb_darkening(wavelength, mu_profile)
    ld = np.interp(mu, mu_profile, ld_profile)
    ld *= delta_omega_center / delta_omega_view
    irradiance_scaling_factor = np.nansum(ld[mask]) / np.nansum(ld)

    return irradiance_scaling_factor
