import os

import numpy as np
import progressbar
import xarray as xr
from sasktran.solarspectrum import SolarSpectrum
# import logging
from scipy.optimize import least_squares

from altius_l1pp_inflight.instrument.srf.uv_srf import AltiusUVSrf, FwhmPerturbations
from altius_l1pp_inflight.srf.blurred_spectra import SAO2010
import matplotlib.pyplot as plt

script_dir = os.path.dirname(__file__)
# rel_path = '../../../altius_l1pp_sphinx_documentation/sphinx/source/special_studies/2020-06-15_wavelength_registration/sphinx/figures/uniform_run/'
# SAVE_DIR = os.path.join(script_dir, rel_path)
FWHM_CONVERT = 2 * np.sqrt(2 * np.log(2))


def fwhm_2_sigma(fwhm):
    return fwhm / FWHM_CONVERT


def sigma_2_fwhm(sigma):
    return FWHM_CONVERT * sigma


spectrum_iterations = []
parameter_iterations = []


def solar_model(fwhm_shift, mu, wv, scale, solar):
    """
    Returns the log of the solar irradiance spectrum for the array of shifted wavelengths,
    defined by :code:`wv - mu`.  The solar irradiance is determined by convolving the high resolution
    solar spectrum with a varying Gaussian SRF, contained in the parameter :code:`solar`.

    :param fwhm_shift:
        Shift to be applied to the FWHM's of all SRF's within the fitting band (to account for the fact
        that SRF's vary in FWHM across a band).
    :param mu:
        Shift in wavelength to be applied uniformly across the band, in nm.
    :param wv:
        Array of wavelengths for the fitting band, in nm
    :param scale:
        Scaling parameter for the SRF, generally not used and defaulted to 1.0
    :param solar:
        Solar model which has precomputed solar spectra for a range of FWHM deltas.  The solar model
        takes in a Gaussian model that can vary in FWHM across wavelength, so it interpolates to a
        given FWHM perturbation.

    :return:
        Logarithm of the resulting solar irradiance spectrum, parallel array to :code:`wv`
    """
    global spectrum_iterations
    global parameter_iterations
    shifted_wavelen = wv - mu
    solar_spectrum = scale * solar(fwhm_shift, shifted_wavelen)
    spectrum_iterations.append((shifted_wavelen, solar_spectrum.squeeze()))
    parameter_iterations.append((fwhm_shift, mu, scale))
    p = np.polyfit(shifted_wavelen, solar_spectrum.squeeze(), 5)
    return np.log(solar_spectrum.squeeze()), p  # - np.log(poly_spectrum(shifted_wavelen, p)), p
    # return solar_spectrum.squeeze() - poly_spectrum(wv, p)


def poly_spectrum(w, p):
    return p[5] + p[4] * w + p[3] * w ** 2 + p[2] * w ** 3 + p[1] * w ** 4 + p[0] * w ** 5


def errfunc(params, wavelen, measured_spectrum, solar):
    """
    The error function used in evaluating the least squares and determining the optimal
    model parameters.  The difference between the measured spectrum and the solar model's spectrum
    for a particular set of :code:`params` gives the error function.  The least squares works by
    minimizing this value through intelligent searching of the model parameter space.

    :param params:
        Set of model parameters for the calibration - currently wavelength shift and FWHM delta.
    :param wavelen:
        The array of wavelengths where the solar spectrum is measured/computed.
    :param measured_spectrum:
        The measured solar spectrum the model is trying to fit to.
    :param solar:
        Solar model that computes the solar spectrum based on the parameters passed in.
    """
    fwhm_shift = params[0]
    shift = params[1]
    scale = 1.0  # params[2]
    model_guess, p = solar_model(fwhm_shift, shift, wavelen, scale, solar)
    log_poly_spectrum = np.log(poly_spectrum(wavelen, p))
    differential_measured_spectrum = np.log(measured_spectrum)  # - log_poly_spectrum
    # differential_measured_spectrum = simulated_measurement_spectrum - poly_spectrum(wavelen, p)
    # plt.figure()
    # plt.plot(wavelen-shift, np.exp(model_guess+np.log(poly_spectrum(wavelen - shift, p))))
    # plt.plot(wavelen, measured_spectrum, linestyle='dashed')
    # plt.legend(['model', 'measured'])
    # plt.title('FWHM delta: {:.4f}, Shift: {:.4f}'.format(fwhm_shift, shift))
    # plt.show()
    return model_guess - differential_measured_spectrum


def make_gaussian(center_wavelength, fwhm, wavelengths):
    sigma = fwhm_2_sigma(fwhm)
    g = np.exp(-0.5 * (wavelengths - center_wavelength) ** 2 / sigma ** 2)
    g /= np.sum(g)
    return g
