import unittest

import numpy as np

from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import VirtualDetectorPixelArray, \
    pixels_to_distance_km
from altius_l1pp_inflight.scene.generation.solar.solar_scan_grid import RasterScanGrid


def raster_grid_test():
    solar_distance = 147.35e6
    solar_pixel_radius = 1954
    rg = RasterScanGrid(solar_radius_angular_size_deg=0.264,
                        scan_rate_deg_per_sec=1.0,
                        time_per_image_sec=1.0,
                        fov_horizontal_deg=2.0,
                        fov_vertical_deg=2.2,
                        v_in=8,
                        h_in=8)

    vp = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                   solar_scan_grid=rg,
                                   boresight_horz_portion=0.0,
                                   boresight_vert_portion=0.0)

    dh = rg.grid_horizontal_step_in_pixels(solar_pixel_radius)
    dv = rg.grid_vertical_step_in_pixels(solar_pixel_radius)

    icf_detector_corners = vp.get_icf_corners(solar_distance)

    for idx in range(0, rg._num_images):
        sun_horz_pix, sun_vert_pix = vp.get_solar_disk_detector_pixel_shift_from_boresight(idx)


def icf_corners_test():
    solar_pixel_radius = 2000
    rg = RasterScanGrid(solar_radius_angular_size_deg=0.26,
                        scan_rate_deg_per_sec=1.0,
                        time_per_image_sec=1.0,
                        fov_horizontal_deg=2.0,
                        fov_vertical_deg=2.2,
                        v_in=8,
                        h_in=8)
    bhp = 0.04
    bvp = -0.02
    vp = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                   solar_scan_grid=rg,
                                   boresight_horz_portion=bhp,
                                   boresight_vert_portion=bvp)

    solar_distance = 147.35e6

    icf_detector_corners = vp.get_icf_corners(solar_distance)

    grid_step_horz = vp._grid_step_horz_px_size
    grid_step_vert = vp._grid_step_vert_px_size

    assert (vp._num_rows == 11)
    assert (vp._num_cols == 10)
    assert (np.isclose(vp._num_horz_virtual_px, grid_step_horz * vp._num_cols))
    assert (np.isclose(vp._num_vert_virtual_px, grid_step_vert * vp._num_rows))
    assert (np.isclose(grid_step_horz, vp._num_horz_virtual_px / vp._num_cols))
    assert (np.isclose(grid_step_vert, vp._num_vert_virtual_px / vp._num_rows))
    assert (np.isclose(vp._boresight_virtual_px_horz,
                       (vp._num_horz_virtual_px / 2 - grid_step_horz) * (1 - bhp)) + grid_step_horz)
    assert (np.isclose(vp._boresight_virtual_px_vert,
                       (vp._num_vert_virtual_px / 2 - grid_step_vert) * (1 + bvp)) + grid_step_vert)

    # compute the azimuth and elevation of the expected ICF corners:
    x = solar_distance
    yl = pixels_to_distance_km(vp._boresight_virtual_px_horz - vp._grid_step_horz_px_size,
                               solar_pixel_radius)
    yr = pixels_to_distance_km(
        vp._boresight_virtual_px_horz - vp._num_horz_virtual_px + vp._grid_step_horz_px_size,
        solar_pixel_radius)
    zu = pixels_to_distance_km(
        vp._num_vert_virtual_px - vp._grid_step_vert_px_size - vp._boresight_virtual_px_vert,
        solar_pixel_radius)
    zb = pixels_to_distance_km(vp._grid_step_vert_px_size - vp._boresight_virtual_px_vert, solar_pixel_radius)

    left_az = np.arctan(yl / x)
    right_az = np.arctan(yr / x)

    tl_elev = np.arctan(zu / np.sqrt(x ** 2 + yl ** 2))
    tr_elev = np.arctan(zu / np.sqrt(x ** 2 + yr ** 2))
    bl_elev = np.arctan(zb / np.sqrt(x ** 2 + yl ** 2))
    br_elev = np.arctan(zb / np.sqrt(x ** 2 + yr ** 2))

    expected = []
    for az, el in zip([left_az, right_az, left_az, right_az], [tl_elev, tr_elev, bl_elev, br_elev]):
        z = np.sin(el)
        y = np.sin(az) * np.cos(el)
        x = np.sqrt(1 - y ** 2 - z ** 2)
        expected.append([x, y, z])

    assert (compare_corner_vectors(expected, icf_detector_corners.T))

    print('ICF corner test passed')


def compare_corner_vectors(expected, actual):
    all_corner_match = []
    for ie, e in enumerate(expected):
        found_match = False
        for ia, a in enumerate(actual):
            agrees = np.isclose(e, a, rtol=1e-5)
            if agrees.all():
                print()
                print('Expected corner ', ie, ' agrees with actual corner ', ia)
                print('Expected: ', e)
                print('Actual: ', a)
                print()
                found_match = True
        if not found_match:
            print('Did not find close match for expected corner: ', e)
        all_corner_match.append(found_match)

    return np.all(all_corner_match)


class RasterGridTest(unittest.TestCase):

    def test_raster_grid_positions(self):
        raster_grid_test()

    def test_icf_corners(self):
        icf_corners_test()


if __name__ == '__main__':
    icf_corners_test()
