import unittest

import numpy as np
import sktime.mjd as mjd_conv

import altius_l1pp_inflight.scene.generation.solar.hmi_files as hmi_files
import altius_l1pp_inflight.scene.generation.solar.interpolate_solar_disk as isd
from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977


def test_simulated_noise(hmi_data, sim_results):
    # Just check if the mean and standard deviation of the central box of
    # each image are reasonably close to the actual data
    ld_model = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1, 50)
    ld_profile = ld_model.limb_darkening(hmi_data[0].wavelength(), mu_profile)

    means = []
    stds = []

    for hmi, sim in zip(hmi_data, sim_results):
        # Data:
        hc, vc = hmi.center_pixel()
        r = hmi.average_solar_radius_in_pixels()
        w = r / 2
        # Apply limb darkening correction
        mu = hmi.mu_solar_disk()
        lv = int(vc - w)
        uv = int(vc + w)
        lh = int(hc - w)
        uh = int(hc + w)
        ld = np.interp(mu[lv:uv, lh:uh], mu_profile, ld_profile)
        fl_hmi = hmi.radiance()[lv:uv, lh:uh] / ld
        mean_hmi = np.nanmean(fl_hmi)
        std_dev_hmi = np.nanstd(fl_hmi)

        # Simulation:
        hc, vc = sim.center_pixel()
        r = sim.average_solar_radius_in_pixels()
        w = int(r / 2)
        lv = int(vc - w)
        uv = int(vc + w)
        lh = int(hc - w)
        uh = int(hc + w)
        mu = sim.mu_solar_disk()
        ld = np.interp(mu[lv:uv, lh:uh], mu_profile, ld_profile)
        fl_sim = sim.radiance()[lv:uv, lh:uh] / ld
        mean_sim = np.nanmean(fl_sim)
        std_dev_sim = np.nanstd(fl_sim)

        means.append((mean_hmi, mean_sim))
        stds.append((std_dev_hmi, std_dev_sim))

    return means, stds


class TestSolarDiskInterpolator(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestSolarDiskInterpolator, cls).setUpClass()
        config = altius_config.AltiusL1ppInflightConfig(test_config=True)
        filemanager = hmi_files.HmiFileManager(config)
        cls.hmi_data = filemanager.load_scenario_files_as_images(hmi_files.SolarDataScenario.SEVERAL_SUNSPOTS)
        cls.interpolator = isd.SimulatedSolarDiskInterpolator(cls.hmi_data)
        cls.mjd_times = cls.interpolator.grid_mjd_times()
        cls.sim_results = cls.interpolator.get_sim_ideal_grid_images()

    def test_simulated_granulation_noise(self):
        means, stds = test_simulated_noise(self.hmi_data, self.sim_results)

        for m, s in zip(means, stds):
            self.assertAlmostEqual(m[0], m[1], delta=0.2 * m[0],
                                   msg='Mean of simulated image {} is >5% from actual image {}'.format(m[1], m[0]))
            self.assertAlmostEqual(s[0], s[1], delta=0.2 * s[0],
                                   msg='Standard deviation of simulated image {} is >5% from actual image {}'.format(
                                       s[1], s[0]))

    def test_solar_disk_interpolation(self):
        time = mjd_conv.mjd_to_datetime((self.mjd_times[0] + self.mjd_times[1]) / 2)

        disk1 = self.sim_results[0].radiance()
        disk2 = self.sim_results[1].radiance()

        interp_disk = self.interpolator.interpolate_disk(time).radiance()

        isclose_array = np.isclose((disk1 + disk2) / 2, interp_disk, rtol=0.001)

        import matplotlib.pyplot as plt
        plt.figure()
        plt.imshow(disk1)
        plt.figure()
        plt.imshow(interp_disk)
        plt.figure()
        plt.imshow(disk2)

        # only worry about the interior pixels:
        R = np.sqrt(self.interpolator._DH ** 2 + self.interpolator._DV ** 2)
        mask = R < 0.9 * self.sim_results[0].average_solar_radius_in_pixels()

        self.assertTrue(np.all(isclose_array[mask]),
                        msg='Some interpolated pixels were more than 0.1% from expected linear interpolated value')


if __name__ == '__main__':
    unittest.main()
