import unittest

import numpy as np

from altius_l1pp_inflight.scene.disk_geometry import center_solar_disk
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977


def center_disk_partially_off_detector(radius, center_x, center_y):
    print('*********************************')
    print('Performing ideal ALTIUS-resolution (1504x1504) solar +scene image')
    print(' with sun partially off detector')
    print('Tolerance is expected to be within 4.0 pixels')
    print('*********************************')
    print()

    # ALTIUS:
    image = create_ideal_solar_image(1504, 1504, center_x, center_y, radius, 335)

    calc_radius, x_c, y_c, _ = center_solar_disk(image, **{'detector_angular_width_degrees': (2.0, 2.0),
                                                           'distance_to_sun_km': 147.24e6})

    return calc_radius, x_c, y_c


def centroid_arbitrary_disk(radius, center_x, center_y):
    print('*********************************')
    print('Performing ideal ALTIUS-resolution (1504x1504) solar +scene image')
    print('Tolerance is expected to be within 3.0 pixels')
    print('*********************************')
    print()

    # ALTIUS:
    image = create_ideal_solar_image(1504, 1504, center_x, center_y, radius, 335)

    calc_radius, x_c, y_c, _ = center_solar_disk(image)

    return calc_radius, x_c, y_c


def create_ideal_solar_image(hsize, vsize, h_center, v_center, radius, wavelen, snr=1000, binning=1):
    # create a disk, apply limb darkening, then add reasonable noise
    high_dn = 14000 * binning
    image = np.zeros((vsize, hsize))

    X, Y = np.meshgrid(np.arange(0, hsize), np.arange(0, vsize))
    R = np.sqrt((X - h_center) ** 2 + (Y - v_center) ** 2)
    mu = np.sqrt(1 - (R / radius) ** 2)
    solar_disk_idx = R <= radius
    image[solar_disk_idx] = high_dn
    darkening_model = LimbDarkening_PS1977()
    limb_profile = darkening_model.limb_darkening(wavelen, np.linspace(0, 1, 100))
    scale = np.interp(mu, np.linspace(0, 1, 100), limb_profile)
    image *= scale
    image[R > radius] = high_dn / snr
    image = np.random.poisson(image * 2, np.shape(image)) / 2 / binning
    return image


class DiskCentering(unittest.TestCase):

    def test_sun_partially_off_detector(self):
        radius = 207
        center_x = 405
        center_y = 1104
        calc_radius, x_c, y_c = center_disk_partially_off_detector(radius, center_x, center_y)
        self.assertAlmostEqual(calc_radius, radius, delta=4.0)
        self.assertAlmostEqual(x_c, center_x, delta=4.0)
        self.assertAlmostEqual(y_c, center_y, delta=4.0)

    def test_sun_fully_on_detector(self):
        radius = 207
        center_x = 405
        center_y = 1104

        calc_radius, x_c, y_c = centroid_arbitrary_disk(radius, center_x, center_y)

        # need better than 1 pixel at ALTIUS resolution
        self.assertAlmostEqual(calc_radius, radius, delta=3.0)
        self.assertAlmostEqual(x_c, center_x, delta=3.0)
        self.assertAlmostEqual(y_c, center_y, delta=3.0)
