import os
import unittest

import numpy as np
import sktime.mjd as mjd_conv
import xarray as xr

from altius_l1pp_inflight.scene.interfaces.lunar_prnu import SpsToUsaskLunarPrnuSceneInterface, LunarPrnuSpotCheckOutput
from altius_l1pp_inflight.scene.disk_geometry import D_1AU, EARTH_TO_MOON_DISTANCE_KM


class TestLunarInterfaces(unittest.TestCase):

    def test_lunar_sps_scene_interface(self):
        scene = SpsToUsaskLunarPrnuSceneInterface()
        nv = 60
        nh = 60
        wavelengths = np.array([400, 450, 500, 550])
        positions = np.array([[10, 20], [10, 40],
                              [30, 20], [30, 40],
                              [50, 20], [50, 40]])

        scene.open_file_for_write('temp_scene.nc',
                                  wavelengths,
                                  (2.0, 2.0),
                                  (nv, nh),
                                  (nv / 2, nh / 2),
                                  positions)

        r = 7.5
        dv, dh = np.meshgrid(np.arange(0, nv), np.arange(0, nh))

        expected_scenes = np.zeros((len(positions), len(wavelengths), nv, nh))
        expected_sat_posn = np.zeros((len(positions), len(wavelengths), 3))
        expected_start_bs = np.copy(expected_sat_posn)
        expected_end_bs = np.copy(expected_sat_posn)
        expected_integration_times = np.zeros((len(positions), len(wavelengths)))
        expected_subsol_lon = np.copy(expected_integration_times)
        expected_subsat_lat = np.copy(expected_integration_times)
        expected_subsat_lon = np.copy(expected_integration_times)
        expected_phase = np.copy(expected_integration_times)
        expected_times = np.zeros_like(expected_integration_times)
        suntomoon = D_1AU * np.ones_like(expected_integration_times)
        sattomoon = EARTH_TO_MOON_DISTANCE_KM * np.ones_like(suntomoon)
        expected_centers = np.zeros((len(positions), len(wavelengths), 2))

        for wl_idx, wl in enumerate(wavelengths):
            for pos_idx, pos in enumerate(positions):
                sat_posn = np.random.random((3,))
                start_bs = np.random.random((3,))
                expected_start_bs[pos_idx, wl_idx] = start_bs
                end_bs = np.random.random((3,))
                expected_end_bs[pos_idx, wl_idx] = end_bs
                t = np.random.rand() * 1.5
                expected_integration_times[pos_idx, wl_idx] = t
                subsol_lon = np.random.rand() * 5.0
                expected_subsol_lon[pos_idx, wl_idx] = subsol_lon
                subsat_lat = np.random.rand() * 5.0
                expected_subsat_lat[pos_idx, wl_idx] = subsat_lat
                subsat_lon = np.random.rand() * 5.0
                expected_subsat_lon[pos_idx, wl_idx] = subsat_lon
                phase = 7.0 + np.random.rand()
                expected_phase[pos_idx, wl_idx] = phase

                wavelen = wl + np.random.rand()
                posn = pos + np.random.rand()
                expected_centers[pos_idx, wl_idx] = posn

                R = np.sqrt((dv - posn[0]) ** 2 + (dh - posn[1]) ** 2)
                next_scene = R <= r
                next_scene = next_scene.astype(int) * np.random.randint(255)
                expected_scenes[pos_idx, wl_idx] = next_scene

                expected_times[pos_idx, wl_idx] = mjd_conv.datetime64_to_mjd(np.datetime64('2019-11-14'))

                scene.write_next_lunar_scene_to_netcdf(np.datetime64('2019-11-14'),
                                                       next_scene,
                                                       wavelen, sat_posn,
                                                       posn,
                                                       r,
                                                       EARTH_TO_MOON_DISTANCE_KM,
                                                       start_bs, end_bs,
                                                       t, subsol_lon, subsat_lat,
                                                       subsat_lon, D_1AU, phase)
        scene.close_write_nc()

        ds = xr.open_dataset('temp_scene.nc')
        vals = ds.data_vars.variables
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS].values.tolist(),
                             wavelengths.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.SUN_TO_MOON_DISTANCE_AU].values.tolist(),
                             suntomoon.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.SAT_TO_MOON_DISTANCE].values.tolist(),
                             sattomoon.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.DETECTOR_POSITIONS].values.tolist(),
                             positions.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.CALIBRATION_WAVELENGTHS].values.tolist(),
                             wavelengths.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.TIMES].values.tolist(),
                             expected_times.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.EXPECTED_LUNAR_CENTER].values.tolist(),
                             expected_centers.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.SUBSOLAR_SELEN_LONGITUDE].values.tolist(),
                             expected_subsol_lon.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.SUBSAT_SELEN_LATITUDE].values.tolist(),
                             expected_subsat_lat.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.SUBSAT_SELEN_LONGITUDE].values.tolist(),
                             expected_subsat_lon.tolist())
        self.assertListEqual(vals[SpsToUsaskLunarPrnuSceneInterface.LUNAR_PHASE_ANGLE_DEG].values.tolist(),
                             expected_phase.tolist())

        ds.close()
        os.remove('temp_scene.nc')

    def test_lunar_prnu_output_interface(self):
        nv = 60
        nh = 60
        wavelengths = np.array([400, 450, 500, 550])
        positions = np.array([[10, 20], [10, 40],
                              [30, 20], [30, 40],
                              [50, 20], [50, 40]])
        prnu = LunarPrnuSpotCheckOutput('temp.nc', nv, nh,
                                        wavelengths, positions)

        expected_ratios = np.zeros((len(wavelengths), len(positions)))
        expected_errors = np.zeros((len(wavelengths), len(positions)))
        expected_mask = np.zeros((len(wavelengths), len(positions), nv, nh))

        r = 5
        dv, dh = np.meshgrid(np.arange(0, nv), np.arange(0, nh))

        for wl_idx, wl in enumerate(wavelengths):
            for pos_idx, pos in enumerate(positions):
                # make sure we can add data with small variations in wavelength/position:
                wavelen = wl + np.random.rand()
                posn = pos + np.random.rand()

                ratio = np.random.rand()
                ratio_error = 0.01 * ratio

                expected_ratios[wl_idx, pos_idx] = ratio
                expected_errors[wl_idx, pos_idx] = ratio_error

                R = np.sqrt((dv - posn[0]) ** 2 + (dh - posn[1]) ** 2)
                mask = R <= r
                expected_mask[wl_idx, pos_idx] = mask

                prnu.write_next_ratio_and_mask(ratio, ratio_error, mask.astype(int), posn, wavelen)

        prnu.close()

        ds = xr.open_dataset('temp.nc')
        vals = ds.data_vars.variables
        self.assertListEqual(vals[LunarPrnuSpotCheckOutput.EXPOSURE_CENTER_PIXEL_POSITIONS].values.tolist(),
                             positions.tolist())
        self.assertListEqual(vals[LunarPrnuSpotCheckOutput.CALIBRATION_WAVELENGTHS].values.tolist(),
                             wavelengths.tolist())
        self.assertListEqual(vals[LunarPrnuSpotCheckOutput.RATIO].values.tolist(),
                             expected_ratios.tolist())
        self.assertListEqual(vals[LunarPrnuSpotCheckOutput.RATIO_ERROR].values.tolist(),
                             expected_errors.tolist())
        self.assertListEqual(vals[LunarPrnuSpotCheckOutput.PIXEL_MASK].values.tolist(),
                             expected_mask.tolist())
        ds.close()
        os.remove('temp.nc')
